/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "misc.h"
#include "user.h"
#include "sock.h"
#include "chan.h"

char globsign;

#ifdef NOTDEF
void FixStamps()
{
  FILE *TheFile;
  FILE *NewFile;
  char ChanName[LINEBUFFSIZE], ModeString[LINEBUFFSIZE], ChanTopic[LINEBUFFSIZE];
  char str[LINEBUFFSIZE + 25];
  int i, s, line = 0;
  int TopicFreq, StrictOp, AutoOp, LamerMode, PublicCmds, EnfModes, PassMode,
  BanTimeout, Protect, ProtLev, Reserved4, Reserved5;
  time_t LastAccess, Stamp, CreationDate;
  ChannelListType ChanPtr;

  Debug(DBGSEVERE, "Were not using this (FixStamps) anymore, right?");
  TheFile = fopen(CHANFILENAME, "r");
  if (!TheFile)
  {
    Debug(DBGSEVERE, "ERROR OPENING chan.dat FILE!");
    return;
  }
  NewFile = fopen("tmp/.tempchan.dat", "w");
  if(!NewFile)
  {
     fclose(TheFile);
     Debug(DBGSEVERE, "Error opening temp file in fixstamps()!");
     return;
  }
  while ((s = fscanf(TheFile, "%s%d%d%d%d%d%d%d%d%d%d%d%d%lu%lu %[^:]%[^\n]\n",
		     ChanName, &TopicFreq, &StrictOp, &AutoOp,
		     &LamerMode, &PublicCmds, &EnfModes, &PassMode,
		     &BanTimeout, &Protect, &ProtLev, &Reserved4, &Reserved5,
		 &CreationDate, &LastAccess, ModeString, ChanTopic)) != EOF)
  {
    line++;
    if (s != 17)
    {
      Debug(DBGSEVERE, "*** ERROR: Param count error in FixStamps line %d! Only %d nybles read.\n", line, s);
      fclose(TheFile);
      fclose(NewFile);
      return;
    }
    ChanPtr = GetChannel(ChanName);
    if (LastAccess == 0l)
    {
      Stamp = time(NULL);
      ChanPtr->OwnersOnline = 0;
      ChanPtr->LastAccess = Stamp;
    }
    else
    {
      Stamp = LastAccess;
      ChanPtr->OwnersOnline = 0;
      ChanPtr->LastAccess = LastAccess;
    }
    fprintf(NewFile, "%s %d %d %d %d %d %d %d %d %d %d %d %d %lu %lu %s:%s\n",
	    ChanName, TopicFreq, StrictOp, AutoOp,
	    LamerMode, PublicCmds, EnfModes, PassMode,
	    BanTimeout, Protect, ProtLev, Reserved4, Reserved5,
	    CreationDate, Stamp, ModeString, &ChanTopic[1]);
  }
  fclose(TheFile);
  fclose(NewFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tempchan.dat", str);
}

#endif

OperSegment GetOpSegment(char *Channel)
{
  int Pos;
  unsigned long Key;
  char Str17[18];
  ChannelListType ChanPtr;
  char *UpCaseChan;
  char *TheChan;

  TheChan = UpCase(Channel);
  strncpy(Str17, &Channel[1], STRTOLSIZE);
  Str17[STRTOLSIZE] = '\0';
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % CHANSEGSIZE;
  if (Channels->Table[Pos])
  {
    ChanPtr = Channels->Table[Pos];
    while (ChanPtr)
    {
      if (!(strcmp((UpCaseChan = UpCase(ChanPtr->Name)), TheChan)))
      {
	free(TheChan);
	free(UpCaseChan);
	return (ChanPtr->Opers);
      }
      free(UpCaseChan);
      ChanPtr = ChanPtr->Next;
    }
  }
  free(TheChan);
  return (NULL);
}

void SetTopics(int period)
{
  int i;
  ChannelListType ChanPtr;

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    if (ChanPtr)
    {
      if (ChanPtr->settings[TOPICFREQ] == period)
      {
	sendtosock(Sock, ":%s TOPIC %s :%s", IrcNick, ChanPtr->Name, ChanPtr->Topic);
      }
      ChanPtr = ChanPtr->Next;
    }
  }
}



ChannelListType GetChannel(char *Channel)
{

  int Pos;
  unsigned long Key;
  char Str17[18];
  ChannelListType ChanPtr;
  char *UpCaseChan;
  char *TheChan;

  TheChan = UpCase(Channel);
  strncpy(Str17, &TheChan[1], STRTOLSIZE);
  Str17[STRTOLSIZE] = '\0';
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % CHANSEGSIZE;
  if (Channels->Table[Pos])
  {
    ChanPtr = Channels->Table[Pos];
    while (ChanPtr)
    {
      if (!(strcmp((UpCaseChan = UpCase(ChanPtr->Name)), TheChan)))
      {
	free(UpCaseChan);
	free(TheChan);
	return (ChanPtr);
      }
      free(UpCaseChan);
      ChanPtr = ChanPtr->Next;
    }
  }
  free(TheChan);
  return (NULL);
}

int IsProt(char *channel, char *hostmask)
{
  ChannelListType ChanPtr;
  OperListType OpList;
  OperSegment OpSeg;
  int i, count = 0;

  ChanPtr = GetChannel(channel);
  if (!ChanPtr)
  {
    Debug(DBGWARNING, "*** ERROR: channel %s not found in IsProt!\n", channel);
    return (0);
  }
  OpSeg = ChanPtr->Opers;
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      count++;
      if (!matches(&OpList->Address[1], hostmask) || !matches(hostmask, &OpList->Address[1]))
      {
	if (OpList->Address[0] != 'L')
	{
	  switch (OpList->Address[0])
	  {
          case 'P':
            return(10);
	  case 'L':
	    return (0);
	  case 'O':
	    return (100);
	  case 'M':
	    return (200);
          case 'C':
            return (290);
	  case 'W':
	    return (300);
	  default:
	    return (0);
	  }
	}
      }
      OpList = OpList->Next;
    }
  }
  return (0);
}

boolean IsChanLamer(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] == 'L')
    return (true);
  return (false);
}

boolean IsChanPeon(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] == 'P')
      return (true);
  return (false);
}  

boolean IsChanOp(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return(false);
  if (TempOperPtr->Address[0] == 'O')
    return (true);
  return (false);
}

int IsChanAOp(char *Channel, char *Nick)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->AutoOp)
    return (1);
  else
    return (0);
}

void SetChanAOp(char *Channel, char *Nick, int setting)
{
  char RegEx[LINEBUFFSIZE * 2];
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  UserListType UserNode;
  OperListType TempOperPtr;
  int i;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  TempOperPtr->AutoOp = setting;
  WriteChanOpFile(Channel);
  return;
}

int SetMaskNameByHost(char *Mask, char *Channel, char *NewHandle, int level)
{
  char RegEx[LINEBUFFSIZE * 2];
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  UserListType UserNode;
  OperListType TempOperPtr;
  int i;

  if (!(TheSegment = GetOpSegment(Channel)))
    return (false);
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
	if ((!matches(&TempOperPtr->Address[1], Mask)))
	{
	  switch (*TempOperPtr->Address)
	  {
	  case 'L':
	    break;
          case 'P':
            if (level > 100)
              break;
            else
              return (-1);
	  case 'O':
	    if (level > 100)
	      break;
	    else
	      return (-1);
	  case 'M':
	    if (level > 200)
	      break;
	    else
	      return (-1);
          case 'C':
            if(level > 290)
              break;
            else
              return (-1);
	  case 'W':
	    if (level >= 300)
	      break;
	    else
	      return (-1);
	  default:
	    break;
	  }
	  strncpy(TempOperPtr->Handle, NewHandle, 19);
	  TempOperPtr->Handle[19] = '\0';
	  WriteChanOpFile(Channel);
	  return (true);
	}
	TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  return (false);
}

void SetMaskName(char *Nick, char *Channel, char *NewHandle)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  strncpy(TempOperPtr->Handle, NewHandle, 19);
  TempOperPtr->Handle[19] = '\0';
  WriteChanOpFile(Channel);
}

void SetInfo(char *Nick, char *Channel, char *Info)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  safecopy(TempOperPtr->Info, Info, 199);
  WriteChanOpFile(Channel);
}

void SetLastSeen(char *Nick, char *Channel, int action)
{
  OperListType TempOperPtr;
  int i;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  TempOperPtr->LastSeen = time(NULL);
  if (action == JOIN)
    TempOperPtr->inchan = 1;
  if (action == PART)
    TempOperPtr->inchan = 0;
  WriteChanOpFile(Channel);
}

time_t GetSeenTime(char *Nick, char *Channel)
{
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  OperListType TempOperPtr;
  int i;
  time_t ts = 0;


  if (!(TheSegment = GetOpSegment(Channel)))
    return (0);
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
	if (!strcasecmp(Nick, TempOperPtr->Handle))
	{
	  if (TempOperPtr->LastSeen > ts)
	    ts = TempOperPtr->LastSeen;
	}
	TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  return (ts);
}

int IsInChan(char *Handle, char *Channel)
{
  OperSegment TheSegment;
  OperListType TempOperPtr;
  int i;
  int inchan;

  if (!(TheSegment = GetOpSegment(Channel)))
    return (FALSE);
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
	if (!strcasecmp(Handle, TempOperPtr->Handle))
	  if (TempOperPtr->inchan > 0)
	    return (TRUE);
	TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  return (FALSE);
}

void SendInfoLine(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if (NETBURSTING == TRUE)
    return;	/* dont want a flood when we first join */
  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if((TempOperPtr->Address[0] == 'W') || 
     (TempOperPtr->Address[0] == 'C') ||
     (TempOperPtr->Address[0] == 'M') || 
     (TempOperPtr->Address[0] == 'O') ||
     (TempOperPtr->Address[0] == 'P'))
  if (strcmp(TempOperPtr->Info, "*"))
     sendtosock(Sock, ":%s PRIVMSG %s :[%s] %s", IrcNick, Channel, Nick, TempOperPtr->Info);
}

char *GetInfo(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return(NULL);
  return (TempOperPtr->Info);
}

char *GetMaskName(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return(false);
  return (TempOperPtr->Handle);
}

OperListType FindBestUserMatchPass(char *Nick, char *Channel)
{
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  UserListType UserNode;
  OperListType TempOperPtr, PrevOperPtr = NULL;
  int i, weight, prevweight, lvl, prevlvl;

  if (!(TheSegment = GetOpSegment(Channel)))
  { /* Maybe no channel given, or regchan.. */
    return (NULL);
  }
  if (!(UserNode = FindNick(Nick, UsersN)))
  {
    return (NULL);
  }
  sprintf(Str, "%s!%s@%s", Nick, UserNode->UserID, UserNode->Address);
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
        if (!matches(&TempOperPtr->Address[1], Str))
        { /* If TempOperPtr matches our user, */
           weight=calcmaskweight(&TempOperPtr->Address[1]);
           lvl=userlevelbychar(TempOperPtr->Address[0]);
           if ( !PrevOperPtr || ( lvl > prevlvl) || (weight > prevweight && lvl >= prevlvl)) 
           {
              PrevOperPtr = TempOperPtr;
              prevweight = weight;
              prevlvl = lvl;
           }
        }
        TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  return(PrevOperPtr);
}

OperListType FindBestUserMatch(char *Nick, char *Channel)
{
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  UserListType UserNode;
  OperListType TempOperPtr, PrevOperPtr = NULL;
  int i, weight, prevweight, lvl, prevlvl;

  if (!(TheSegment = GetOpSegment(Channel)))
  { /* Maybe no channel given, or regchan.. */
    return (NULL);
  }
  if (!(UserNode = FindNick(Nick, UsersN)))
  {
    return (NULL);
  }
  sprintf(Str, "%s!%s@%s", Nick, UserNode->UserID, UserNode->Address);
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
        if (!matches(&TempOperPtr->Address[1], Str))
        { /* If TempOperPtr matches our user, */
           weight=calcmaskweight(&TempOperPtr->Address[1]);
           lvl=userlevelbychar(TempOperPtr->Address[0]);
           if(lvl != 0 && strcmp(TempOperPtr->Pass, "*") && !IsValidatedNode(UserNode, Channel))
           {
              weight=0;  /* If they need a pass and havnt validated, set the weight on this one low as possible */
              lvl = 0;
           }
           if ( !PrevOperPtr || ( lvl > prevlvl) || ( weight > prevweight && lvl >= prevlvl)) 
           {
              PrevOperPtr = TempOperPtr;
              prevweight = weight;
              prevlvl = lvl;
           }
        }
        TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  return(PrevOperPtr);
}

boolean IsChanOwner(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] == 'W')
     return (true);
  return (false);
}

boolean IsChanCoOwner(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] == 'C')
      return (true);
  return (false);
}   

boolean IsChanManager(char *Nick, char *Channel)
{
  OperListType TempOperPtr;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] == 'M')
     return (true);
  return (false);
}

boolean IsValidatedNode(UserListType UserNode, char *Channel)
{
  char *vstr;
  char *freeptr;
  char *tmpPtr;

  if (UserNode->Validated)
  {
    freeptr = vstr = (char *) malloc(strlen(UserNode->Validated) + 1);
    strcpy(vstr, UserNode->Validated);
/*    tmpPtr = (char *) strtok(vstr, " "); */
    tmpPtr = (char *) mysep(&vstr, " ");
    while (tmpPtr)
    {
      if (!strcasecmp(tmpPtr, Channel))
      {
         free(freeptr);
         return (true);
      }
/*      tmpPtr = (char *) strtok(NULL, " "); */
    tmpPtr = (char *) mysep(&vstr, " ");
    }
    free(freeptr);
    return (false);
  }
  return(false);
}

boolean IsValidated(char *Nick, char *Channel)
{
  UserListType UserNode;
  OperListType TempOperPtr;
  boolean FoundStar = false;
  char *tmpPtr;
  char *vstr, *freeptr;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return (false);
  if(!(TempOperPtr = FindBestUserMatchPass(Nick, Channel)))
    return;
  if (!strcmp(TempOperPtr->Pass, "*"))
    FoundStar = true;
  else if (UserNode->Validated)
  {
    freeptr = vstr = (char *) malloc(strlen(UserNode->Validated) + 1);
    strcpy(vstr, UserNode->Validated);
    tmpPtr = (char *) mysep(&vstr, " ");
    while (tmpPtr)
    {
      if (!strcasecmp(tmpPtr, Channel))
      {
         free(freeptr);
         return (true);
      }
      tmpPtr = (char *) mysep(&vstr, " ");
    }
    free(freeptr);
    return (false);
  } 
  if (FoundStar)
    return (true);
  return (false);
}

boolean PassOk(char *Nick, char *Channel, char *Pass)
{
  UserListType UserNode;
  OperListType TempOperPtr;
  int i;
  boolean FoundStar = false;
  char *tmpPtr;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return (false);
  if(!(TempOperPtr = FindBestUserMatchPass(Nick, Channel)))
    return;
  if (!strcmp(Pass, "*"))
    return (false);
  if (!strcmp(TempOperPtr->Pass, Pass))
  {
    if (!IsValidatedNode(UserNode, Channel))
    {
      tmpPtr = (char *) malloc(strlen(UserNode->Validated) + strlen(Channel) + 2);
      sprintf(tmpPtr, "%s %s", UserNode->Validated, Channel);
      free(UserNode->Validated);
      UserNode->Validated = tmpPtr;
    }
    return (true);
  }
  else
  {
    if (!strcmp(TempOperPtr->Pass, "*"))
      FoundStar = true;
  }
  if (FoundStar)
    return (true);
  return (false);
}

void ChangeAllHandlePass(char *Channel, char *OldPass, char *NewPass, char *Handle)
{
  int i;
  OperSegment TheSegment;
  OperListType TempOperPtr;

 if (!(TheSegment = GetOpSegment(Channel)))
    return;

  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
        if (!strcasecmp(TempOperPtr->Handle, Handle))
        {
           if(!strcmp(TempOperPtr->Pass, OldPass))
               strcpy(TempOperPtr->Pass, NewPass);
        }
        TempOperPtr = TempOperPtr->Next;
      }
    }
  }
  WriteChanOpFile(Channel);
}

boolean ChangePass(char *Nick, char *Channel, char *OldPass, char *NewPass)
{
  char OpAddress[LINEBUFFSIZE];
  char Pass[MAXPASS];
  UserListType UserNode;
  OperListType TempOperPtr;
  boolean FoundStar = false;
  char *tmpPtr;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return (false);
  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (!strcmp(TempOperPtr->Pass, OldPass))
  {
      ChangeAllHandlePass(Channel, OldPass, NewPass, TempOperPtr->Handle);
      tmpPtr = (char *) malloc(strlen(UserNode->Validated) + strlen(Channel) + 2);
      sprintf(tmpPtr, "%s %s", UserNode->Validated, Channel);
      free(UserNode->Validated);
      UserNode->Validated = tmpPtr;
      WriteChanOpFile(Channel);
      return (true);
  }
  return (false);
}

void ResetPass(char *Nick, char *Channel)
{
  OperListType TempOperPtr;
  boolean FoundStar = false;

  if(!(TempOperPtr = FindBestUserMatch(Nick, Channel)))
    return;
  if (TempOperPtr->Address[0] != 'L')
  {
    strcpy(TempOperPtr->Pass, "*");
    WriteChanOpFile(Channel);
    return;
  }
}

boolean IsDeop(char *Modes)
{

  char Operator = '+';
  int Len, i;

  Len = strlen(Modes);
  for (i = 0; i < Len; i++)
  {
    if (Modes[i] == '-')
      Operator = '-';
    if ((Modes[i] == 'o') && (Operator == '-'))
      return (true);
  }
  return (false);
}

boolean IsBan(char *Modes)
{
  char Operator = '+';
  int Len, i;

  Len = strlen(Modes);
  for (i = 0; i < Len; i++)
  {
    if (Modes[i] == '-')
      Operator = '-';
    if (Modes[i] == '+')
      Operator = '+';
    if ((Modes[i] == 'b') && (Operator == '+'))
      return (true);
  }
  return (false);
}

void ProcessKick(char *Nick, char *Kicked, char *Channel)
{
  ChannelListType ChanPtr;

  if (strcasecmp(Kicked, IrcNick) == 0)
  {				/* If its me who got kicked.. */
    sendtosock(Sock, ":%s PART %s", IrcNick, Channel);		/* This is required by protocol */
    sendtosock(Sock, ":%s WALLOPS :\002Ok, Very funny. Now tell %s to stop kicking me from %s\002", IrcNick, Nick, Channel);
    sendtosock(Sock, ":%s JOIN %s", IrcNick, Channel);
    sendtosock(Sock, ":%s MODE %s +o %s", ServiceName, Channel, IrcNick);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    Debug(DBGINFO, "ProcessKick chanel not found");
    return;                     /* chan not found?? its ok.. were not there*/
  }    
  ProcessPart(Nick, Channel);
  if (IsModeK(Nick))
    return;
  if(!ChanPtr->settings[INCHAN])
    return;
  if (!kickok(&ChanPtr, Nick, Kicked))
    punish(&ChanPtr, Nick, Kicked, 'K');
}

/*
   void ProcessChanMode(char *Nick, char *Channel, char *Modes, char *Arg4, char *Ban) {

   ChannelListType ChanNode;

   ChanNode = GetChannel(Channel);
   if (!ChanNode)
   return; 
   if (IsDeop(Modes)) {
   if ( (ENFORCEOPS) && (ChanNode->EnforceOps) && (IsChanOp(Arg4,Channel)) && (!IsChanOwner(Nick,Channel))) {
   sendtosock(Sock,":%s MODE %s +o %s",IrcNick,Channel,Arg4);
   if (!IsChanOp(Nick,Channel))
   sendtosock(Sock,":%s MODE %s -o %s",IrcNick,Channel,Nick);
   }
   else if ((ENFORCEOPS) && (ChanNode->EnforceOps) && (IsChanOwner(Arg4,Channel)) && (!IsChanOwner(Nick,Channel))) {
   sendtosock(Sock,":%s MODE %s +o %s",IrcNick,Channel,Arg4);
   sendtosock(Sock,":%s MODE %s -o %s",IrcNick,Channel,Nick); 
   }
   }

   if (IsBan(Modes)) 
   {
   sendtosock(Sock, ":%s PRIVMSG %s :\002%s sets ban: %s on %s\002", IrcNick, Channel, Nick, Modes, Arg4);

   }
   }
 */

void on_mode(char *from, char *rest)
{
  int did_change = FALSE;	/* If we have to change anything */
  int netsplit = FALSE;
  char *channel;
  char *mode;

  char *chanchars;		/* i.e. +oobli */
  char *params;
  char sign = '+';		/* +/-, i.e. +o, -b etc */

  char unminmode[MAXLEN];	/* used to undo modechanges */
  char unminparams[MAXLEN];
  char unplusmode[MAXLEN];	/* used to undo modechanges */
  char unplusparams[MAXLEN];

  /* these are here merely to make things clearer */
  char *nick = NULL;
  char *banstring;
  char *key;
  char *limit;
  static time_t currenttime;
  char buffer[MAXLEN];
  ChannelListType ChanNode;

  currenttime = getthetime();

  strcpy(unminmode, "");	/* min-modes, i.e. -bi */
  strcpy(unminparams, "");	/* it's parameters */
  strcpy(unplusmode, "");	/* plus-modes, i.e. +oo */
  strcpy(unplusparams, "");	/* parameters */

/*  channel = (char *) strtok(rest, " "); */
  channel = (char *) mysep(&rest, " ");
  ChanNode = GetChannel(channel);
  if (!ChanNode)		/* if I'm not in that chan */
    return;
  if(!ChanNode->settings[INCHAN])
    return;
  mode = rest; 
  params = (char *) strchr(mode, ' ') + 1;
  chanchars = (char *) mysep(&mode, " ");
  while (*chanchars)
  {
    switch (*chanchars)
    {
    case '+':
      sign = '+';
      globsign = '+';
      break;
    case '-':
      sign = '-';
      globsign = '-';
      break;
    case 'o':
      Debug(DBGINFO, "doing strtok(\"%s\",\" \")", params);
      nick = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (!nick)
	break;
      if (IsModeK(from))
	break;
      if (GetServer(from))
	break;
      if (sign == '+')
      {
	/* Do checks if we should remove the o */
	if (!opok(&ChanNode, from, nick))
	{
	  strcat(unminmode, "o");
	  strcat(unminparams, nick);
	  strcat(unminparams, " ");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	/* Do checks if we should give back the o */
	if (!deopok(&ChanNode, from, nick))
	{
	  strcat(unplusmode, "o");
	  strcat(unplusparams, nick);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	  punish(&ChanNode, from, nick, 'o');
	}
      }
      break;
#ifdef HYBRIDTS4
    case 'h':
      nick = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (!nick)
        break;
      if (IsModeK(from))
        break;
      if (GetServer(from))
        break;
      if (sign == '+')
      {
        /* Do checks if we should remove the h */
        if (!opok(&ChanNode, from, nick))
        {
          strcat(unminmode, "h");
          strcat(unminparams, nick);
          strcat(unminparams, " ");
          did_change = TRUE;
        }
      }
      else
        /* sign == '-' */
      {
        /* Do checks if we should give back the h */
        if (!deopok(&ChanNode, from, nick))
        {
          strcat(unplusmode, "h");
          strcat(unplusparams, nick);
          strcat(unplusparams, " ");
          did_change = TRUE;
          punish(&ChanNode, from, nick, 'o');
        }
      }
      break;
#endif
    case 'b':
      banstring = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (!banstring)
	break;
      if (sign == '+')
      {
	if (!IsModeK(from))	/* If its not a service doing it */
	  if (!banok(&ChanNode, from, banstring))
	  {
	    strcat(unminmode, "b");
	    strcat(unminparams, banstring);
	    strcat(unminparams, " ");
	    did_change = TRUE;
	    punish(&ChanNode, from, banstring, 'b');
	    break;
	  }
	add_ban(&ChanNode->banned, banstring);
      }
      else
	/* sign == '-' */
      {
	delete_ban(&ChanNode->banned, banstring);
      }
      break;
    case 'v':
      nick = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      break;
    case 'p':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEp)
	{
	  strcat(unminmode, "p");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEp)
	{
	  strcat(unplusmode, "p");
	  did_change = TRUE;
	}
      }
      break;
    case 's':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEs)
	{
	  strcat(unminmode, "s");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEs)
	{
	  strcat(unplusmode, "s");
	  did_change = TRUE;
	}
      }
      break;
    case 'm':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEm)
	{
	  strcat(unminmode, "m");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEm)
	{
	  strcat(unplusmode, "m");
	  did_change = TRUE;
	}
      }
      break;
    case 't':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEt)
	{
	  strcat(unminmode, "t");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEt)
	{
	  strcat(unplusmode, "t");
	  did_change = TRUE;
	}
      }
      break;
    case 'i':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEi)
	{
	  strcat(unminmode, "i");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEi)
	{
	  strcat(unplusmode, "i");
	  did_change = TRUE;
	}
      }
      break;
    case 'n':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	if (ChanNode->MinModes & CMODEn)
	{
	  strcat(unminmode, "n");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEn)
	{
	  strcat(unplusmode, "n");
	  did_change = TRUE;
	}
      }
      break;
    case 'k':
      key = (char *) strtok(params, " ");
      params = (char *) strtok(NULL, "\0\n");
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	strncpy(ChanNode->CurrentKey, key, 25);
	if (ChanNode->MinModes & CMODEk)
	{
	  strcat(unminmode, "k");
	  strcat(unminparams, key);
	  strcat(unminparams, " ");
	  did_change = TRUE;
	}
      }
      else
	/* sign == '-' */
      {
	strcpy(ChanNode->CurrentKey, "");
	if (ChanNode->PlusModes & CMODEk)
	{
	  strcat(unplusmode, "k");
	  strcat(unplusparams, ChanNode->Key);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      break;
#ifdef HYBRIDTS4
    case 'c':
      break;
#endif /* HYBRIDTS4 */
    case 'l':
      if (modeok(&ChanNode, from))
	break;
      if (sign == '+')
      {
	limit = (char *) strtok(params, " ");
	params = (char *) strtok(NULL, "\0\n");
	if (ChanNode->MinModes & CMODEl)
	{
	  strcat(unminmode, "l");
	  did_change = TRUE;
	}
	else if ((ChanNode->PlusModes & CMODEl) && strcmp(limit, ChanNode->Limit))
	{
	  strcat(unplusmode, "l");
	  strcat(unplusparams, ChanNode->Limit);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      else /* sign == '-' */
      {
	if (ChanNode->PlusModes & CMODEl)
	{
	  strcat(unplusmode, "l");
	  strcat(unplusparams, ChanNode->Limit);
	  strcat(unplusparams, " ");
	  did_change = TRUE;
	}
      }
      break;
    default:
      Debug(DBGWARNING, "Unknown channel mode %c [%x] in %s", *chanchars, *chanchars, channel);
      break;
    }
    chanchars++;
  }
  /* restore unwanted modechanges */
  if (did_change)
  {
    sendtosock(Sock, ":%s MODE %s -%s+%s %s %s", IrcNick, channel, unminmode,
	       unplusmode, unminparams, unplusparams);
  }
  return;
}

void ConvertModeString(char *ModeString, int *PlusModes, int *MinModes,
		       char **Key, char **limit)
{
  int *modeptr, *optr;
  char *modes;
  char *tmpstr;

  if(!ModeString || !*ModeString)
     return;
  modes = (char *) mysep(&ModeString, " ");
  modeptr = PlusModes;
  optr = MinModes;
  *PlusModes = 0;
  *MinModes = 0;
  if (*Key)
  {
    free(*Key);
    *Key = NULL;
  }
  if (*limit)
  {
    free(*limit);
    *limit = NULL;
  }
  while (*modes)
  {
    switch (*modes)
    {
    case '+':
      modeptr = PlusModes;
      optr = MinModes;
      break;
    case '-':
      modeptr = MinModes;
      optr = PlusModes;
      break;
    case 'n':
      *modeptr = *modeptr | CMODEn;
      *optr = *optr & ~CMODEn;
      break;
    case 't':
      *modeptr = *modeptr | CMODEt;
      *optr = *optr & ~CMODEt;
      break;
    case 's':
      *modeptr = *modeptr | CMODEs;
      *optr = *optr & ~CMODEs;
      break;
    case 'p':
      *modeptr = *modeptr | CMODEp;
      *optr = *optr & ~CMODEp;
      break;
    case 'i':
      *modeptr = *modeptr | CMODEi;
      *optr = *optr & ~CMODEi;
      break;
    case 'm':
      *modeptr = *modeptr | CMODEm;
      *optr = *optr & ~CMODEm;
      break;
    case 'k':
      if (modeptr == MinModes)
      {
	*modeptr = *modeptr | CMODEk;
	*optr = *optr & ~CMODEk;
	break;
      }
      if (tmpstr = (char *) mysep(&ModeString, " "))
      {
	if (KeyOk(tmpstr))
	{
	  if (*Key)
	    break;		/* already got first one.. */
	  *Key = malloc(strlen(tmpstr) + 2);
	  if (!(*Key))
	    return;		/* out of memory! */
	  strcpy(*Key, tmpstr);
	  *modeptr = *modeptr | CMODEk;
	}
      }
      break;
    case 'l':
      if (modeptr == MinModes)
      {
	*modeptr = *modeptr | CMODEl;
	*optr = *optr & ~CMODEl;
	break;
      }
      if (tmpstr = (char *) mysep(&ModeString, " "))
      {
	if (IsAllDigits(tmpstr) && (atol(tmpstr) != 0))
	{
	  if (*limit)
	    break;		/* already got firstone */
	  *limit = malloc(strlen(tmpstr) + 2);
	  if (!(*limit))
	    return;		/* out of memory! */
	  strcpy(*limit, tmpstr);
	  *modeptr = *modeptr | CMODEl;
	}
      }
      break;
    default:
      break;
    }
    modes++;
  }
}

#ifdef HYBRIDTS4
void ProcessJoin(char *Nick, char *Chan, char *SJmodes)
#else  /* HYBRIDTS4 */
void ProcessJoin(char *Nick, char *Chan)
#endif /* HYBRIDTS4 */
{				/* on_join */
  /* note anything that calls strtok cannot be used
   * in this function or it will break the calling
   * function. I'll fix this at some point */
  ChannelListType ChanNode;
  UserListType UserNode;
  OwnerListType NewChanList;
  char RegEx[LINEBUFFSIZE * 2];
  char Str[LINEBUFFSIZE];
  OperSegment TheSegment;
  OperListType TempOperPtr, UsrPtr;
  int i, gaveops=0;
  time_t seentime;
  char *MaskNick;

  Debug(DBGINFO, "Processing Join: %s joined %s", Nick, Chan);
  UserNode = FindNick(Nick, UsersN);
  if (!UserNode)
    return;
  if (strlen(Chan) < 2)
  {
    return;
  }
  if (!(ChanNode = GetChannel(Chan)))
  {                                            /* AutoOp Everyone.. */
     return;
  }
  if(!ChanNode->settings[INCHAN])                                                                   
    return; /* Channel not active */
  if(strcmp(ChanNode->URL, "*"))
  {
     sendtosock(Sock, ":%s 328 %s %s %s", ServiceName, Nick, ChanNode->Name, ChanNode->URL);
  }
#ifdef HYBRIDTS4
    if ((ChanNode->settings[AUTOOP] == 3) && !(strchr(SJmodes, 'o')))
#else
    if (ChanNode->settings[AUTOOP] == 3)
#endif /* HYBRIDTS4 */
    {
      sendtosock(Sock, ":%s MODE %s +o %s", IrcNick, Chan, Nick);
      gaveops++;
    }
  if(UsrPtr = FindBestUserMatch(Nick, Chan))
  {
    seentime = GetSeenTime(UsrPtr->Handle, Chan);
  }
  else 
    seentime = 0;
  if (seentime != 0)
    if ((seentime < NetMessage.Timestamp) && userlevel(Nick, Chan) >= 200)
      SendNetMessage(Nick);
  SetLastSeen(Nick, Chan, JOIN);

/* If they are (co)owner, mark that one has joined. */
  if(IsChanOwner(Nick, Chan) || IsChanCoOwner(Nick, Chan))
  {
     ChanNode->OwnersOnline++;   
     if(ChanNode->OwnersOnline > 0)
        ChanNode->LastAccess = (unsigned long) 0;
     WriteChanOpFile(Chan);
  }
  if (IsChanOwner(Nick, Chan) || IsChanCoOwner(Nick, Chan) || IsChanManager(Nick, Chan))
  {
    if (((ChanNode->settings[PASSMODE] == 1) || (IsValidated(Nick, Chan))) && IsChanAOp(Chan, Nick)
#ifdef HYBRIDTS4
 && !(strchr(SJmodes, 'o'))
#endif
        )
    {
      switch (ChanNode->settings[AUTOOP])
      {
      case 0:
	break;
      case 1:
	if (userlevel(Nick, Chan) >= 200)
        {
	  sendtosock(Sock, ":%s MODE %s +o %s", IrcNick, Chan, Nick);
          gaveops++;
        }
	break;
      case 2:
	if (userlevel(Nick, Chan) >= 100)
        {
	  sendtosock(Sock, ":%s MODE %s +o %s", IrcNick, Chan, Nick);
          gaveops++;
        }
	break;
      case 3:
	break;			/* handled above prior to this */
      default:
	break;
      }
    }
    SendInfoLine(Nick, Chan);
    NewChanList = (OwnerListType) malloc(sizeof(struct OwnerStruct));

    if (!NewChanList)
      return;			/* no memory free */
    NewChanList->Channel = (char *) malloc(strlen(Chan) + 1);
    strcpy(NewChanList->Channel, Chan);
    NewChanList->Next = UserNode->Chans;
    UserNode->Chans = NewChanList;
  }
  else if (IsChanOp(Nick, Chan))
  {
    if (((ChanNode->settings[PASSMODE] == 1) || (IsValidated(Nick, Chan))) && IsChanAOp(Chan, Nick) 
#ifdef HYBRIDTS4
&& !(strchr(SJmodes, 'o'))
#endif /* HYBRIDTS4 */
          )
    {
      switch (ChanNode->settings[AUTOOP])
      {
      case 0:
	break;
      case 1:
	break;
      case 2:
	if (userlevel(Nick, Chan) >= 100)
        {
	  sendtosock(Sock, ":%s MODE %s +o %s", IrcNick, Chan, Nick);
          gaveops++;
        }
	break;
      case 3:
	break;			/* handled above prior to this */
      default:
	break;
      }
    }
    if(ChanNode->settings[USERINFO])
      SendInfoLine(Nick, Chan);
    NewChanList = (OwnerListType) malloc(sizeof(struct OwnerStruct));

    if (!NewChanList)
      return;			/* no memory free */
    NewChanList->Channel = (char *) malloc(strlen(Chan) + 1);
    strcpy(NewChanList->Channel, Chan);
    NewChanList->Next = UserNode->Chans;
    UserNode->Chans = NewChanList;
  }
  else if (IsChanPeon(Nick, Chan))
  {
    if (((ChanNode->settings[PASSMODE] == 1) || (IsValidated(Nick, Chan))) && IsChanAOp(Chan, Nick))
    {
       if(ChanNode->settings[VOICE] == 1)
       {
         sendtosock(Sock, ":%s MODE %s +v %s", IrcNick, Chan, Nick);
         gaveops++;
       }
    }
    SendInfoLine(Nick, Chan);
    NewChanList = (OwnerListType) malloc(sizeof(struct OwnerStruct));

    if (!NewChanList)
      return;                   /* no memory free */
    NewChanList->Channel = (char *) malloc(strlen(Chan) + 1);
    strcpy(NewChanList->Channel, Chan);
    NewChanList->Next = UserNode->Chans;
    UserNode->Chans = NewChanList;
  }  
  else if (IsChanLamer(Nick, Chan))
  {
    if (!(TheSegment = GetOpSegment(Chan)))
      return;
    if (!(UserNode = FindNick(Nick, UsersN)))
      return;
    if (IsModeK(Nick))
      return;
    if(!(TempOperPtr = FindBestUserMatch(Nick, Chan)))
      return;
    deopban(Chan, Nick, &TempOperPtr->Address[1]);
    sendtosock(Sock, ":%s KICK %s %s :%s", IrcNick, Chan, Nick, TempOperPtr->Info);
    NewChanList = (OwnerListType) malloc(sizeof(struct OwnerStruct));

    if (!NewChanList)
    {
      Debug(DBGSEVERE, "No memory free to allocate NewChanList!");
      return;			/* no memory free */
    }
    if(NewChanList->Channel = (char *) malloc(strlen(Chan) + 1))
    {
      strcpy(NewChanList->Channel, Chan);
      NewChanList->Next = UserNode->Chans;
      UserNode->Chans = NewChanList;
    }
    else
    {
        Debug(DBGSEVERE, "No memory free to callocate NewChanList->Channel!");
        return;
    }
    return;
  }
#ifdef HYBRIDTS4
  if((ChanNode->settings[VOICE] == 2) && !(strchr(SJmodes, 'v')))
#else
  if(ChanNode->settings[VOICE] == 2)
#endif /* HYBRIDTS4 */
  {
     if(!gaveops)
       sendtosock(Sock, ":%s MODE %s +v %s", IrcNick, Chan, Nick);
  }
}

void ProcessPart(char *Nick, char *Chan)
{
  ChannelListType ChanPtr;
  UserListType UserNode;
  OwnerListType OwnerList, ChanTail;
  char *UpCaseChan;
  char *UpCaseStr;

  ChanPtr = GetChannel(Chan);
  if (!ChanPtr)
    return;
  if(!ChanPtr->settings[INCHAN])
    return;
  SetLastSeen(Nick, Chan, PART);
  UserNode = FindNick(Nick, UsersN);
  if (!UserNode)
    return;
  if(IsChanOwner(Nick, Chan) || IsChanCoOwner(Nick, Chan))
  {
     ChanPtr->OwnersOnline--;
     if(ChanPtr->OwnersOnline < 0)
        ChanPtr->OwnersOnline = 0;
     if(ChanPtr->OwnersOnline == 0)
        ChanPtr->LastAccess = time(NULL);
     WriteChanOpFile(Chan);
  }           
  UpCaseChan = UpCase(Chan);
  OwnerList = UserNode->Chans;
  ChanTail = NULL;
  while (OwnerList)
  {
    if (!strcasecmp(OwnerList->Channel, UpCaseChan))
      break;
    else
    {
      ChanTail = OwnerList;
      OwnerList = OwnerList->Next;
    }
  }
  if (!OwnerList)
    return;
  if (!ChanTail)
    UserNode->Chans = OwnerList->Next;
  else
    ChanTail->Next = OwnerList->Next;
  free(OwnerList->Channel);
  free(OwnerList);
  free(UpCaseChan);
}

void KillChanList()
{
  ChannelCount = 0;
}

int DelChan(char *Channel)
{
  ChannelListType CurrPtr, TailPtr;
  OperListType OpList, TempOpList, TempAdd;
  OperSegment OpSeg;

  int Pos, i;
  long unsigned Key;
  char *UpCaseChan;
  char *UpCaseStr;
  char Str17[18];

  TailPtr = NULL;

  UpCaseChan = UpCase(Channel);
  strncpy(Str17, &UpCaseChan[1], STRTOLSIZE);
  Str17[STRTOLSIZE] = '\0';
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % CHANSEGSIZE;

  CurrPtr = Channels->Table[Pos];
  while (CurrPtr)
  {
    if (!strcmp(UpCaseStr = UpCase(CurrPtr->Name), UpCaseChan))
    {
      free(UpCaseStr);
      break;
    }
    free(UpCaseStr);
    TailPtr = CurrPtr;
    CurrPtr = CurrPtr->Next;
  }
  free(UpCaseChan);
  if (!CurrPtr)
    return (0);
  if (!TailPtr)
    Channels->Table[Pos] = CurrPtr->Next;
  else
    TailPtr->Next = CurrPtr->Next;
  OpSeg = CurrPtr->Opers;
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      TempOpList = OpList;
      OpList = OpList->Next;
      if (TempOpList->Address[0] == 'P')
        PeonCount--;
      else if (TempOpList->Address[0] == 'O')
	OpCount--;
      else if (TempOpList->Address[0] == 'M')
	ManagerCount--;
      else if (TempOpList->Address[0] == 'L')
	LamerCount--;
      else if (TempOpList->Address[0] == 'C')
        CoOwnerCount--;
      else if (TempOpList->Address[0] == 'W')
	OwnerCount--;
      free(TempOpList->Address);
      free(TempOpList);
    }
  }
  if (OpSeg)
    free(OpSeg);
  free(CurrPtr->Name);
  free(CurrPtr->Topic);
  free(CurrPtr->URL);
  free(CurrPtr->Description);
  free(CurrPtr->Reggedby);
  if (CurrPtr->Key)
    free(CurrPtr->Key);
  if (CurrPtr->Limit)
    free(CurrPtr->Limit);
  free(CurrPtr);
  ChannelCount--;
}

int AddChan(char *ChanName, char *ModeString, char *ChanTopic, 
            char *Description, char *URL, char *ReggedBy,
            int settings[NUMOFCHANSETTINGS],
	    time_t CreationDate, time_t LastAccess,
	    OperSegment NewOperSeg)
{
  ChannelListType NewChan;
  char Str17[18];
  unsigned long Key;
  int Pos, i;
  char *UpCaseChan;

  NewChan = (ChannelListType) malloc(sizeof(struct ChanStruct));

  if (!NewChan)
  {
    printf("Could not malloc ChanStruct! Terminating AddChan\n");
    return (0);
  }

  NewChan->Name = (char *) malloc(strlen(ChanName) + 1);
  if (!NewChan->Name)
  {
    printf("Could not malloc Name! Terminating AddChan\n");
    return (0);
  }
  NewChan->Topic = (char *) malloc(strlen(ChanTopic) + 1);
  if (!NewChan->Topic)
  {
    printf("Could not malloc Topic!  Terminating AddChan\n");
    return (0);
  }
  NewChan->URL = (char *) malloc(strlen(URL) + 1);
  if(!NewChan->URL)
  {
    printf("Could not malloc URL! Terminating AddChan\n");
    return;
  }
  NewChan->Description = (char *) malloc(strlen(Description) + 1);
  if(!NewChan->Description)
  {
    printf("Could not malloc Description! Terminating AddChan\n");
    return;
  }
  NewChan->Reggedby = (char *) malloc(strlen(ReggedBy) + 1);                               
  if(!NewChan->Reggedby)
  {                                                                                                 
    printf("Could not malloc Reggedby! Terminating AddChan\n");                               
    return;                                                                                         
  }
  strcpy(NewChan->Name, ChanName);
  strcpy(NewChan->Topic, ChanTopic);
  strcpy(NewChan->URL, URL);
  strcpy(NewChan->Description, Description);
  strcpy(NewChan->Reggedby, ReggedBy);
  NewChan->Key = NULL;
  NewChan->Limit = NULL;
  ConvertModeString(ModeString, &NewChan->PlusModes, &NewChan->MinModes, &NewChan->Key, &NewChan->Limit);
  for(i=0;i<NUMOFCHANSETTINGS;i++)
    NewChan->settings[i] = settings[i];
  NewChan->OwnersOnline = 0;
  NewChan->Opers = NewOperSeg;
  NewChan->LastAccess = LastAccess;
  NewChan->CreationDate = CreationDate;
  NewChan->banned = NULL;

  ChannelCount++;
  UpCaseChan = UpCase(ChanName);
  strncpy(Str17, &UpCaseChan[1], STRTOLSIZE);
  Str17[STRTOLSIZE] = '\0';
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % CHANSEGSIZE;
  free(UpCaseChan);
  NewChan->Next = Channels->Table[Pos];
  Channels->Table[Pos] = NewChan;
  return 1;
}

OperListType FindOp(OperSegment TheSegment, char *OperAddy)
{
  char Str17[18];
  int Pos;
  unsigned long OperKey;
  OperListType OperList, TempListPtr;
  char *UpCaseAdd;
  char *UpCaseStr;

  UpCaseAdd = UpCase(OperAddy);
  GetStr17(Str17, UpCaseAdd);
  OperKey = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = OperKey % OPERSEGSIZE;
  if (TheSegment->Table[Pos] != NULL)
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if (!(strcmp(UpCaseStr = UpCase(&TempListPtr->Address[1]), UpCaseAdd)))
      {
	free(UpCaseAdd);
	free(UpCaseStr);
	return (TempListPtr);
      }
      else
      {
	free(UpCaseStr);
	TempListPtr = TempListPtr->Next;
      }
    }
  }
  free(UpCaseAdd);
  return (0);
}

int AddOp(OperSegment NewOperSeg, char *OperAddy,
	  char *Nick, char *Pass, char *Info,
	  int AutoOp, time_t LastSeen)
{
  char Str17[18];
  unsigned long OperKey;
  char *NewAddPtr;
  OperListType NewOperList;
  char CleanAdd[LINEBUFFSIZE];
  int Pos;
  char *UpCaseAdd;

  if (!AddyOk(&OperAddy[1]))
    return (0);
  if (!HandleOk(Nick))
    return (0);
  if (NewOperList = FindOp(NewOperSeg, &OperAddy[1]))
    return (-1);		/* alreayd in list */
  if (!(NewAddPtr = (char *) malloc(strlen(OperAddy) + 1)))
    return (0);
  if (!(NewOperList = (OperListType) malloc(sizeof(struct OperStruct))))
     return (0);

  UpCaseAdd = UpCase(&OperAddy[1]);
  GetStr17(Str17, UpCaseAdd);
  free(UpCaseAdd);
  OperKey = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = OperKey % OPERSEGSIZE;
  strcpy(NewAddPtr, OperAddy);
  NewOperList->Address = NewAddPtr;
  safecopy(NewOperList->Pass, Pass, 19);
  safecopy(NewOperList->Handle, Nick, 19);
  safecopy(NewOperList->Info, Info, MAXINFOLEN);
  NewOperList->AutoOp = AutoOp;
  NewOperList->LastSeen = LastSeen;
  NewOperList->inchan = 0;
  NewOperList->Next = NewOperSeg->Table[Pos];
  NewOperSeg->Table[Pos] = NewOperList;
  if (OperAddy[0] == 'P')
    PeonCount++;
  else if (OperAddy[0] == 'O')
    OpCount++;
  else if (OperAddy[0] == 'M')
    ManagerCount++;
  else if (OperAddy[0] == 'C')
    CoOwnerCount++;
  else if (OperAddy[0] == 'W')
    OwnerCount++;
  else if (OperAddy[0] == 'L')
    LamerCount++;
  else
  {
    Debug(DBGSEVERE, "Serious data problem. No P, O, M, C, W , or L start in an oper file entry.\n");
    return (0);
  }
  return (1);
}

void WriteChanOpFile(char *Channel)
{
  char sets_str[MAXLEN];
  ChannelListType ChanPtr;
  OperListType OpList;
  OperSegment OpSeg;
  int i;
  FILE *FilePtr;
  char str[MAXLEN], *UpCaseChan;
  char ModeString[MAXLEN];

  if (!ChanOk(Channel))
    return;  /* Not ok channel name*/
  ChanPtr = GetChannel(Channel);
  if (!ChanPtr)
  {
      Debug(DBGSEVERE, "*** ERROR WRITING CHANNEL USERFILE!! %s not found in memory!\n", Channel);
      return;
  }
  FilePtr = fopen("./tmp/.tmpopfile", "w");
  if(!FilePtr)
  {
     Debug(DBGSEVERE, "*** ERROR OPENING TEMP OPFILE! :%s", strerror(errno));
     return;
  }
  clearerr(FilePtr);
/* Compile a string of setting values... */
  strcpy(sets_str, "");
  for(i=0;i<NUMOFCHANSETTINGS;i++)
  {
    char buff[MAXLEN];
    sprintf(buff, "%d", ChanPtr->settings[i]);
    if(i>0)
      strcat(sets_str, " ");
    strcat(sets_str, buff);
  }
  /* These MUST all start with different letters! the read fucntion is 
   * optomized to look only at (T) for topic, (M) for MOTD etc.
   * Commented lines are not yet implimented (but this format
   * is great coz when they are nothing will break!)
   */
  fprintf(FilePtr, "VERSION 1\n");
  fprintf(FilePtr, "CHANNEL %s\n", ChanPtr->Name);
  fprintf(FilePtr, "TOPIC %s\n", ChanPtr->Topic);
  fprintf(FilePtr, "DESCRIPTION %s\n", ChanPtr->Description);
  fprintf(FilePtr, "WEB %s\n", ChanPtr->URL);
  fprintf(FilePtr, "MODES %s\n", GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
  fprintf(FilePtr, "SETTINGS %s\n", sets_str);
  fprintf(FilePtr, "LASTVISITED %lu\n", ChanPtr->LastAccess);
  fprintf(FilePtr, "REGEDBY %s %lu\n", ChanPtr->Reggedby , ChanPtr->CreationDate);
  /* End of header section.. on to USER section */
  OpSeg = ChanPtr->Opers;
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      if (!OpList->Address)
      {
        sendtosock(Sock, ":%s WALLOPS :\002Channel memory corruption in %s terminating!!",
                   IrcNick, Channel);
        Debug(DBGSEVERE, "No data in non-null OpList variable. Memory is corrupted bailing out without saving channel data!");
        exit(2);
      }
      fprintf(FilePtr, "USER %s %d %lu %s %s %s\n",
              OpList->Address, OpList->AutoOp,
              OpList->LastSeen, OpList->Handle,
              OpList->Pass, OpList->Info);
      OpList = OpList->Next;
    }
  }
  if(ferror(FilePtr))
  {
     Debug(DBGSEVERE, "Error writing to a tempopfile (Out of disk space?). %s data NOT saved.", Channel);
     fclose(FilePtr);
     sleep(4);
     exit(2);
  }
  fclose(FilePtr);
  UpCaseChan = UpCase(Channel);
  sprintf(str, "./CHANS/%s", &UpCaseChan[1]);
  unlink(str);
  rename("./tmp/.tmpopfile", str);
  free(UpCaseChan);
}

void OldWriteChanOpFile(char *Channel)
{
  ChannelListType ChanPtr;
  OperListType OpList;
  OperSegment OpSeg;
  int i;
  FILE *FilePtr;
  char str[MAXLEN], *UpCaseChan;

  if (!ChanOk(Channel))
    return;
  ChanPtr = GetChannel(Channel);
  if (!ChanPtr)
  {
    Debug(DBGSEVERE, "*** ERROR WRITING CHANNEL USERFILE!! %s not found in memory!\n", Channel);
    return;
  }
  FilePtr = fopen("./tmp/.tmpopfile", "w");
  if(!FilePtr)
  {
     Debug(DBGSEVERE, "*** ERROR OPENING TEMP OPFILE! :%s", strerror(errno));
     return;
  }
  OpSeg = ChanPtr->Opers;
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      if (!OpList->Address)
      {
	sendtosock(Sock, ":%s WALLOPS :\002Channel memory corruption in %s terminating!!",
		   IrcNick, Channel);
        Debug(DBGSEVERE, "No data in non-null OpList variable. Memory is corrupted bailing out without saving channel data!");
	exit(2);
      }
      fprintf(FilePtr, "%s %d %lu %s %s %s\n",
	      OpList->Address, OpList->AutoOp,
	      OpList->LastSeen, OpList->Handle,
	      OpList->Pass, OpList->Info);
      OpList = OpList->Next;
    }
  }
  fclose(FilePtr);
  UpCaseChan = UpCase(Channel);
  sprintf(str, "./CHANS/%s", &UpCaseChan[1]);
  unlink(str);
  rename("./tmp/.tmpopfile", str);

  free(UpCaseChan);
}

int FDelUserByHost(OperSegment OpSeg, char *OperAddy, char *Channel, char *Nick)
{
  ChannelListType ChanPtr;
  OperListType OpList, tmp;
  int i, count = 0;

  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      tmp = OpList->Next;
      if (!(matches(&OpList->Address[1], &OperAddy[1])))
	if (OpList->Address[0] == OperAddy[0])
	{
	  send_to_user(Nick, "\002Deleting matching address [%s]", &OpList->Address[1]);
	  DelUserSeg(OpSeg, OpList->Address);
	  count++;
	}
      OpList = tmp;
    }
  }
  if (count < 1)
    return (FALSE);
  else
    WriteChanOpFile(Channel);
  return (TRUE);
}

int FDelUser(OperSegment OpSeg, char *OperAddy, char *Channel, char *Nick)
{
  ChannelListType ChanPtr;
  OperListType OpList, tmp;
  int i, count = 0;

  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      tmp = OpList->Next;
      if (!(matches(&OperAddy[1], &OpList->Address[1])))
      {
	if (OpList->Address[0] == OperAddy[0])
	{
	  send_to_user(Nick, "\002Deleting matching address [%s]", &OpList->Address[1]);
	  DelUserSeg(OpSeg, OpList->Address);
	  count++;
	}
      }
      OpList = tmp;
    }
  }
  if (count < 1)
    return (FALSE);
  else
  {
    WriteChanOpFile(Channel);
    return (TRUE);
  }
}

int DelUser(OperSegment OpSeg, char *OperAddy, char *Channel, char *Nick)
{
  if (DelUserSeg(OpSeg, OperAddy))
  {
    WriteChanOpFile(Channel);
    return (TRUE);
  }
  return (FALSE);
}

int DelUserSeg(OperSegment OpSeg, char *OperAddy)
{	/* Oper Address must be upcase.  first char is W for owner L for lamer etc */
  char Str17[18];
  unsigned long OperKey;
  OperListType OperList, CurrPtr, TailPtr;
  int Pos;

  if (!OperAddy)
    return (FALSE);

  TailPtr = NULL;
  if (!(OperList = FindOp(OpSeg, &OperAddy[1])))
    return (FALSE);

  GetStr17(Str17, &OperAddy[1]);
  OperKey = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = OperKey % OPERSEGSIZE;

  CurrPtr = OpSeg->Table[Pos];
  while (CurrPtr)
  {
    if (!strcasecmp(CurrPtr->Address, OperAddy))
      break;
    else
    {
      TailPtr = CurrPtr;
      CurrPtr = CurrPtr->Next;
    }
  }
  if (!CurrPtr)
    return (FALSE);
  if (!TailPtr)
    OpSeg->Table[Pos] = CurrPtr->Next;
  else
    TailPtr->Next = CurrPtr->Next;
  free(CurrPtr);
  return (TRUE);
}

char *ConvertChanMode(char *StrModes, unsigned char Modes)
{

  unsigned char Mode;
  char *tmp;

  tmp = StrModes;
  Mode = CMODEn;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'n';

  Mode = CMODEt;
  if ((Mode & Modes) != 0)
    *StrModes++ = 't';

  Mode = CMODEs;
  if ((Mode & Modes) != 0)
    *StrModes++ = 's';

  Mode = CMODEp;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'p';

  Mode = CMODEi;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'i';

  Mode = CMODEm;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'm';

  Mode = CMODEk;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'k';

  Mode = CMODEl;
  if ((Mode & Modes) != 0)
    *StrModes++ = 'l';

  *StrModes = '\0';
  return (tmp);
}

char *GetModeString(char *ModeString, int PlusModes, int MinModes, char *key, char *limit)
{
  char tmpstr[LINEBUFFSIZE];

  strcpy(ModeString, "");
  if (PlusModes)
  {
    strcat(ModeString, "+");
    strcat(ModeString, ConvertChanMode(tmpstr, PlusModes));
  }
  else
  {  /* Always put the + */
      strcat(ModeString, "+");
  }
  if (MinModes)
  {
    strcat(ModeString, "-");
    strcat(ModeString, ConvertChanMode(tmpstr, MinModes));
  }
  if (PlusModes & CMODEk)
  {
    strcat(ModeString, " ");
    strcat(ModeString, key);
  }
  if (PlusModes & CMODEl)
  {
    strcat(ModeString, " ");
    strcat(ModeString, limit);
  }
  if(!*ModeString) /* Empty mode is bad in the data files.. put a + in there */
     strcat(ModeString, "+");
  return (ModeString);
}

void add_ban(BAN_list ** b_list, char *banstring)
{
  BAN_list *New_ban;

  if ((New_ban = (BAN_list *) malloc(sizeof(*New_ban))) == NULL)
    return;
  if (search_ban(b_list, banstring) != NULL)
    return;
  strncpy(New_ban->banstring, banstring, 69);
  New_ban->banstring[69] = '\0';
  New_ban->time = time(NULL);
  Debug(DBGINFO, "*** Adding a ban to the list. banstring=%s\n", banstring);
  New_ban->next = *b_list;
  *b_list = New_ban;
}

int delete_ban(BAN_list ** b_list, char *banstring)
{
  BAN_list **old;
  BAN_list *Dummy;

  if ((Dummy = (BAN_list *) search_ban(b_list, banstring)) == NULL)
    return (FALSE);

  for (old = b_list; *old; old = &(*old)->next)
  {
    Debug(DBGINFO, "*** In delete_ban loop: %s\n", &(*old)->banstring);
    if (*old == Dummy)
    {
      Debug(DBGINFO, "*** Found the match.");
      *old = Dummy->next;
      free(Dummy);
      return (TRUE);
    }
  }
  Debug(DBGINFO, "*** NULL encountered. ban not found. returning.\n");
  return (FALSE);
}

int DeleteExpiredBans(char *Channel, BAN_list ** b_list, time_t period)
{
  BAN_list *Dummy, *tmp;
  char unbanmode[10];
  char unbanstring[1024];
  int i;
  int found = 0;
  time_t currtime;

  Debug(DBGNOTICE, "*** Checking for bans in %s older than %lu\n", Channel, period);
  Dummy = *b_list;
  currtime = time(NULL);
  while (Dummy)
  {
    /* We can't just -bbb ban1 ban2,
       because the third b will make the server
       show the banlist etc. */
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (Dummy && (i < 6))
    {
      if ((currtime - Dummy->time) > period)
      {
	strcat(unbanmode, "b");
	strcat(unbanstring, " ");
	strcat(unbanstring, Dummy->banstring);
	i++;
	found++;
	tmp = Dummy->next;
	delete_ban(b_list, Dummy->banstring);
	Dummy = tmp;
      }
      else
	Dummy = Dummy->next;
    }
    if (unbanmode[1])
      sendtosock(Sock, ":%s MODE %s %s %s", IrcNick, Channel, unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

void SaveDel(char *Channel)
{
  UserListType UserNode;
  ChannelListType ChanPtr;
  FILE *ChanFile;
  FILE *NewChanFile;
  char *UpCaseChan;
  char str[LINEBUFFSIZE];
  char ChanName[LINEBUFFSIZE], rest[LINEBUFFSIZE];
  time_t LastAccess, CreationDate;
  int s, line = 0;

  ChanFile = fopen(CHANFILENAME, "r");
  if (ChanFile < 0)
  {
    Debug(DBGSEVERE, "Error: op file doesn't not exist!");
    return;
  }
  NewChanFile = fopen("tmp/.tmpchanfile.dat", "w");
  if(!NewChanFile)
  {  
     Debug(DBGSEVERE, "Error opening temp file in unreg()!!");
     return;
  }
  while ((s = fscanf(ChanFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    { 
      Debug(DBGERROR, "Error with param count in chan.dat! s=%d line=%d", s, line);
      fclose(ChanFile);
      fclose(NewChanFile);
      return;
    }
    if (strcasecmp(ChanName, Channel))
      fprintf(NewChanFile, "%s\n", ChanName);
  }
  if(ferror(NewChanFile))
  {  
     Debug(DBGSEVERE, "Error writing to .tmpchanfile.dat (Out of disk space?). Unregistration NOT saved.", Channel);
     fclose(NewChanFile);
     sleep(4);
  }
  fclose(ChanFile);
  fclose(NewChanFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tmpchanfile.dat", str);                                                        
  sprintf(str, "./CHANS/%s", UpCaseChan = UpCase(&Channel[1]));                                   
  unlink(str);                                                                                
  free(UpCaseChan);                
}

void ExpireChannels(char *Nick)
{
  ChannelListType TempChanPtr, NextChanPtr;
  int i;
  char ChanName[MAXLEN];

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    if (Channels->Table[i])
    {
      TempChanPtr = Channels->Table[i];
      while (TempChanPtr)
      {
        NextChanPtr = TempChanPtr->Next;
        if( !(TempChanPtr->settings[NODELETE]) && (TempChanPtr->LastAccess > 0)) /* If its not marked nodelete */
        {
          if(  (time(NULL) - TempChanPtr->LastAccess) > 86400 * CHANEXPIREDAYS) /* older than CHANEXPIREDAYS days */
          {
             char str[MAXLEN];

             Debug(DBGNOTICE, "Channel %s not visited since %lu. Removing..", TempChanPtr->Name,
                                                                              TempChanPtr->LastAccess);
             if(Nick)
                send_to_user(Nick, "Removing %s", TempChanPtr->Name);
             sendtosock(Sock, ":%s PART %s", IrcCID, TempChanPtr->Name);
             DelChan(TempChanPtr->Name);
             SaveDel(TempChanPtr->Name);
          }
        }
        /* work here */
	TempChanPtr = NextChanPtr;
      }
    }
  }
}

void CheckBanTimeouts()
{
  char ChanModes[50];
  char ModeString[LINEBUFFSIZE];
  int i;
  ChannelListType TempChanPtr;
  time_t period;

  Debug(DBGINFO, "*** Scanning for expired bans...\n");
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    if (Channels->Table[i])
    {
      TempChanPtr = Channels->Table[i];
      while (TempChanPtr)
      {
	switch (TempChanPtr->settings[BANTIMEOUT])
	{
	case 0:
	  period = -1;
	  break;
	case 1:
	  period = 600;		/* 10 mins */
	  break;
	case 2:
	  period = 7200;	/* 2 hours */
	  break;
	case 3:
	  period = 86400;	/* 24 hours */
	  break;
	case 4:
	  period = 604800;	/* one week */
	  break;
	default:
	  period = -1;
	  break;
	}
	if (period > 0)
	  DeleteExpiredBans(TempChanPtr->Name, &(TempChanPtr->banned), period);
	TempChanPtr = TempChanPtr->Next;
      }
    }
  }
}

int channel_massunban(char *channel)
{
  ChannelListType Channel;
  BAN_list *Dummy;
  char unbanmode[10];
  char unbanstring[1024];
  int i;
  int found = 0;

  if ((Channel = GetChannel(channel)) == NULL)
  {
    return;
  }
  Dummy = Channel->banned;
  while (Dummy)
  {
    /* We can't just -bbb ban1 ban2,
       because the third b will make the server
       show the banlist etc. */
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (Dummy && (i < 6))
    {
      strcat(unbanmode, "b");
      strcat(unbanstring, " ");
      strcat(unbanstring, Dummy->banstring);
      i++;
      Dummy = Dummy->next;
      found++;
    }
    Channel->banned = NULL;
    sendtosock(Sock, ":%s MODE %s %s %s", IrcNick, channel, unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

int channel_unban(char *channel, char *user)
{
  ChannelListType Channel;
  BAN_list *Dummy, *tmp;
  char unbanmode[5];
  char unbanstring[MAXLEN];
  int i;
  int found = 0;

  if ((Channel = GetChannel(channel)) == NULL)
    return;

  Dummy = Channel->banned;
  while (Dummy)
  {
    strcpy(unbanmode, "-");
    strcpy(unbanstring, "");
    i = 0;
    while (Dummy && (i < 3))
    {
      tmp = Dummy->next;
      if (!matches(Dummy->banstring, user) || !strcmp(user, Dummy->banstring))
      {
	strcat(unbanmode, "b");
	strcat(unbanstring, " ");
	strcat(unbanstring, Dummy->banstring);
	Debug(DBGINFO, "*** channel_unban calling delete_ban for %s", Dummy->banstring);
	delete_ban(&Channel->banned, Dummy->banstring);
	i++;
	found++;
      }
      Dummy = tmp;
    }
    sendtosock(Sock, ":%s MODE %s %s %s", IrcNick, channel, unbanmode, unbanstring);
  }
  if (found > 0)
    return (1);
  else
    return (0);
}

BAN_list *search_ban(BAN_list ** b_list, char *banstring)
{
  BAN_list *Banned;

  for (Banned = *b_list; Banned; Banned = Banned->next)
    if (!strcasecmp(Banned->banstring, banstring))
      return (Banned);
  return (NULL);
}
