/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "chan.h"
#include "init.h"
#include "user.h"
#include "misc.h"
#include "sock.h"
#include "server.h"
#include "commands.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifndef MSDOS
#include <strings.h>
#endif
#include <time.h>
#include <signal.h>

int FILTER = 0;
char LISTTYPE = '\0';

unsigned long int TotalCommands = 0, TotalLines = 0, TotalUserCount = 0;
char *chan_to;
/* *INDENT-OFF* */
struct
{
        char  *name;
        int  (*function)();
        int   userlevel;
        int   shitlevel;
        int   publicOK;
        int   needChan;
	int   needInChan;
        int   ircop;
} on_msg_commands[] =
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
{
    { "JUMP",      do_jump,            500,   0,     TRUE, FALSE, FALSE, FALSE },
    { "UPLINKS",   do_uplinks,         500,   0,     TRUE, FALSE, FALSE, FALSE },
    { "SEX",       do_sex,               0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "LAST10",    do_last10,          500,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "TEST",      do_test,              0,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "CLUSTER",   do_cluster,           0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "NEWS",      do_news,              0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "REJOIN",    do_rejoin,          500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "REHASH",    do_rehash,          500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "WIPESETS",  do_wipesets_,       500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "SEARCHCHAN",do_searchchan,        0,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "TOPIC",     do_topic,           100,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "MODE",      do_mode,            100,   0,     TRUE, TRUE,  TRUE,  TRUE  },
    { "UPTIME",    do_uptime,            0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "HELP",      ProcessHelp,          0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "IRCOPS",    ShowIrcOps,           0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "SUSPEND",   ProcessSuspend,     500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "NETSTAT",   NetStat,              0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "NETSTATS",  NetStat,              0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "SERVERS",   ShowServers,          0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "MAP",       do_map,               0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "LINKS",     ShowServers,          0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "SETTINGS",  Settings,             0,   0,     TRUE, FALSE, FALSE, FALSE },
/*  { "REROUTE",   Reroute,            500,   0,     TRUE, FALSE, FALSE, TRUE  },
*/
    { "ROUTING",   do_routing,         500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "LOADROUT",  do_loadrout,        500,   0,     TRUE, FALSE, FALSE, TRUE  },
    { "OPEN",      ProcessOpen,        100,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "INVITE",    ProcessInvite,      200,   0,     TRUE, TRUE,  FALSE, FALSE },
    { "INVITEME",  do_inviteme,        100,   0,    FALSE, TRUE,  FALSE, FALSE },
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
    { "USERLIST",  ShowOps,             10,   0,     TRUE, TRUE,  FALSE, TRUE  },  
    { "USERS",     ShowOps,             10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "LAMERLIST", do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "LAMELIST",  do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "LAMERS",    do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "SHITLIST",  do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },  
    { "PLIST",     do_listpeons,        10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "OLIST",     do_listops,          10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "MLIST",     do_listmanagers,      0,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "CLIST",     do_listcoowners,      0,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "WLIST",     do_listowners,        0,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "BLIST",     do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "LLIST",     do_listlamers,       10,   0,     TRUE, TRUE,  FALSE, TRUE  },
    { "CHANSTAT",  ChanStat,             0,   0,     TRUE, TRUE,  FALSE, FALSE },
    { "CHANSTATS", ChanStat,             0,   0,     TRUE, TRUE,  FALSE, FALSE },
    { "AUTH",      ProcessAuth,          0,   0,    FALSE, TRUE,  FALSE, FALSE },
    { "TOGOP",     do_togop,            10,   0,     TRUE, TRUE,  FALSE, FALSE },
    { "PASS",      ProcessPass,          0,   0,    FALSE, TRUE,  FALSE, FALSE },
    { "UNREG",     UnregChan,          500,   0,    FALSE, TRUE,  FALSE, TRUE },
    { "EXPIRECHANS",ExpireChans,       500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SPOOFSERV", ProcessSpoofServ,   500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "RESETPASS", ProcessResetPass,   200,   0,    FALSE, TRUE,  FALSE, TRUE },
    { "OP",        OpComm,             100,   0,     TRUE, TRUE,  TRUE,  FALSE },
    { "VOICE",     do_voice,           100,   0,     TRUE, TRUE,  TRUE,  FALSE},
    { "V",         do_voice,           100,   0,     TRUE, TRUE,  TRUE,  FALSE},
    { "DEVOICE",   do_devoice,         100,   0,     TRUE, TRUE,  TRUE, FALSE},
    { "OPCHAN",    do_opchan,          500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "DEOPCHAN",  do_deopchan,        500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "DOWN",      do_down,              0,   0,     TRUE, TRUE,  TRUE, FALSE },
    { "UP",        do_up,               10,   0,     TRUE, TRUE,  TRUE, FALSE },
    { "DOME",      do_up,               10,   0,     TRUE, TRUE,  TRUE, FALSE },
    { "KICK",      do_kick,            100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "K",         do_kick,            100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "DEOP",      DeopComm,           100,   0,     TRUE, TRUE,  TRUE, FALSE},
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
    { "SET",       do_set,             290,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "XSET",      ProcessXSET,        500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "GLOBAL",    ProcessGlobal,      500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "UNVISITED", Unvisited,          500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SETTIME",   do_settime,         500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "GLINE",     ProcessGline,       500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SGLINE",    ProcessSGline,      500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "UNGLINE",   ProcessUngline,     500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "UNSGLINE",  ProcessUnsgline,    500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "UNGLINEALL",ProcessUnglineAll,  500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "GLINES",    ProcessGlines,      500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SGLINES",   ProcessSGlines,     500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SETGLINES", do_setglines,       500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "LOADCHAN",  do_loadchan,        500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "REGCHAN",   RegisterChan,       500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SCRIPTREG", ScriptReg,          500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "SAVECHAN",  do_savechan,        500,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "LEVELS",    do_showlevels,        0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "ADDLAMER",  ProcessAddLamer,    200,   0,     TRUE, TRUE,  FALSE, FALSE},
    { "ADDPEON",   ProcessAddPeon,     200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "ADDOP",     ProcessAddOp,       200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "ADDMASTER", ProcessMasterInfo,    0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "ADDMANAGER",ProcessAddManager,  290,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "ADDCOOWNER",ProcessAddCoOwner,  300,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "ADDCO",     ProcessAddCoOwner,  300,   0,     TRUE, TRUE,  FALSE, TRUE },  
    { "ADDOWNER",  ProcessAddOwner,    500,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "IDENTIFY",  do_identify,          0,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "CLVL",      do_clvl,            200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "ACCESS",    ProcessAccess,        0,   0,     TRUE, TRUE,  FALSE, FALSE},
    { "SEEN",      do_seen,              0,   0,     TRUE, TRUE,  FALSE, FALSE},
    { "DELLAMER",  ProcessDelLamer,    200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELLAMER", ProcessFDelLamer,   200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "DELPEON",   ProcessDelPeon,     200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELPEON",  ProcessFDelPeon,    200,   0,     TRUE, TRUE,  FALSE, TRUE }, 
    { "DELOP",     ProcessDelOp,       200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELOP",    ProcessFDelOp,      200,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELMASTER",ProcessMasterInfo,    0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "DELMASTER", ProcessMasterInfo,    0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "DELMANAGER",ProcessDelManager,  290,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELMANAGER",ProcessFDelManager,290,   0,     TRUE, TRUE,  FALSE, TRUE }, 
    { "DELCOOWNER",ProcessDelCoOwner,  300,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELCOOWNER",ProcessFDelCoOwner,300,   0,     TRUE, TRUE,  FALSE, TRUE }, 
    { "DELCO",     ProcessDelCoOwner,  300,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELCO",    ProcessFDelCoOwner, 300,   0,     TRUE, TRUE,  FALSE, TRUE },     
    { "DELOWNER",  ProcessDelOwner,    500,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "FDELOWNER", ProcessFDelOwner,   500,   0,     TRUE, TRUE,  FALSE, TRUE }, 
    { "SETHANDLE", do_sethandle,       100,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "SETINFO",   do_setinfo,          10,   0,     TRUE, TRUE,  FALSE, FALSE},
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
    { "QUIT",      ProcessQuit,        500,   0,    FALSE, FALSE, FALSE, TRUE },
    { "D",         do_roll,              0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "8BALL",     Process8ball,         0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "8",         Process8ball,         0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "CALC",      do_calc,              0,   0,     TRUE, FALSE, FALSE, FALSE },
    { "WHOIS",     do_whois,           500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "WHOAMI",    do_whoami,            0,   0,     TRUE, FALSE, FALSE, TRUE },
    { "UBA",       do_massunban,       100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "UB",        do_unban,           100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "UNBAN",     do_unban,           100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "UBM",       do_unbanme,         100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "B",         do_ban,             100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "BAN",       do_ban,             100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "KB",        do_kickban,         100,   0,     TRUE, TRUE,  TRUE, TRUE },
    { "KICKBAN",   do_kickban,         100,   0,     TRUE, TRUE,  TRUE, TRUE }, 
    { "MATCH",     do_match,           500,   0,     TRUE, TRUE,  FALSE, TRUE },
    { "VERSION",   do_version,           0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "AUTHORS",   do_authors,           0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "MOTD",      do_motd,              0,   0,     TRUE, FALSE, FALSE, FALSE},
    { "DUMP",      do_raw,             500,   0,     TRUE, FALSE, FALSE, TRUE },
#ifdef debugon
    { "RAW",       do_raw,             500,   0,     TRUE, FALSE, FALSE, TRUE },
#endif
#ifndef AFTERNET /* we have a mode in our ircd for this */
    { "GOD",       do_god,             500,   0,     TRUE, FALSE, FALSE, TRUE },
#endif
    { "MEMDUMP",   do_memdump,         500,   0,     TRUE, FALSE, FALSE, TRUE },
    { "BACKUP",    do_backup,          500,   0,     TRUE, FALSE, FALSE, TRUE },
/* ... */
    { NULL,        null(int(*)()),      0,   0,     FALSE, FALSE, FALSE, FALSE}
/*    Command      function        userlvl   na    public   Chan  InChan  ircopOverride */
};

/* *INDENT-ON* */

void on_msg(char *from, char *to, char *done_in, boolean public, char *command, char *msg)
{
  int i, ii, to_all, temp;
  FILE *logfile;
  static time_t currenttime;
  extern char *parmstr;
  char bigbuf[MAXLEN] = "\0";
  char chan[MAXCHANLEN] = "\0";
  char blank[3] = "\0";
  char buf1[MAXLEN/2], buf2[MAXLEN/2], tmp[MAXLEN];
  int foundcommand = 0;
  int commandlinechan = 0;
  ChannelListType ChanNode;
  UserListType tmpuserptr;

  chan_to = done_in;
  strncpy(chan, to, 200);
  TotalCommands++;
  if (!msg)
    msg = blank;
  parmstr = msg;
  if (!(tmpuserptr = FindNick(from, UsersN)))
  {
    Debug(DBGERROR, "ERROR: cannot find %s in memory!! SERIOUS ERROR\n", from);
    return;
  }

  Debug(DBGNOTICE, "*** on_msg(from:%s, to:%s, chan%s, public:%d, command:%s, msg:%s)", from, to, done_in, public, command, msg);

  for (; *msg && *msg == ' '; msg++);
  if (!(*command) || !(command))
    return;			/* NULL-command */
/****************************************************/
/** Find the command in the list and check access ***/
/****************************************************/
  for (i = 0; on_msg_commands[i].name != NULL; i++)
  {
    if (!strcasecmp(on_msg_commands[i].name, command))
    {
      foundcommand++;
/****************************************************/
/*** If its sent to a chan check if its publickOK ***/
/****************************************************/
      if (public && !(on_msg_commands[i].publicOK))
      {
	send_to_user(from, "\002This command must be done in /msg format.\002");
	return;
      }
/****************************************************/
/*** Check if chan is required & if we have one   ***/
/****************************************************/
      if (!((to) && (*to)) && on_msg_commands[i].needChan)
      {
	send_to_user(from, "\002What channel?\002");
	return;
      }
      ChanNode = GetChannel(chan);
      if (on_msg_commands[i].needChan && !(ChanNode))
      {
	send_to_user(from, "\002Channel %s not registered\002", chan);
	return;
      }
/****************************************************/
/*** Check if its INCHAN, and if it needs to be   ***/
/****************************************************/
      if(ChanNode)
      {
         if(!ChanNode->settings[INCHAN] && on_msg_commands[i].needInChan)
         {
             if(ChanNode->settings[SUSPENDED])
               send_to_user(from, "\002X2 is currently suspended in %s by the CService administrators. email cservice@afternet.org for more information\002", ChanNode->Name);
             else
               send_to_user(from, "\002X2 is currently disabled in %s. To Re-Enable do: \002 /MSG %s SET INCHAN 1",
                                  ChanNode->Name, IrcNick); 
             return;
         }
      }
/****************************************************/
/*** Check if we have authenticated (or need to)  ***/
/****************************************************/
      if (on_msg_commands[i].needChan &&
	  !(IsValidated(from, chan)) &&
	  (on_msg_commands[i].userlevel > 0) &&
	  (userlevel(from, chan) > 0) &&
	  (!IsGodOper(from)) &&
	  (on_msg_commands[i].userlevel != 500))
      {
	send_to_user(from, "\002Access denied: Authentication required.\002");
	send_to_user(from, "\002/msg %s auth %s <password>\002", IrcNick, chan);
	return;
      }
      currenttime = time(NULL);
      strncpy(tmp, ctime(&currenttime), MAXLEN);
      tmp[strlen(tmp) - 1] = '\0';
          /*  :time:<nick!user@host> */ /* command [#chan] rest  */
      snprintf(buf1, MAXLEN/2, ":%s:<%s!%s@%s>", tmp,from,tmpuserptr->UserID,tmpuserptr->Address);
      snprintf(buf2, MAXLEN/2, "[%s] %s %s\n", command, chan, msg);
/****************************************************/
/** Check if we have access and are not shitlisted **/
/****************************************************/
      /* if we have access in the channel, or
       * we are +g and the command allows irc override, or
       * were an oper and the commadn is a level 500 command */
      if ((on_msg_commands[i].userlevel <1) || chan && *chan && (userlevel(from, chan) >= on_msg_commands[i].userlevel))
      {				/* we have normal access to the channel */
        if( on_msg_commands[i].function(from, chan, msg))
        {
            logfile = fopen(AccountingLog, "a");
            if (logfile != 0)
            {
               fprintf(logfile, "%s %s", buf1, buf2);
               fclose(logfile);
            }
        }
	return;
      }
      else if (IsOper(from) && (on_msg_commands[i].userlevel == 500))
      {				/* we have normal access to an ircop command */

        if(on_msg_commands[i].function(from, chan, msg))
        {
            logfile = fopen(AccountingLog, "a");
            if (logfile != 0)
            {
               fprintf(logfile, "%s used ircop to %s", buf1, buf2);
               fclose(logfile);
            }
        }
        return;
      }
      else if (IsGodOper(from) && on_msg_commands[i].ircop)
      {  /* we are using +g mode to over-ride channel access */
        send_to_user(from, "\002WARNING:\002 You are using GOD MODE to over-ride channel security!");
        if(on_msg_commands[i].function(from, chan, msg))
        {
            logfile = fopen(AccountingLog, "a");
            if (logfile != 0)
            {
               fprintf(logfile, "%s used godmode to %s", buf1, buf2);
               fclose(logfile);
            }
        }
        return;
      }
      else
      {				/* we have no access at all */
        send_to_user(from, "\002Sorry, access to \002%s\002 is denied.", command);
	return;
      }
/****************************************************/
    }
  }
/****************************************************/
/** Check if we found a hit. If not tell them      **/
/****************************************************/
  if (foundcommand == 0)
  {
    char *TmpPtr;
    int valid = 1;

    for (TmpPtr = command; *TmpPtr && *TmpPtr != ' '; TmpPtr++)
    {				/* This checks for '<WeirdKev> ...' */
      if(!isalnum(*TmpPtr))
      {
       valid = 0;
       break;
      }
    }
    if (valid)
      send_to_user(from, "\002Unknown Command [%s]. Try HELP for help.\002", command);
    return;
  }
/****************************************************/
}

set_list set_commands[NUMOFCHANSETTINGS+1] =
/*   Command        min max def description set0 set1 set2 set3 set4 set5 */
{
  {
    /* #define TOPICFREQ   0 */
    "TOPICFREQ", TOPICFREQ, 1,0, 4, 1, "How frequently x2 sets the channel topic to its default",
      "The Topic is refreshed: ",
      "Never.",
      "Once a day.",
      "Twice a day.",
      "Every 4 hours.",
      "Every hour.",
      ""
  }
  ,
  {
/* #define STRICTOP    1 */
    "STRICTOP", STRICTOP, 1,0, 2, 0, "If X2 lets non-users get oped by others",
      "Unknown users can be opped by: ",
      "Anyone.",
      "Only OPs, MANAGERs, and (CO)OWNERs.",
      "Only MANAGERs and (CO)OWNERs.",
      "", "", ""
  }
  ,
  {
    "AUTOOP", AUTOOP, 1,0, 3, 2, "Who does X2 autoop",
      "X2 will Auto-op: ",
      "Nobody",
      "Only MANAGERs and (CO)OWNERs.",
      "OPs, MANAGERs, and (CO)OWNERs.",
      "Everyone.",
      "", ""
  }
  ,
  {
    "PROTECT", PROTECT, 1,0, 4, 0, "Who does X2 protect from who?",
      "X2 will protect: ",
      "No one from anyone.",
      "users from anyone of lower rank.",
      "users from anyone of equal or lower rank.",
      "Only MANAGERs, and (CO)OWNERs from everyone.",
      "Only (CO)OWNERS from everyone.",
      ""
  }
  ,
  {
    "ENFMODES", ENFMODES, 1,0, 2, 0, "Should X2 force the default modes",
      "Default modes can be overridden by: ",
      "Everyone.",
      "Only OPs, MANAGERs, and (CO)OWNERs.",
      "Only MANAGERs and (CO)OWNERs.",
      "", "", ""
  }
  ,
  {
    "PASSMODE", PASSMODE, 1,0, 1, 0, "Should X2 give ops to passworded users who are not verified?",
      "Users who have set a password, but have NOT authenticated: ",
      "are NOT opped.",
      "are given ops anyway.",
      "", "", "", ""
  }
  ,
  {
    "BANTIMEOUT", BANTIMEOUT, 1, 0, 5, 0, "Should X2 remove old bans?",
      "Bans are removed: ",
      "Never.",
      "After 10 minutes.",
      "After 2 hours.",
      "After 4 hours.",
      "After 24 hours.",
      "After one week."
  }
  ,
  {
    "VOICE", VOICE, 1, 0, 2, 1, "Should X2 give voice?",
      "Mode +v (voice) given to: ",
      "Noone.",
      "Peons.",
      "Everyone.",
      "",
      "",
      ""
  }      
  ,
  {
    "USERINFO", USERINFO, 1, 0, 1, 1, "Show userinfo's",
      "Show userinfo lines when users join?: ",
      "No",
      "Yes",
      "",
      "",
      "",
      ""
  }
  ,
  {
    "INCHAN", INCHAN, 1, 0, 1, 1, "Is it joined",
     "Should X2 go into the channel?: ",
     "No",
     "Yes",
     "",
     "",
     "",
     ""
  }
  ,
  {
    "SUSPENDED", SUSPENDED, 0, 0, 1, 0, "Is it suspended",
     "Has the channel been suspended by CService Admin?: ",
     "No",
     "Yes",
     "",
     "",
     "",
     ""
  }
  ,
  {
    "NODELETE", NODELETE, 0, 0, 1, 0, "Is it void of deletion",
     "Exempt from unvisited deletion rules?: ",
     "No",
     "Yes",
     "",
     "",
     "",
     ""
  }
  ,
  {
    NULL, 0, 0, 0, 0, 0, NULL,
      "",
      "", "", "", "", "", ""
  }
};

/* Begin command descriptions */

int do_ban(char *from, char *channel, char *rest)
{
  char nuh[MAXLEN];
  UserListType UserNode;
  ChannelListType ChanNode;
  char *nick, *user, *host;
  char tmpbuf[MAXLEN];

  rest = (char *) strtok(rest, " ");
  if (!rest)
  {
    send_to_user(from, "\002Who do you want to ban?\002");
    return(FALSE);
  }
  if (!(ChanNode = GetChannel(channel)))
  {
    send_to_user(from, "\002Warning: Error getting channel!! Internal error.\002");
    return(FALSE);
  }
  if (UserNode = FindNick(rest, UsersN))
  {
    if (IsModeK(rest))
    {
      send_to_user(from, "\002%s is a Service.\002", rest);
      return(FALSE);
    }
    sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);

    if (!banok(&ChanNode, from, nuh))
    {
      send_to_user(from, "\002Sorry, that user is protected\002");
      return(FALSE);
    }
    ban_user(channel, nuh);
    send_to_user(from, "\x2%s banned on %s\x2", nuh, channel);
    sendtosock(Sock, ":%s WALLCHOPS %s :\002%s banned %s on %s\002", IrcCID, channel, from, nuh, channel);
  }
  else
  {
    if (!strchr(rest, '@') || !strchr(rest, '!'))
    {
      send_to_user(from, "\002Bans must be either the nick of someone currently online, or a mask in the form of *!*@*\002");
      return(FALSE);
    }
    strncpy(tmpbuf, rest, MAXLEN);
    tmpbuf[MAXLEN - 1] = '\0';
    nick = strtok(tmpbuf, "!");
    user = strtok(NULL, "@");
    host = strtok(NULL, " ");
    if (!nick || !user || !host || !*nick || !*user || !*host)
    {
      send_to_user(from, "\002Each ban must contain _something_ for the nick, user and host. even if that thing is a *.\002");
      return(FALSE);
    }
    if (!banok(&ChanNode, from, rest))
    {
      send_to_user(from, "\002Sorry, that user is protected\002");
      return(FALSE);
    }
    else
    {
      ban_mask(channel, rest);
      send_to_user(from, "\x2%s banned on %s\x2", rest, channel);
      sendtosock(Sock, ":%s WALLCHOPS %s :\002%s banned %s on %s\002", IrcCID, channel, from, rest, channel);
      return(TRUE);
    }
  }
}

int do_jump(char *from, char *Channel, char *rest)
{
  UpLinkList *UpLinkPtr;
  char *server, *portstr;
  int port = 0;
  int sockholder;

  server = (char *) strtok(rest, " ");
  portstr = (char *) strtok(NULL, " ");
  if(!portstr)
     portstr = "6667";

  if(!server)
  {
     send_to_user(from, "Syntax: JUMP <server> [port]");
     return(FALSE);
  }
  for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
  {
    if(!strcasecmp(server, UpLinkPtr->ServName))
    {
      if(portstr)
        port = atol(portstr);
      if(port < 0)
        port = UpLinks->Port;
      
      send_to_user(from, "Attempting to swith my uplink to %s..", UpLinkPtr->ServName);
      sendtosock(Sock, ":%s WALLOPS :Changing to server %s", IrcCID, UpLinkPtr->ServName);
      sendtosock(Sock, ":%s QUIT :Changing servers", IrcCID);
      /* Disconnecting... */
      sendtosock(Sock, ":%s SQUIT %s :Changing to server %s", ServiceName, Uplink, UpLinkPtr->ServName);
      DiscTime = time(NULL);
      close(Sock);
      Debug(DBGINFO, "Changing uplink servers..");
      ServerConnected = FALSE;
      NETBURSTING = TRUE;
      SentBurst = FALSE;
      Authenticated = FALSE;
      KillServerList();
      KillUserList();
      OperCount = 0;
      CurUpLink = UpLinkPtr;
      if(TryToConnect(UpLinkPtr))
      {
          Authenticate(CurUpLink);
          Authenticated = TRUE;
          SentBurst = TRUE;
      }
      else
      {
         Debug(DBGINFO, "Uplink server change failed.. trying to reconnect somewhere..");
         SockConnected = FALSE;
         return(TRUE);
      }
      return(TRUE);
    }
  }
  send_to_user(from, "Server %s not found in Uplinks list.", server);
  return(FALSE);
}

int do_uplinks(char *from, char *Channel, char *rest)
{
   UpLinkList *UpLinkPtr;

   send_to_user(from, "=------- Servers I can link to: --------=");
   for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
   {
     send_to_user(from, "%40s[%16s] %4d", UpLinkPtr->ServName, UpLinkPtr->ServIp, UpLinkPtr->Port);
   }
   send_to_user(from, "=---------------- End ------------------=");
   return(TRUE);
}

int do_sex(char *Nick, char *Channel, char *rest)
{
  char *to;

  if (chan_to)
  {
    if (userlevel(Nick, chan_to) > 0)
      to = chan_to;
    else
      to = Nick;
  }
  else
  {
    to = Nick;
  }
  send_to_user(to, "\002%s:\002 I don't want to be a part of your sick fantasies.\002", Nick);
  return(TRUE);
}

int do_last10(char *Nick, char *Channel, char *rest)
{
  char buf[MAXLEN];
  char Buffer[MAXLEN];
  int c;
  FILE *LastFile;

  sprintf(buf, "last10 %s", Channel);
  system(buf);
  if (LastFile = fopen("last10.tmp", "r"))
  {
    send_to_user(Nick, "=----------- Last 10 commands --------------=");
    while (c = fgets(Buffer, LINEBUFFSIZE, LastFile) != NULL)
    {
      Buffer[strlen(Buffer) - 1] = '\0';
      send_to_user(Nick, " %s", Buffer);
    }
    send_to_user(Nick, "=------------------ End --------------------=");
    fclose(LastFile);
  }
  unlink("last10.tmp");
  return(TRUE);
}

int do_togop(char *Nick, char *Channel, char *rest)
{
  ChannelListType ChanPtr;
  char *phrase;

  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered", Channel);
    return(FALSE);
  }
  rest = strtok(rest, " ");
  if(IsChanPeon(Nick, Channel))
     phrase = "Voice";
  else
     phrase = "Op";
  if (!rest)
  {
    if (IsChanAOp(Channel, Nick))
    {
      send_to_user(Nick, "X2 will now no longer Auto %s you when you join %s", phrase, Channel);
      SetChanAOp(Channel, Nick, FALSE);
      return(TRUE);
    }
    else
    {
      send_to_user(Nick, "X2 will now Automaticaly %s you when you join %s", phrase, Channel);
      SetChanAOp(Channel, Nick, TRUE);
      return(TRUE);
    }
  }
  else if (!strcasecmp(rest, "ON") || !strcasecmp(rest, "TRUE"))
  {
    if (IsChanAOp(Channel, Nick))
      send_to_user(Nick, "Your Auto %s status on %s is already on.", phrase, Channel);
    else
      send_to_user(Nick, "X2 will now Automaticaly %s you when you join %s", phrase, Channel);
    SetChanAOp(Channel, Nick, TRUE);
    return(TRUE);
  }
  else if (!strcasecmp(rest, "OFF") || !strcasecmp(rest, "FALSE"))
  {
    if (!IsChanAOp(Channel, Nick))
      send_to_user(Nick, "Your Auto %s status on %s is already off.", phrase, Channel);
    else
      send_to_user(Nick, "X2 will now no longer Auto %s you when you join %s", phrase, Channel);
    SetChanAOp(Channel, Nick, FALSE);
    return(TRUE);
  }
  send_to_user(Nick, "Error: Use keywords ON, or OFF");
  return(FALSE);
}

int do_kickban(char *from, char *channel, char *rest)
{
  char nuh[MAXLEN];
  char nick[MAXLEN], reason[MAXLEN], *buff;
  UserListType UserNode;
  ChannelListType ChanNode;

/*  buff = (char *) strtok(rest, " "); */
  buff = (char *) mysep(&rest, " ");
  if (buff)
    safecopy(nick, buff, MAXLEN);
  else
  {
    send_to_user(from, "\002Who do you want to ban?\002");
    return(FALSE);
  }
/*  buff = (char *) strtok(NULL, "\0"); */
  buff = (char *) mysep(&rest, "\0");
  if (buff)
    sprintf(reason, "\002<%s> %s\002", from, buff);
  else
    sprintf(reason, "\002<%s> Later %s\002", from, nick);
  if (!(ChanNode = GetChannel(channel)))
  {
    send_to_user(from, "\002Warning: Error getting channel!! Internal error.\002");
    return(FALSE);
  }
  if (UserNode = FindNick(nick, UsersN))
  {
    if (IsModeK(nick))
    {
      send_to_user(from, "\002Cannot Kick or deop channel service\002");
      return(FALSE);
    }
    sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);

    if ((!kickok(&ChanNode, from, nick)) || (!banok(&ChanNode, from, nuh)))
    {
      send_to_user(from, "\x2Sorry, that user is protected\x2");
      return(FALSE);
    }

    ban_user(channel, nuh);
    sendtosock(Sock, ":%s KICK %s %s :%s", IrcCID, channel, nick, reason);
    send_to_user(from, "\x2%s kick-banned on %s\x2", nuh, channel);
  }
  else
    send_to_user(from, "\002Unknown user %s\002", nick);
  return(TRUE);
}

int do_unbanme(char *from, char *channel, char *rest)
{
  char nuh[MAXLEN];
  UserListType UserNode;

  if (UserNode = FindNick(from, UsersN))
  {
    sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
    if (channel_unban(channel, nuh))
    {
      send_to_user(from, "\x2You are unbanned on %s\x2", channel);
      return(TRUE);
    }
    else
    {
      send_to_user(from, "\002No bans matching %s found.\002", nuh);
      return(FALSE);
    }
  }
}

int do_massunban(char *Nick, char *Channel, char *rest)
{
  if (channel_massunban(Channel))
  {
    send_to_user(Nick, "\002Unbanning everyone on %s\002", Channel);
    return(TRUE);
  }
  else
  {
    send_to_user(Nick, "\002No bans in %s\002", Channel);
    return(FALSE);
  }
}

int do_unban(char *from, char *channel, char *rest)
{
  char nuh[MAXLEN];
  UserListType UserNode;

  if (!(rest = (char *) strtok(rest, " ")))
  {
    send_to_user(from, "\002Usage: UB <nick/user@host>\002");
    return(FALSE);
  }
  if (UserNode = FindNick(rest, UsersN))
  {
    sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
    if (channel_unban(channel, nuh))
    {
      send_to_user(from, "\002%s unbanned on %s\002", nuh, channel);
      sendtosock(Sock, ":%s WALLCHOPS %s :\002%s Unbanned %s in %s\002", IrcCID, channel, from, nuh, channel);
    }
    else
    {
      send_to_user(from, "\002No matching bans.\002");
      return(FALSE);
    }
  }
  else
  {
    if (channel_unban(channel, rest))
    {
      send_to_user(from, "\002%s unbanned on %s\002", rest, channel);
      sendtosock(Sock, ":%s PRIVMSG %s :\002%s Unbanned %s in %s\002", IrcCID, channel, from, rest, channel);
    }
    else
    {
      send_to_user(from, "\002No matching bans.\002");
      return(FALSE);
    }
  }
  return(TRUE);
}

int do_cycle(char *from, char *channel, char *rest)
{
/*    sendtosock(Sock, ":%s PART %s", IrcNick, channel);
   sendtosock(Sock, ":%s JOIN %s", IrcNick, channel);
 */
   return(FALSE);
}

int do_news(char *nick, char *channel, char *rest)
{
    SendNetMessage(nick);
    return(true);
}

int do_cluster(char *nick, char *channel, char *rest)
{
   if(!rest || !*rest)
   {
      send_to_user(nick, "USAGE: cluster <nick/nick!user@host>");
      send_to_user(nick, "This command is just a quick access for testing the proper responce of the x2 hostmask guessing functions.");
      return;
   }
   if(strchr(rest, '!') || strchr(rest, '@'))
     send_to_user(nick, "That string clusteres to: %s", cluster(rest));
   else
   {
     UserListType UserPtr;

     if (!(UserPtr = FindNick(rest, UsersN)))
        send_to_user(nick, "Cannot find %s on IRC.", rest);
     else
     {
        char str[MAXLEN];
        snprintf(str, MAXLEN, "%s!%s@%s", UserPtr->Nick, UserPtr->UserID, UserPtr->Address);
        send_to_user(nick, "That user (%s) clusters to: %s", str, cluster(str));
     }
   }
}

int ProcessMasterInfo(char *Nick, char *channel, char *rest)
{
   send_to_user(Nick, "The MASTER level has been changed to MANAGER, to avoid confusion. Please see the LEVELS command for a description of all the levels available. Please use MANAGER in all places you previously used MASTER. Thanks -AfterNET CService Staff"); 

}

int do_test(char *nick, char *channel, char *rest)
{
  OperListType OpPtr;
  OperListType OpPtr2;

  if(!rest || !*rest)
  {
     send_to_user(nick, "test <nick>");
     return(FALSE);
  }
  OpPtr = FindBestUserMatch(rest, channel);
  OpPtr2 = FindBestUserMatchPass(rest, channel);
  send_to_user(nick, "Normal check:");
  if(OpPtr)
    send_to_user(nick, "\002%s\002 - [%c] \002%11s\002 %s <pass>", 
          levelnamefromint(userlevelbychar(OpPtr->Address[0])),
                   OpCh(OpPtr->AutoOp), NameCh(OpPtr->Handle), &OpPtr->Address[1]);
  else
    send_to_user(nick, "None found");
  send_to_user(nick, "Check without lamer override");
  if(OpPtr2)
    send_to_user(nick, "\002%s\002 - [%c] \002%11s\002 %s <pass>", 
                   levelnamefromint(userlevelbychar(OpPtr2->Address[0])), OpCh(OpPtr2->AutoOp), NameCh(OpPtr2->Handle), 
                   &OpPtr2->Address[1]); 
  else
    send_to_user(nick, "None found");

  return(TRUE);
}

int do_topic(char *from, char *Channel, char *rest)
{
  ChannelListType ChanNode;

  if (!(ChanNode = GetChannel(Channel)))
  {
     Debug(DBGWARNING, "CANT FIND CHANNEL IN do_topic()!");
     return(FALSE);
  }
  if(rest && *rest)
  {
     if((!IsGodOper(from)) && userlevel(from, Channel) < 200)
     {
        send_to_user(from, "You don't have access to change the default topic in %s, sorry.", Channel);
        return(FALSE);
     }
     if(!strcmp(rest, "*"))
     {
        send_to_user(from, "Default topic cleared in %s", Channel);
        strcpy(ChanNode->Topic, "*");
        return(TRUE);
     }
     else
     {
       free(ChanNode->Topic);
       if (ChanNode->Topic = (char *) malloc(strlen(rest) + 10))
       { 
         safecopy(ChanNode->Topic, rest, MAXTOPICLEN);
         send_to_user(from, "Default topic in %s changed to: %s", Channel, ChanNode->Topic);
	 WriteChanOpFile(Channel);
       }
       else
       {
           Debug(DBGSEVERE, "Out of memory allocating topic in do_topic()");
           return(FALSE);
       }
     }
  }
  if (ChanNode->settings[INCHAN] && (strcmp(ChanNode->Topic, "*")))
    sendtosock(Sock, ":%s TOPIC %s :%s", IrcCID, Channel, ChanNode->Topic);
  else
  {
    send_to_user(from, "\002No default topic for %s.\002", Channel);
    return(FALSE);
  }
  return(TRUE);
}

int do_topics(char *from, char *Channel, char *rest)
{
  ChannelListType ChanNode;

  if (*rest)
  {
    send_to_user(from, "\002Usage: TOPICS\002");
    send_to_user(from, "\002TOPICS Restores the default topic to all channels.\002");
    return(FALSE);
  }
  ResetTopics();
  return(TRUE);
}

#ifndef AFTERNET
int do_god(char *Nick, char *Channel, char *rest)
{
  UserListType UserNode;

  if (UserNode = (FindNick(Nick, UsersN)))
  {
    if (!strcasecmp(rest, "ON"))
    {
      UserNode->Modes = (UserNode->Modes | UMODEgod);
      send_to_user(Nick, "\002You now have authority to override channel security. Use carefully.\002");
    }
    else if (!strcasecmp(rest, "OFF"))
    {
      UserNode->Modes = (UserNode->Modes & ~UMODEgod);
      send_to_user(Nick, "\002You have given up authority to override channel security.\002");
      return(TRUE);
    }
    else
    {
      send_to_user(Nick, "\002Incorrect syntax.\002");
      send_to_user(Nick, "\002Usage: /msg x2 GOD ON/OFF\002");
      return(FALSE);
    }
  }
  else
  {
    send_to_user(Nick, "\002Error: cant find you in database!\002");
    return(FALSE);
  }
}
#endif

int do_whoami(char *Nick, char *Channel, char *rest)
{
  do_whois(Nick, Channel, Nick);
  return(TRUE);
}

int do_whois(char *Nick, char *Channel, char *rest)
{
  UserListType UserNode;
  char modetmp[20] = "";
  OwnerListType ChanList, TempChanList;

  /* The ownerlist is really for what chans were in,
   * but it used to only keep for owners.. the name is misleading
   * its for anyone whos LAMER/OP/MANAGER/(CO)OWNER etc */
  ChannelListType ChanPtr;
  char chans[MAXLEN];
  int ccount = 0;

  if (!rest)
  {
    send_to_user(Nick, "\002Usage: whois <nick>\002");
    return(FALSE);
  }
  if (UserNode = (FindNick(rest, UsersN)))
  {
    send_to_user(Nick, "\002-----: %s :-----\002", rest);
    send_to_user(Nick, "\002Nick: %s\002", UserNode->Nick);
#ifdef HYBRIDTS4
    send_to_user(Nick, "\002CID: %s\002", UserNode->CID);
#endif /* HYBRIDTS4 */
    send_to_user(Nick, "\002UserID: %s\002", UserNode->UserID);
    send_to_user(Nick, "\002Address: %s\002", UserNode->Address);
    send_to_user(Nick, "\002Server: %s\002", UserNode->Serv);
    send_to_user(Nick, "\002Clones: %d\002", UserNode->Clones);
    if (UserNode->Validated && (*UserNode->Validated != 0))
      send_to_user(Nick, "\002Validated on:%s\002", UserNode->Validated);
    else
      send_to_user(Nick, "\002Validated on: <NONE>\002");
    if (UserNode->Modes & UMODEd)
      strcat(modetmp, "d");
    if (UserNode->Modes & UMODEi)
      strcat(modetmp, "i");
    if (UserNode->Modes & UMODEo)
      strcat(modetmp, "o");
    if (UserNode->Modes & UMODEw)
      strcat(modetmp, "w");
    if (UserNode->Modes & UMODEk)
      strcat(modetmp, "k");
    if (UserNode->Modes & UMODEg)
      strcat(modetmp, "g");
    if (UserNode->Modes & UMODEx)
      strcat(modetmp, "x");
    if (UserNode->Modes & UMODEgod)
      strcat(modetmp, "(GOD)");    
    send_to_user(Nick, "\002Modes: +%s\002", modetmp);
    ChanList = UserNode->Chans;
    strcpy(chans, "");
    while (ChanList && (ccount < 10))
    {
      ccount++;
      strcat(chans, " ");
      strcat(chans, ChanList->Channel);
      ChanList = ChanList->Next;
    }
    send_to_user(Nick, "\002Known on channels:%s\002", chans);
    send_to_user(Nick, "\002-------- End ----------\002");
    return(TRUE);
  }
  else
  {
    send_to_user(Nick, "\002No such user %s\002", rest);
    return(FALSE);
  }
}

int do_roll(char *from, char *channel, char *rest)
{
  int my_rand;
  int size = 1;
  char *to;

  if (chan_to)
  {
    if (userlevel(from, chan_to) > 0)
      to = chan_to;
    else
      to = from;
  }
  else
  {
    to = from;
  }
  rest = strtok(rest, " ");
  if (rest && *rest)
  {
    size = atoi(rest);
  }
  else
    size = 6;
  if (size < 2)
  {
    send_to_user(from, "\002Die must have at least 2 sides.\002");
    return(FALSE);
  }
  my_rand = (rand() % size);
  my_rand++;
  send_to_user(to, "\002%s:\002 a \002%d\002 shows on the %d sided die.",
	       from, my_rand, size);
  return(TRUE);
}
#define LOCKE8BALL
#ifdef LOCKE8BALL

char *eightball(char *outcome, int method)
{
   int answer = 0;
   int test1[] = {1,2,5,4};

#define NUMOFCOLORS 17
   char ballcolors[40][17] = {"blue", "red", "green", "yellow",
        "white", "black","gray","brown",
        "yellow","pink","purple","orange","teal","burgandy",
        "fuchsia","turquoise","magenta"};
#define NUMOFLOCATIONS 45
   char balllocations[50][50] = { "Locke's house", "Oregon", "California", "Indiana", "Canada",
        "Russia", "Japan", "Florida", "The Bahamas", "Hiroshima",
        "The Carribean", "The Everglades", "your head", "your pants", "your School",
        "The Statue of Liberty", "Mt. Fugi", "your mother's house", "IRC", "OSU",
        "Locke's cat", "the closet", "the washroom", "a lake", "Spain"
        "a bathtub", "a toilet", "a sewer", "a horse", "Jupiter",
        "Uranis", "Pluto", "a dark place", "your undies", "your shirt",
        "your bra", "your hair", "bed", "the couch", "the wall",
        "your skin", "your brain", "Italy", "The Netherlands", "Germany"};
#define NUMOFPREPS 10
   char ballpreps[50][50] = { "Near", "In", "On", "Under", "Next to", 
        "above", "In place of", "Besides", "Far away from", "North of"};
#define NUMOFNUMS 10
   char ballnums[50][50] = { "Ten", "A hundred", "A few", "Aproximately 7",
        "about 9", "-42", "4", "8", "3", "2"};
#define NUMOFMULTS 4
   char ballmults[4][50] = { "Million", "", "thousand", ""};

   /* Method:
    * 0: normal
    * 1: color
    * 2: where is
    * 3: how many
    */
 
    if(method == 0)
    {
      answer = (rand() % 11); /* Make sure this is the # of entries */
      switch(answer)
      {
        case 0: strcpy(outcome, "Are you out of your MIND?");
                return;
        case 1: strcpy(outcome, "It won't happen, not a chance, definately no.");
                return;
        case 2: strcpy(outcome, "Outlook seems bleak.");
                return;
        case 3: strcpy(outcome, "My sources say no.");
                return;
        case 4: strcpy(outcome, "You bet!!");
                return;
        case 5: strcpy(outcome, "It is decidedly so.");
                return;
        case 6: strcpy(outcome, "It's hard to be sure.");
                return;
        case 7: strcpy(outcome, "Most definately.");
                return;
        case 8: strcpy(outcome, "In your Dreams.");
                return;
        case 9: strcpy(outcome, "If the prophets wish it...");
                return;
        case 10: strcpy(outcome, "Forcast hazey, try again.");
                return;
        case 11: strcpy(outcome, "I don't know!");
                return;
        default: strcpy(outcome, "An invalid random number was generated.");
                return;
      }
    }
    else if (method == 1) /* A Color */
    {
      char tmp[MAXLEN];

      answer = (rand() % 5); /* Make sure this is the # of entries */
      switch(answer)
      {
        case 0: strcpy(tmp, "Very bright %s I'd say.");
                break;
        case 1: strcpy(tmp, "Sort of a light %s color.");
                break;
        case 2: strcpy(tmp, "Dark and dreary %s.");
                break;
        case 3: strcpy(tmp, "Quite pale %s.");
                break;
        case 4: strcpy(tmp, "Mucky %s.");
                break;
        default: strcpy(outcome, "An invalid random number was generated.");
                return;
      }
      sprintf(outcome, tmp, ballcolors[rand() % NUMOFCOLORS]);
      return;
    }
    else if (method == 2)  /* Location */
    {
       sprintf(outcome, "%s %s.", ballpreps[rand() % NUMOFPREPS], balllocations[rand() % NUMOFPREPS]);
    }
    else if (method == 3)  /* Number of ___ */
    {
       sprintf(outcome, "%s %s.", ballnums[rand() % NUMOFNUMS], ballmults[rand() % NUMOFMULTS]);
    }
    else
    {
      Debug(DBGWARNING, "Error in eightball");
    }
    return(outcome);
}

int do_calc(char *Nick, char *Channel, char *rest)
{
  char *to;

  if(chan_to)
  {
     if(userlevel(Nick, chan_to) > 0)
       to = chan_to;
    else                                                                                            
      to = Nick;                                                                                 
  }                                                                                                 
  else                                                                                           
  {                                                                                                 
    to = Nick;                                                                                   
  }     
  if(!rest || !*rest)
  {
      send_to_user(Nick, "SYNTAX: CALC <expression>");
      send_to_user(Nick, "EX: CALC 5 + 3 / 6");
      return(TRUE);
  }
  do_math(Nick, to, rest);
}

int Process8ball(char *Nick, char *Channel, char *rest)
{
  char *word1, *word2, *word3;
  static char eb[MAXLEN];
  char *to;
  
  if (chan_to)
  {
    if (userlevel(Nick, chan_to) > 0)
      to = chan_to;
    else
      to = Nick;
  }
  else
  {
    to = Nick;
  }
  if(!rest || !*rest)
  {
    send_to_user(to, "\002%s:\002 The syntax is 8BALL [how many...?|where is...?|what color...?|yes or no question]", Nick);
    return;
  }
  word1 = (char *) strtok(rest, " ,.\"_[]()");
  word2 = (char *) strtok(NULL, " ,.\"_[]()");
  word3 = (char *) strtok(NULL, " ,.\"_[]()");

  if(!word1)
	  word1 = "";

/*** COLOR *****/
  if((word2) && strcasecmp(word1, "what") == 0 && strcasecmp(word2, "color") == 0)
     eightball(eb, 1);
  else if((word3) && strcasecmp(word1, "what's") == 0 && strcasecmp(word2, "the") == 0 && strcasecmp(word3, "color") == 0)
     eightball(eb, 1);
  else if((word3) && strcasecmp(word1, "whats") == 0 && strcasecmp(word2, "the") == 0 && strcasecmp(word3, "color") == 0)
     eightball(eb, 1);
/*** LOCATION *****/
  else if((word2) && (strcasecmp(word1, "where") == 0 && strcasecmp(word2, "is") == 0 || strcasecmp(word1, "where's") == 0))
     eightball(eb, 2);
/*** NUMBER *****/
  else if((word2) && strcasecmp(word1, "how") == 0 && strcasecmp(word2, "many") == 0)
     eightball(eb, 3);
/*** GENERIC *****/
  else
     eightball(eb, 0);
  send_to_user(to, "\002%s:\002 %s", Nick, eb);
return;
}

#endif

int OldProcess8ball(char *Nick, char *Channel, char *rest)
{
  int my_rand = 0;
  char outcome[MAXLEN];
  char *to;

  if (chan_to)
  {
    if (userlevel(Nick, chan_to) > 0)
      to = chan_to;
    else
      to = Nick;
  }
  else
  {
    to = Nick;
  }
  my_rand = (rand() % 11);
  switch (my_rand)
  {
  case 0:
    {
      strcpy(outcome, "Your outlook seems bleak.");
      break;
    }
  case 1:
    {
      strcpy(outcome, "Most definately.");
      break;
    }
  case 2:
    {
      strcpy(outcome, "It is decidedly so.");
      break;
    }
  case 3:
    {
      strcpy(outcome, "My sources say no.");
      break;
    }
  case 4:
    {
      strcpy(outcome, "Forcast hazey, try again.");
      break;
    }
  case 5:
    {
      strcpy(outcome, "It wont happen, not a chance, definately no.");
      break;
    }
  case 6:
    {
      strcpy(outcome, "Are you out of your MIND?");
      break;
    }
  case 7:
    {
      strcpy(outcome, "You bet!!");
      break;
    }
  case 8:
    {
      strcpy(outcome, "In your Dreams.");
      break;
    }
  case 9:
    {
      strcpy(outcome, "Its hard to be sure.");
      break;
    }
  case 10:
    {
      strcpy(outcome, "If the prophets wish it...");
      break;
    }
  }
  send_to_user(to, "\002%s\002: %s", Nick, outcome);
  return(TRUE);
}

int ProcessTest(char *Nick, char *Channel, char *rest)
{
  send_to_user(Channel, "\002Test passed.\002");
  send_to_user(Nick, "\002Test passed.\002");
  send_to_user(Channel, "\002Message from:%s on channel:%s\002", Nick, Channel);
  return(TRUE);
}

int do_inviteme(char *Nick, char *Channel, char *rest)
{
  sendtosock(Sock, ":%s INVITE %s %s", IrcCID, Nick, Channel);
  send_to_user(Nick, "\002You are invited to %s\002", Channel);
  return(TRUE);
}

int ProcessOpen(char *Nick, char *Channel, char *rest)
{
  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name '%s'", Channel);
    return(FALSE);
  }
  sendtosock(Sock, ":%s MODE %s", IrcCID, Channel);
  send_to_user(Nick, "Clearing %s of all modes.", Channel);
  channel_massunban(Channel);
  return(TRUE);
}

int ProcessInvite(char *Nick, char *Channel, char *UserNick)
{
  UserListType UserPtr;

/* This command was disabled due to abuse.. */
  UserNick = strtok(UserNick, " ");
  if(!UserNick || !*UserNick)
  {
     send_to_user(Nick, "USAGE: INVITE <nick>");
     return(FALSE);
  }
  UserPtr = FindNick(UserNick, UsersN);
  if(!UserPtr)
  {
     send_to_user(Nick, "Cant find %s on IRC", UserNick);
     return(FALSE);
  }

  sendtosock(Sock, ":%s INVITE %s %s", IrcCID, UserNick, Channel);
  send_to_user(UserNick, "%s has invited you to %s", Nick, Channel);
  send_to_user(Nick, "Inviting %s to %s", UserNick, Channel);
  return(TRUE);
}

int ProcessResetPass(char *Nick, char *Channel, char *UserNick)
{

  UserNick = (char *) strtok(UserNick, " ");
  if(!UserNick)
  {
      send_to_user(Nick, "USAGE: RESETPASS <Nick>");
      send_to_user(Nick, "where <Nick> is a user currently online.");
      return(FALSE);
  }
  if (!NickOk(UserNick))
  {
    send_to_user(Nick, "Illegal nick!");
    return(FALSE);
  }
  if (
       (!IsChanOwner(UserNick, Channel)) 
     &&
       (!IsChanCoOwner(UserNick, Channel))
     &&
       (!IsChanManager(UserNick, Channel)) 
     && 
       (!IsChanOp(UserNick, Channel))
     &&
       (!IsChanPeon(UserNick, Channel))
     )
  {
    send_to_user(Nick, "%s is not a user on %s.", UserNick, Channel);
    return(FALSE);
  }
  if((!IsGodOper(Nick)) && (userlevel(Nick, Channel) <= userlevel(UserNick, Channel)) )
  {
      send_to_user(Nick, "You cannot modify users at equal or higher access levels.");
      return(FALSE);
  } 
  ResetPass(UserNick, Channel);
  send_to_user(Nick, "Password for %s set to '*' (cleared).", UserNick);
  send_to_user(UserNick, "Your password has been cleared. To set a new password type: /msg %s %s pass * <New Pass>", IrcNick, Channel);
  return(TRUE);
}

int ProcessPass(char *Nick, char *Channel, char *rest)
{
  char *OldPass, *NewPass;

  OldPass = strtok(rest, " ");
  NewPass = strtok(NULL, " ");
  if (!NewPass)
  {
    send_to_user(Nick, "\002Usage: PASS <oldpass> <newpass>\002");
    send_to_user(Nick, "\002If you have no password, use * for oldpass.\002");
    return(FALSE);
  }
  if (userlevel(Nick, Channel) < 1)
  {
    send_to_user(Nick, "\002You are not on the %s userlist. You must be added with each address you use.\002", Channel);
    return(FALSE);
  }
  if(strlen(NewPass) > PASSWORDLEN)
  {
    send_to_user(Nick, "\002New password too long. Must be $d or less charactors.\002", PASSWORDLEN);
    return(FALSE);
  }
  if (!ChangePass(Nick, Channel, OldPass, NewPass))
  {
    send_to_user(Nick, "Your old password is incorrect! If you have NOT set a password before, your oldpassword is '*'.");
    return(FALSE);
  }
  send_to_user(Nick, "Your password has been changed to %s.", NewPass);
  return(TRUE);
}

int do_setinfo(char *Nick, char *Channel, char *info)
{
  OperListType OperPtr;
  ChannelListType ChanPtr;
  UserListType UserPtr;

  if (!info || !info[0])
  {
    send_to_user(Nick, "SYNTAX: SETINFO <infoline>");
    send_to_user(Nick, "Set your info to \"*\" to disable it.");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
    return(FALSE);
  if (!(UserPtr = FindNick(Nick, UsersN)))
    return(FALSE);
  if (strlen(info) > MAXINFOLEN);
    info[MAXINFOLEN + 1] = '\0';
  SetInfo(Nick, Channel, info);
  send_to_user(Nick, "Info changed to: %s", info);
  return(TRUE);
}

int do_rejoin(char *Nick, char *Channel, char *rest)
{
  sendtosock(Sock, ":%s WALLOPS :\002%s issued REJOIN\002", IrcCID, Nick);
  JoinChans();
  return(TRUE);
}

int do_routing(char *Nick, char *Channel, char *rest)
{
  int i;
  ServerListType TempRoutPtr, TempServPtr;

  send_to_user(Nick, "- Routing table -");
  for (i = 0; i < ROUTSEGSIZE; i++)
  {
    if (Routs->Table[i])
    {
      TempRoutPtr = Routs->Table[i];
      while (TempRoutPtr)
      {
	send_to_user(Nick, "\002%s <-> %s\002", TempRoutPtr->Name, TempRoutPtr->LinkName);
	TempRoutPtr = TempRoutPtr->Next;
      }
    }
  }
  send_to_user(Nick, "---- end ----");
  return(TRUE);
}

int do_loadrout(char *Nick, char *Channel, char *rest)
{
  char filename[MAXLEN];
  char buf[MAXLEN];
  int i = 0;

  if (!(rest = strtok(rest, " ")))
  {
    send_to_user(Nick, "\002Usage: LOADROUT <routfilename>\002");
    return(FALSE);
  }
  if ((strstr(rest, "..")) || (strstr(rest, "/")))
  {
    send_to_user(Nick, "\002Invalid parameter\002");
    return(FALSE);
  }
  sprintf(buf, "ROUTES/%s.dat", rest);
  KillRouts();
  i = ReadRoutFile(buf);
  if (i == 0)
  {
    send_to_user(Nick, "\002Error reading route file %s", buf);
    return(FALSE);
  }
  else if (i == 2)
  {
    send_to_user(Nick, "\002Error: %s not found.\002", buf);
    return(FALSE);
  }
  sendtosock(Sock, ":%s WALLOPS :\002%s is loading routing file %s\002", IrcCID, Nick, buf);
  do_routing(Nick, Channel, buf);
  return(TRUE);
}

int do_searchchan(char *Nick, char *Channel, char *rest)
{
  int i,j, count = 0;
  ChannelListType TempChanPtr;
  char chanmask[MAXLEN];

  if(!(rest && *rest))
  {
      send_to_user(Nick, "SYNTAX: SEARCHCHAN *name*!*topic*");
      send_to_user(Nick, "mask is similar to a hostmask.");
      return(FALSE);
  }
  count = 0;
  send_to_user(Nick, "Searching channels matcching: %s", rest);
  send_to_user(Nick, "=--Name-----------------------=----Topic-----------=");
  for(i = 0; i< CHANSEGSIZE; i++)
  {
     if(Channels->Table[i])
     {
        TempChanPtr = Channels->Table[i];
        while(TempChanPtr)
        {
           sprintf(chanmask, "%s!%s", TempChanPtr->Name, TempChanPtr->Topic);
           if(!matches(rest, chanmask))
           {
              if(((TempChanPtr->PlusModes & CMODEs) || (TempChanPtr->PlusModes & CMODEp)))
              {
                if(IsOper(Nick))
                {
                   send_to_user(Nick, "\002(SECRET)\002 %21s : %s", TempChanPtr->Name, TempChanPtr->Topic);
                   count++;
                }
              }
              else
              {
                send_to_user(Nick, "%30s : %s", TempChanPtr->Name, TempChanPtr->Topic);
                count++;

              }
           }
           TempChanPtr = TempChanPtr->Next;
        }
     }
  }
  if(!count)
    send_to_user(Nick, "No matching channels found.");
  send_to_user(Nick, "=-----------------------------=--------------------=");
  return(TRUE);
}

int do_wipesets_(char *Nick, char *Channel, char *rest)
{
  int i,j, count = 0;
  ChannelListType TempChanPtr;

  for(i = 0; i< CHANSEGSIZE; i++)
  {
     if(Channels->Table[i])
     {
        TempChanPtr = Channels->Table[i];
        while(TempChanPtr)
        {
           count++;
           for(j=0;j<NUMOFCHANSETTINGS;j++)
           {
               TempChanPtr->settings[j] = set_commands[j].def;
           }
           WriteChanOpFile(TempChanPtr->Name);
           TempChanPtr = TempChanPtr->Next;
        }
     }
  }
  send_to_user(Nick, "%d channel settings wiped to default values.", count);
  return(TRUE);
}

int do_rehash(char *Nick, char *Channel, char *rest)
{
  struct stat StatStruct;

  sendtosock(Sock, ":%s WALLOPS :\002%s is rehashing me!\002", IrcCID, Nick);
  send_to_user(Nick, "\002Killing routing database...\002");
  KillRouts();
  send_to_user(Nick, "\002Reloading routing database...\002");
  if (!ReadRoutFile())
  {
    sendtosock(Sock, ":%s WALLOPS :\002Error encountered while reading routing file.\002", IrcCID);
  }
  send_to_user(Nick, "\002Re-scanning news file...\002");
  stat(MESSAGEFILENAME, &StatStruct);
  NetMessage.Timestamp = StatStruct.st_mtime;
  send_to_user(Nick, "   NetMessage timestamp is %lu", NetMessage.Timestamp);

  send_to_user(Nick, "\002Completed rehash.\002");
  return(TRUE);
}

int ProcessQuit(char *Nick, char *Channel, char *rest)
{
  if (rest && *rest)
    sendtosock(Sock, ":%s QUIT :Quit by %s: \002%s\002", IrcCID, Nick, rest);
  else
    sendtosock(Sock, ":%s QUIT :QUIT by %s", IrcCID, Nick);
  sendtosock(Sock, ":%s SQUIT %s %s :Quitting", ServiceName, Uplink, ServiceName);
  KillChanList();
  KillUserList();
  KillServerList();
  KillRouts();
  unlink("./x2.pid");
/*   system("rm -f ./x2.pid"); */
  exit(0);
}

int ProcessSuspend(char *Nick, char *Channel, char *rest)
{
  sendtosock(Sock, ":%s QUIT :SUSPENDED by %s (Port: %d)", IrcCID, Nick, SERVERPORT);
  sendtosock(Sock, ":%s SQUIT %s :SUSPENDED by %s (Port: %d)", ServiceName, Uplink, Nick, SERVERPORT);
  AUTOCONNECT = 0;
  SockConnected = FALSE;
  return(TRUE);
}

int ProcessAuth(char *Nick, char *Channel, char *Pass)
{

  ChannelListType ChanPtr;

  Pass = strtok(Pass, " ");
  if (!Pass)
  {
    send_to_user(Nick, "\002Usage: /msg %s PASS <oldpass> <newpass>\002", IrcNick);
    send_to_user(Nick, "\002Note: If you have no password use '*'.  To disable password, set it to '*'.\002");
    send_to_user(Nick, "\002EXAMPLE: /msg %s PASS #afternet * dreamboat\002", IrcNick);
    return(FALSE);
  }
/*
  if (userlevel(Nick, Channel) < 1)
  {
    send_to_user(Nick, "\002You dont have access on channel %s.  Unable to authorize you.\002", Channel);
    return(FALSE);
  }
*/
  if (PassOk(Nick, Channel, "I hope nobody ever sets there password to this: ddddd"))
  {
    send_to_user(Nick, "Your password is disabled (set to '*').");
    return(TRUE);
  }
  if (PassOk(Nick, Channel, Pass))
  {
    send_to_user(Nick, "Your password for %s has been validated. Thank you.", Channel);
    return(TRUE);
  }
  else
  {
    send_to_user(Nick, "Incorrect Password!");
    return(FALSE);
  }
}

int do_seen(char *Nick, char *Channel, char *rest)
{
  time_t seentime, difference;
  char *to;
  char buf[MAXLEN] = "", tmp[MAXLEN] = "";

  int days, hours, minutes, seconds;

  if (chan_to)
  {
    if (userlevel(Nick, chan_to) > 0)
      to = chan_to;
    else
      to = Nick;
  }
  else
  {
    to = Nick;
  }
  rest = strtok(rest, " ");
  if (!rest)
  {
    send_to_user(Nick, "Usage: SEEN <handle>");
    send_to_user(Nick, "where <handle> is the handle in the channels userlist.");
    return(FALSE);
  }
  seentime = GetSeenTime(rest, Channel);
  if (seentime == 0)		/* not found */
  {
    send_to_user(to, "Sorry nobody on the %s userlist has a handle of %s.", Channel, rest);
    send_to_user(to, "This seen command only works for people in the channels userlist.");
    return(FALSE);
  }
  if (IsInChan(rest, Channel))
  {
    send_to_user(to, "\002%s\002: %s is in %s right now.", Nick, rest, Channel);
    return(TRUE);
  }

  difference = time(NULL) - seentime;
  days = difference / 86400;
  difference = difference - (days * 86400);
  hours = difference / 3600;
  difference = difference - (hours * 3600);
  minutes = difference / 60;
  difference = difference - (minutes * 60);
  seconds = difference;
  /* put our time together */
  if (days == 1)
    strcat(buf, "1 day ");
  else if (days > 1)
  {
    sprintf(tmp, "%d days ", days);
    strcat(buf, tmp);
  }
  if (hours == 1)
    strcat(buf, "1 hour ");
  else if (hours > 1)
  {
    sprintf(tmp, "%d hours ", hours);
    strcat(buf, tmp);
  }
  if (!days)	/* why show minutes if it's been over a day? */
  {
    if (minutes == 1)
      strcat(buf, "1 minute ");
    else if (minutes > 1)
    {
      sprintf(tmp, "%d minutes ", minutes);
      strcat(buf, tmp);
    }
    if (!hours)			/* no reason to display seconds if it has been over an hour */
    {
      if (seconds == 1)
	strcat(buf, "1 second ");
      else if (seconds > 1)
      {
	sprintf(tmp, "%d seconds ", seconds);
	strcat(buf, tmp);
      }
    }
  }
  send_to_user(to, "\002%s\002: I last saw %s on %s \002%s\002ago.", Nick, rest, Channel, buf);
  return(TRUE);
}

int ProcessAccess(char *Nick, char *Channel, char *UserNick)
{
  UserListType UserNode, OperUserNode;
  ChannelListType ChanPtr;
  char level[10] = "";
  char oper[10] = "";
  char *Info;

  if (!strcmp(UserNick, ""))
    UserNick = Nick;
  if (!(UserNode = FindNick(UserNick, UsersN)))
  {
    send_to_user(Nick, "Unknown user [%s]", UserNick);
    return(FALSE);
  }
  ChanPtr = GetChannel(Channel);
  if (!strcasecmp(IrcNick, UserNick))
  {
    send_to_user(Nick, "\002%s\002 is the channel services bot. /msg %s HELP for info on how to use it.", IrcNick, IrcNick);
    if (IsOper(UserNick))
      send_to_user(Nick, "%s is an AfterNet \002IRC Operator\002.", UserNode->Nick);
    if (IsGodOper(UserNick))
      send_to_user(Nick, "%s has \002security over-ride (+g) \002enabled.", UserNode->Nick);
    if (IsModeK(UserNick))
      send_to_user(Nick, "%s is a \002no-kick service (+k) \002", UserNode->Nick);
    return(FALSE);
  }
  if (IsChanOwner(UserNick, Channel))
    strcpy(level, "an OWNER");
  else if (IsChanCoOwner(UserNick, Channel))
    strcpy(level, "a CO-OWNER");
  else if (IsChanManager(UserNick, Channel))
    strcpy(level, "a MANAGER");
  else if (IsChanOp(UserNick, Channel))
    strcpy(level, "an OP");
  else if (IsChanPeon(UserNick, Channel))
    strcpy(level, "a PEON");
  else if (IsChanLamer(UserNick, Channel))
    strcpy(level, "LAMER");
  if (userlevel(UserNick, Channel) > 0)
  {
    send_to_user(Nick, "%s is \002%s\002 on %s, under the name \002%s\002.",
	UserNode->Nick, level, Channel, GetMaskName(UserNick, Channel));
    Info = GetInfo(UserNick, Channel);
    if (Info)
      if (Info[0] != '*')
	send_to_user(Nick, "[%s] %s", UserNick, Info);
    if (IsValidated(UserNick, Channel))
      send_to_user(Nick, "%s is authenticated on channel %s", UserNode->Nick, Channel);
    else
      send_to_user(Nick, "%s is NOT authenticated on channel %s", UserNode->Nick, Channel);
    if (IsChanAOp(Channel, UserNode->Nick))
      send_to_user(Nick, "%s has auto ops turned \002ON\002", UserNode->Nick);
    else
      send_to_user(Nick, "%s has auto ops turned \002OFF\002", UserNode->Nick);

  }
  else if (IsChanLamer(UserNick, Channel))
    send_to_user(Nick, "%s is %s from %s!", UserNode->Nick, level, Channel);
  else
    send_to_user(Nick, "%s is not on the %s userlist.", UserNode->Nick, Channel);
  if (IsOper(UserNick))
    send_to_user(Nick, "%s is an AfterNet \002IRC Operator\002.", UserNode->Nick);
  if (IsGodOper(UserNick))
    send_to_user(Nick, "%s has \002security over-ride (+g) \002enabled.", UserNode->Nick);
  if (IsModeK(UserNick))
    send_to_user(Nick, "%s is a \002no-kick service (+k) \002", UserNode->Nick);
  return(TRUE);
}

int do_settime(char *Nick, char *Channel, char *rest)
{
  time_t now, l;
  char buf[MAXLEN];
  char *param;
  char *ammount;

  now = time(NULL);

  param = (char *) strtok(rest, " ");
  ammount = (char *) strtok(NULL, " ");
  if(ammount)
  {
    if(!strcmp(param, "POWER"))
    {
       if(atol(ammount) != 0)
       {
         sendtosock(Sock, ":%s WALLOPS :%s is bringing the timestamp up SLOWLY", 
                                    ServiceName, Nick);
         for(l=atol(ammount);l<now;l+=999999)
            sendtosock(Sock, ":%s SETTIME %lu :%s", ServiceName, l, ServiceName);
       }
    }
    return(TRUE);
  }
  sendtosock(Sock, ":%s SETTIME %lu :%s", ServiceName, now, ServiceName);
  send_to_user(Nick, "\002Time set to %lu: %s\002", now, t2s(buf, now));
  NetOffset = 0;
  return(TRUE);
}

int ProcessGline(char *Nick, char *Channel, char *rest)
{
  int i;
  int Value, len;
  char *Arg1;
  char *Arg2;
  char *Arg3;
  char *Arg4;
  char *Address;
  char *UpCasePer;
  UserListType UserNode;
  time_t gtime;
  char buff[MAXLEN];

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;
  Address = strtok(rest, " ");
  Arg1 = strtok(NULL, " ");
  Arg2 = strtok(NULL, "\0");
  Address = strtok(Address, "$");
  if (!Arg1 || !Arg2 || !Address)
  {
    send_to_user(Nick, "\002Need more parameters.\002");
    return(FALSE);
  }
  if (!GlineAddyOk(Address))
  {
    send_to_user(Nick, "Illegal mask!");
    return(FALSE);
  }
  if (!strchr(Arg1, ':'))
  {
    send_to_user(Nick, "Illegal time Argument!");
    return(FALSE);
  }
  Arg3 = (char *) strtok(Arg1, ":");
  Arg4 = (char *) strtok(NULL, " \0");
  if (!Arg3 || !Arg4)
  {
    send_to_user(Nick, "\002Invalid time argument.\002");
    return(FALSE);
  }
  len = strlen(Arg3);

  if (len > 5)
  {
    send_to_user(Nick, "Illegal Time Value!");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg3[i]))
    {
      send_to_user(Nick, "Illegal Time Value!");
      return(FALSE);
    }
  }
  Value = atoi(Arg3);

  if ((strlen(Arg4) > 4))
  {
    send_to_user(Nick, "Illegal time period");
    return(FALSE);
  }
  UpCasePer = UpCase(Arg4);

  /* 11-21-97 GLINE <TS> is now GLINE <target (* for all)> ...
   * to be compatable with ircu2.10.  */
  if (!strcmp(UpCasePer, "MIN"))
    gtime = Value * 60;
  else if (!strcmp(UpCasePer, "HOUR"))
    gtime = Value * 3600;
  else if (!strcmp(UpCasePer, "DAY"))
    gtime = Value * 86400;
  else
  {
    send_to_user(Nick, "Illegal time Period. Use <#>:MIN/HOUR/DAY");
    free(UpCasePer);
    return(FALSE);
  }
  sprintf(buff, "<%s> %s", Nick, Arg2);
  sendtosock(Sock, ":%s GLINE * +%s %lu :%s", ServiceName, Address, gtime, buff);
  ProcessAddGline(Address, gtime, buff);

  send_to_user(Nick, "Adding GLINE for %s :<%s> %s", Address, Nick, Arg2);
  WriteGlineDat();
  free(UpCasePer);
  return(TRUE);
}

int ProcessSGline(char *Nick, char *Channel, char *rest)
{
  int i;
  int Value, len;
  char *Arg1;
  char *Arg2;
  char *Arg3;
  char *Arg4;
  char *Address;
  char *UpCasePer;
  UserListType UserNode;
  time_t gtime;
  char buff[MAXLEN];

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;
  Address = strtok(rest, " ");
  Arg1 = strtok(NULL, " ");
  Arg2 = strtok(NULL, "\0");
  if (!Arg1 || !Arg2 || !Address)
  {
    send_to_user(Nick, "\002Need more parameters.\002");
    return(FALSE);
  }
  if (!GlineAddyOk(Address))
  {
    send_to_user(Nick, "Illegal mask!");
    return(FALSE);
  }
  if (!strchr(Arg1, ':'))
  {
    send_to_user(Nick, "Illegal time Argument!");
    return(FALSE);
  }
  Arg3 = (char *) strtok(Arg1, ":");
  Arg4 = (char *) strtok(NULL, " \0");
  if (!Arg3 || !Arg4)
  {
    send_to_user(Nick, "\002Invalid time argument.\002");
    return(FALSE);
  }
  len = strlen(Arg3);

  if (len > 5)
  {
    send_to_user(Nick, "Illegal Time Value!");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg3[i]))
    {
      send_to_user(Nick, "Illegal Time Value!");
      return(FALSE);
    }
  }
  Value = atoi(Arg3);

  if ((strlen(Arg4) > 4))
  {
    send_to_user(Nick, "Illegal time period");
    return(FALSE);
  }
  UpCasePer = UpCase(Arg4);

  /* 11-21-97 GLINE <TS> is now GLINE <target (* for all)> ...
   * to be compatable with ircu2.10.  */
  if (!strcmp(UpCasePer, "MIN"))
    gtime = Value * 60;
  else if (!strcmp(UpCasePer, "HOUR"))
    gtime = Value * 3600;
  else if (!strcmp(UpCasePer, "DAY"))
    gtime = Value * 86400;
  else
  {
    send_to_user(Nick, "Illegal time Period. Use <#>:MIN/HOUR/DAY");
    free(UpCasePer);
    return(FALSE);
  }
  sprintf(buff, "<%s> %s", Nick, Arg2);
  ProcessAddSGline(Address, gtime, buff);
  send_to_user(Nick, "Adding SuperGLINE for %s :<%s> %s", Address, Nick, Arg2);
  WriteSGlineDat();
  free(UpCasePer);
  return(TRUE);
}


int ProcessUnglineAll(char *Nick, char *Channel, char *Address)
{
   UserListType UserNode;
   GlineListType *gptr, *gtmp, *glast;

  gptr = Glines;
  while(gptr)
  {
    gtmp = gptr->Next;
    safecopy(Address, gptr->address, MAXLEN);
    sendtosock(Sock, ":%s GLINE -%s", ServiceName, gptr->address);
    free(gptr);
    gptr = gtmp;
  }
  Glines = NULL;
  WriteGlineDat();
  return(TRUE);
}

int ProcessUngline(char *Nick, char *Channel, char *Address)
{

  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;
  if (!GlineAddyOk(Address))
  {
    send_to_user(Nick, "Illegal mask!");
    return(FALSE);
  }
  if(!GetGline(Address))
  {
     send_to_user(Nick, "Gline \"%s\" not found in memory. Clearing from server..", Address);
  }
  sendtosock(Sock, ":%s GLINE * -%s", ServiceName, Address);
  ProcessRemoveGline(Address);
  send_to_user(Nick, "Gline removed.");
  return(TRUE);
}

int ProcessUnsgline(char *Nick, char *Channel, char *Address)
{
  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;
  if (!GlineAddyOk(Address))
  {
    send_to_user(Nick, "Illegal mask!");
    return(FALSE);
  }
  if(!GetSuperGline(Address))
  {
     send_to_user(Nick, "No SuperGline: %s", Address);
     return;
  }
  ProcessRemoveSGline(Address);
  send_to_user(Nick, "SuperGline %s removed.", Address);
  return(TRUE);
}

int do_setglines(char *Nick, char *Channel, char *rest)
{
  GlineListType *gptr;
  int FORCE = 0;

  gptr = Glines;
  if (!gptr)
  {
    send_to_user(Nick, "There Are currently no glines in my database");
    return(FALSE);
  }
  if (!strcasecmp("HARD", rest))
  {
    FORCE = TRUE;
  }
  sendtosock(Sock, ":%s WALLOPS :\002%s is resynching the glines..", IrcCID, Nick);
  while (gptr)
  {
    if (FORCE)
      sendtosock(Sock, ":%s GLINE * -%s", ServiceName, gptr->address);
    sendtosock(Sock, ":%s GLINE * +%s %lu :%s", ServiceName, gptr->address, gptr->time - NetTime(), gptr->reason);
    gptr = gptr->Next;
  }
  return(TRUE);
}

int ProcessGlines(char *Nick, char *Channel, char *rest)
{
  GlineListType *gptr;
  char buff[MAXLEN];

  gptr = Glines;
  if (!gptr)
  {
    send_to_user(Nick, "There are no glines in my database");
    return(FALSE);
  }
  send_to_user(Nick, " Gline List: %s", rest);
  send_to_user(Nick, "=-------------Address--------------Time------------=");

  while (gptr)
  {
    if(!(*rest) || strstr(gptr->address, rest))
    {
       send_to_user(Nick, " %s  %s", gptr->address, t2s(buff, gptr->time));
       if(*rest)
         send_to_user(Nick, "      Reason : %s", gptr->reason);
    }
    gptr = gptr->Next;
  }
  send_to_user(Nick, "=----------------- End of Glines ----------------------=");
  return(TRUE);
}

int ProcessSGlines(char *Nick, char *Channel, char *rest)
{
  GlineListType *gptr;
  char buff[MAXLEN];

  gptr = SuperGlines;
  if (!gptr)
  {
    send_to_user(Nick, "There are no superglines in my database");
    return(FALSE);
  }
  send_to_user(Nick, " SuperGline List");
  send_to_user(Nick, "=-------------Address--------------Time------------=");

  while (gptr)
  {
    send_to_user(Nick, " %23s  %s", gptr->address, t2s(buff, gptr->time));
    send_to_user(Nick, "                  Reason : %s", gptr->reason);
    gptr = gptr->Next;
  }
  send_to_user(Nick, "=----------------- End of SuperGlines ---------------------=");
  return(TRUE);
}

int GetChanStats(char *Channel, int *w, int *c, int *m, int *o, int *p, int *b)
{
  OperSegment OpSeg;
  int i;
  OperListType OpList;

  *w = *m = *o = *b = *p = *c = 0;
  OpSeg = GetOpSegment(Channel);
  if (!OpSeg)
  {
    Debug(DBGERROR, "Channel not found but should be");
    return(FALSE);
  }
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      if (OpList->Address[0] == 'W')
	w[0]++;
      else if (OpList->Address[0] == 'C')
        c[0]++;
      else if (OpList->Address[0] == 'M')
	m[0]++;
      else if (OpList->Address[0] == 'O')
	o[0]++;
      else if (OpList->Address[0] == 'P')
        p[0]++;
      else if (OpList->Address[0] == 'L')
	b[0]++;
      else
      {
	Debug(DBGERROR, "I seem to have corruption in %s! Tell the admin GetChanStats found an unknown entry!!", Channel);
        return(FALSE);
      }
      OpList = OpList->Next;
    }
  }
  return(TRUE);
}

int Unvisited(char *Nick, char *Channel, char *Args)
{
  int i;
  ChannelListType ChanPtr;
  int Value, len;
  char *Arg1;
  char *Arg2;
  int count = 0;
  char *UpCasePer;
  int w, c, m, o, p, b;

  if (!strchr(Args, ':'))
  {
    send_to_user(Nick, "Illegal Argument! ");
    return(FALSE);
  }
  Arg1 = (char *) strtok(Args, ":");
  Arg2 = (char *) strtok(NULL, " \r\n");
  len = strlen(Arg1);
  if (!Arg2)
  {
    send_to_user(Nick, "Illegal Period. Use DAY, HOUR, or MIN.");
    return(FALSE);
  }
  if (len > 5)
  {
    send_to_user(Nick, "Illegal Time Value!");
    return(FALSE);
  }

  for (i = 0; i < len; i++)
  {
    if (!isdigit(Arg1[i]))
    {
      send_to_user(Nick, "Illegal Time Value!");
      return(FALSE);
    }
  }
  Value = atoi(Arg1);

  if (strlen(Arg2) > 4)
  {
    send_to_user(Nick, "Illegal Period");
    return(FALSE);
  }
  UpCasePer = UpCase(Arg2);

  if ((strcmp(UpCasePer, "MIN")) && (strcmp(UpCasePer, "HOUR")) && (strcmp(UpCasePer, "DAY")))
  {
    send_to_user(Nick, "Illegal Period");
    free(UpCasePer);
    return(FALSE);
  }
  send_to_user(Nick, "------------------------------------------------------");
  send_to_user(Nick, " Channels unvisited by an owner for %d %s(S) or more :", Value, UpCasePer);
  send_to_user(Nick, " ");

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    while (ChanPtr)
    {
      if (ChanPtr->LastAccess == 0)
      {
	ChanPtr = ChanPtr->Next;
	break;
      }
      if(ChanPtr->settings[NODELETE])
      {
        ChanPtr = ChanPtr->Next;
        break;
      }
      GetChanStats(ChanPtr->Name, &w, &c, &m, &o, &p, &b);
      if (!strcmp(UpCasePer, "MIN"))
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 60) >= Value)
	{
	  count++;
	  send_to_user(Nick, " %s   -   %d Min(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 60,
		       w, c, m, o, p, b);
	}
      }
      else if (!strcmp(UpCasePer, "HOUR"))
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 3600) >= Value)
	{
	  count++;
	  send_to_user(Nick, " %s   -   %d Hour(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 3600,
		       w, c, m, o, p, b);
	}
      }
      else
      {
	if (((time(NULL) - ChanPtr->LastAccess) / 86400) >= Value)
	{
	  count++;
	  send_to_user(Nick, " %s   -   %d Day(s): [%dw %dc %dm %do %dp %db]", ChanPtr->Name, (time(NULL) - ChanPtr->LastAccess) / 86400,
		       w, c, m, o, p, b);
	}
      }
      ChanPtr = ChanPtr->Next;
    }
  }
  if (!count)
    send_to_user(Nick, " None");
  send_to_user(Nick, "------------------------------------------------------");
  free(UpCasePer);
  return(TRUE);
}

int ProcessGlobal(char *Nick, char *Channel, char *rest)
{

  int i;
  ChannelListType ChanPtr;

  if(!rest)
  {
       send_to_user(Nick, "\002Usage: GLOBAL <message>\002");
       return(FALSE);
  }
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    while (ChanPtr)
    {
      sendtosock(Sock, ":%s PRIVMSG %s :\002Channel Service Announcment from <%s>:", IrcCID, ChanPtr->Name, Nick);
      sendtosock(Sock, ":%s PRIVMSG %s :%s", IrcCID, ChanPtr->Name, rest);
      ChanPtr = ChanPtr->Next;
    }
  }
  return(TRUE);
}

int ChanStat(char *Nick, char *Channel, char *rest)
{
  OperSegment OpSeg;
  OperListType OpList;
  ChannelListType ChanPtr;
  char ModeString[LINEBUFFSIZE];
  char timestring[LINEBUFFSIZE];
  time_t GoneTime;
  int OwnerNumber = 0, CoOwnerNumber = 0, ManagerNumber = 0, OpNumber = 0, PeonNumber = 0, LamerNumber = 0;
  int i, j, k, sval;
  BAN_list *bptr;

  ChanPtr = GetChannel(Channel);
  if (!ChanPtr)
  {
    Debug(DBGERROR, "Cannot find channel that should exist in ChanStat().");
    return(FALSE);
  }
  OpSeg = GetOpSegment(Channel);
  if (!OpSeg)
  {
    Debug(DBGERROR, "Cannot find opseg in ChanStat()");
    return(FALSE);
  }
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      if (OpList->Address[0] == 'W')
	OwnerNumber++;
      else if (OpList->Address[0] == 'C')
        CoOwnerNumber++;
      else if (OpList->Address[0] == 'M')
	ManagerNumber++;
      else if (OpList->Address[0] == 'O')
	OpNumber++;
      else if (OpList->Address[0] == 'P')
        PeonNumber++;
      else if (OpList->Address[0] == 'L')
	LamerNumber++;
      else
      {
	send_to_user(Nick, "\002I seem to have corruption in %s! Tell the admin!!\002", Channel);
      }
      OpList = OpList->Next;
    }
  }
  strcpy(timestring, ctime(&ChanPtr->CreationDate));
  timestring[strlen(timestring) - 1] = '\0';
  GoneTime = time(NULL) - ChanPtr->LastAccess;
  send_to_user(Nick, "=---------------------------------=");
  send_to_user(Nick, "        Channel Statistics         ");
  send_to_user(Nick, "=---------------------------------=");
  send_to_user(Nick, " Channel Name  : \002%s\002 ", ChanPtr->Name);
  send_to_user(Nick, " Created       : \002%s\002 by \002%s\002", timestring, ChanPtr->Reggedby);
  if (!strcmp(ChanPtr->Topic, "*"))
    send_to_user(Nick, " No default topic");
  else
    send_to_user(Nick, " Default Topic : %s", ChanPtr->Topic);
  if(strcmp(ChanPtr->Description, "*"))
    send_to_user(Nick, " Description   : %s", ChanPtr->Description);
  else
    send_to_user(Nick, " No Description");
  if(strcmp(ChanPtr->URL, "*"))
    send_to_user(Nick, " URL           : %s", ChanPtr->URL);
  else
    send_to_user(Nick, " No URL");
  send_to_user(Nick, " Default mode  : %s", GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, "????", ChanPtr->Limit));
  send_to_user(Nick, " ");
  send_to_user(Nick, " Num Of Owners  : %3d  | CoOwners : %3d", OwnerNumber, CoOwnerNumber);
  send_to_user(Nick, " Num Of Managers: %3d", ManagerNumber);
  send_to_user(Nick, " Num Of Ops     : %3d  | Peons    : %3d", OpNumber, PeonNumber);
  send_to_user(Nick, " Num Of Lamers  : %3d", LamerNumber);
  send_to_user(Nick, " ");
/*
  for (k = 0; set_commands[k].name != NULL; k++)
  {
    send_to_user(Nick, " \002%s\002%s", set_commands[k].pre, set_commands[k].settings[ChanPtr->settings[set_commands[k].index]]);
  }
*/
  if (OwnerNumber > 0)
  {
    if (ChanPtr->LastAccess == 0)
      send_to_user(Nick, "A (Co)Owner or Manager is currently in the channel.", GoneTime);
    else if (GoneTime < 60)
      send_to_user(Nick, "\002 Last Time Visited by a (Co)Owner/Manager:\002 %d secs ago.", GoneTime);
    else if (GoneTime < 3600)
      send_to_user(Nick, "\002 Last Time Visited by a (Co)Owner/Manager:\002 %d mins ago.", GoneTime / 60);
    else if (GoneTime < 86400)
      send_to_user(Nick, "\002 Last Time Visited by a (Co)Owner/Manager:\002 %d hours ago.", GoneTime / 3600);
    else
      send_to_user(Nick, "\002 Last Time visited by an Owner/Manager:\002 %d days ago.", GoneTime / 86400);
  }
  return(TRUE);
}

int do_mode(char *Nick, char *Channel, char *rest)
{
  ChannelListType ChanPtr;
  char ModeString[LINEBUFFSIZE];

  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "\002%s is not registered!\002", Channel);
    return(FALSE);
  }
  sendtosock(Sock, ":%s MODE %s -k %s", IrcCID, Channel, ChanPtr->CurrentKey);
  sendtosock(Sock, ":%s MODE %s %s", IrcCID, Channel,
	     GetModeString(ModeString, ChanPtr->PlusModes, 
             ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
  return(TRUE);
}

int do_set(char *Nick, char *Channel, char *rest)
{
  char *option, *param;
  int i, j, value, curset;
  ChannelListType ChanPtr;
  char ModeString[LINEBUFFSIZE];
  int ChangedINCHAN = 0;

  option = strtok(rest, " ");
  param = strtok(NULL, "\n");

  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "\002%s is not registered!\002", Channel);
    return(FALSE);
  }
  if (!option)
  {
    send_to_user(Nick, "=----------- %s Settings -------------=", ChanPtr->Name);
    send_to_user(Nick, "      \002Topic \002: %s", ChanPtr->Topic);
    send_to_user(Nick, "      \002Modes \002: %s", GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
    send_to_user(Nick, "        \002URL \002: %s", ChanPtr->URL);
    send_to_user(Nick, "\002Description \002: %s", ChanPtr->Description);
    for (i = 0; set_commands[i].name != NULL; i++)
    {
      if(set_commands[i].user)  /* If its not an IRCOP only command */
      send_to_user(Nick, "\002%10s\002 : %d  %s%s", set_commands[i].name,
	      ChanPtr->settings[set_commands[i].index], set_commands[i].pre,
	set_commands[i].settings[ChanPtr->settings[set_commands[i].index]]);
    }
    send_to_user(Nick, "=- The following are set by IRCOPS only -=");
    for (i = 0; set_commands[i].name != NULL; i++)
    {
      if(!set_commands[i].user)  /* If it IS an IRCOP only command */
      send_to_user(Nick, "%10s : %d  \002%s\002%s", set_commands[i].name,
              ChanPtr->settings[set_commands[i].index], set_commands[i].pre,
        set_commands[i].settings[ChanPtr->settings[set_commands[i].index]]);
    }

    send_to_user(Nick, "=----------------- End ------------------=");
    return(TRUE);
  }
  if(!strcasecmp(option, "URL"))
  {
    if(!param)
     send_to_user(Nick, "\002URL\002: %s", ChanPtr->URL);
    else
    {
      free(ChanPtr->URL);
      if(ChanPtr->URL = (char *) malloc(strlen(param) + 1))
      {
         safecopy(ChanPtr->URL, param, MAXURLLEN);
         send_to_user(Nick, "\002URL set to\002: %s", ChanPtr->URL);
      }
      else
      {
         send_to_user(Nick, "\002Error allocating memory for new URL!\002");
         return(FALSE); 
      }
      WriteChanOpFile(ChanPtr->Name);
    }
    return(TRUE);
  }
 if(!strcasecmp(option, "DESCRIPTION"))
  {
    if(!param)
     send_to_user(Nick, "\002Description\002: %s", ChanPtr->Description);
    else
    {
      free(ChanPtr->Description);
      if(ChanPtr->Description = (char *) malloc(strlen(param) + 1))
      {
         safecopy(ChanPtr->Description, param, MAXDESCLEN);
         send_to_user(Nick, "\002Description set to\002: %s", ChanPtr->Description);
      }
      else
      {
         send_to_user(Nick, "\002Error allocating memory for new description!\002");
         return(FALSE);
      }
      WriteChanOpFile(ChanPtr->Name);
    }
    return(TRUE);
  }

  if (!strcasecmp(option, "TOPIC"))
  {
    if (!param)
      send_to_user(Nick, "\002Topic\002: %s", ChanPtr->Topic);
    else
    {
      free(ChanPtr->Topic);
      if(strlen(param)>MAXTOPICLEN)
        param[MAXTOPICLEN] = '\0';
      if (ChanPtr->Topic = (char *) malloc(strlen(param) + 10))
      {
	safecopy(ChanPtr->Topic, param, MAXTOPICLEN);
	send_to_user(Nick, "\002Default topic set to\002: %s", ChanPtr->Topic);
      }
      else
      {
	send_to_user(Nick, "\002Error allocating memory for new topic!");
	return(FALSE);
      }
    }
    WriteChanOpFile(ChanPtr->Name);
    return(TRUE);
  }
  else if (!strcasecmp(option, "MODES") || !strcasecmp(option, "MODE"))
  {
    int n;
    if(param)
      for(n=strlen(param)-1;n>=0;n--)
      {
         if(isblank(param[n]))
           param[n] = '\0';
         else 
           break;
      }
    if (!param || !*param)
    {
      send_to_user(Nick, "\002Mode: %s",
		   GetModeString(ModeString, ChanPtr->PlusModes,
			  ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
    }
    else
    {
      if (userleveloper(Nick, Channel) < 200)
      {
	send_to_user(Nick, "\002You must be OWNER or MANAGER to set perm modes.\002");
	return(FALSE);
      }
      ConvertModeString(param, &ChanPtr->PlusModes, &ChanPtr->MinModes, &ChanPtr->Key, &ChanPtr->Limit);
      if(ChanPtr->settings[INCHAN])
      {
        sendtosock(Sock, ":%s MODE %s -k %s", IrcCID, Channel, ChanPtr->CurrentKey);
        sendtosock(Sock, ":%s MODE %s %s", IrcCID, Channel,
	    GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes,
			  ChanPtr->Key, ChanPtr->Limit));
      }
      send_to_user(Nick, "\002Default mode set to: %s",
		   ModeString);
      if (ChanPtr->Key)
	safecopy(ChanPtr->CurrentKey, ChanPtr->Key, MAXCHANKEY);
      else
	strcpy(ChanPtr->CurrentKey, "");

    }
    WriteChanOpFile(ChanPtr->Name);
    return(TRUE);
  }
  if (param)
  {
    if (sscanf(param, "%d", &value) != 1)
    {
      send_to_user(Nick, "\002Error: Parameter must be a number\002");
      return(FALSE);
    }
  }
  for (i = 0; set_commands[i].name != NULL; i++)
  {
    if (!strcasecmp(option, set_commands[i].name))
    {
      if (!param)		/* No parameter. */
      {
	curset = ChanPtr->settings[set_commands[i].index];
	send_to_user(Nick, "=-----------------------------=");
	send_to_user(Nick, "\002%s\002: %s\002", set_commands[i].name, set_commands[i].pre);
	send_to_user(Nick, "=---- * indicates current ----=");
	for (j = 0; j < 6; j++)
	{
	  if (set_commands[i].settings[j][0] == 0)
	    break;
	  if (j == curset)
	    send_to_user(Nick, "  %d: [*] %s", j, set_commands[i].settings[j]);
	  else
	    send_to_user(Nick, "  %d: [ ] %s", j, set_commands[i].settings[j]);
	}
	send_to_user(Nick, "=------------ End ------------=");
	return(TRUE);
      }
      else if (userleveloper(Nick, Channel) < 200)
      {
	send_to_user(Nick, "\002You must be OWNER or MANAGER to change settings.\002");
	return(FALSE);
      }
      if (!set_commands[i].user && !IsGodOper(Nick))
      {
        send_to_user(Nick, "\002%s can only be modified by the CService administrators.\002", set_commands[i].name);
	return(FALSE);
      }
      if ((value < set_commands[i].min) || (value > set_commands[i].max))
      {
	send_to_user(Nick, "\002Error, %s may only be %d-%d", set_commands[i].name,
		     set_commands[i].min, set_commands[i].max);
	return(FALSE);
      }
      if(!strcasecmp(option, "INCHAN"))
      {
         if(ChanPtr->settings[SUSPENDED])
         {
            send_to_user(Nick, "Channel %s has been suspended by the CService Admins. Email cservice@afternet.org for more information.", ChanPtr->Name);
            return;
         }
         ChangedINCHAN = TRUE;
      }
      ChanPtr->settings[set_commands[i].index] = value;                                             
      WriteChanOpFile(ChanPtr->Name);                                                            
      send_to_user(Nick, "\002%s\002 set to %d. \"\002%s\002%s\"", set_commands[i].name,         
                  value, set_commands[i].pre, set_commands[i].settings[value]);                        
      if(!strcasecmp(option, "SUSPENDED"))
      {
         ChangedINCHAN = TRUE;
         if(ChanPtr->settings[SUSPENDED] == 1)
         {
            send_to_user(Nick, "Channel %s marked as suspended.", ChanPtr->Name);
            ChanPtr->settings[INCHAN] = 0;
         }
         else
         {
            send_to_user(Nick, "Suspend cleared for channel %s", ChanPtr->Name);
            ChanPtr->settings[INCHAN] = 1;
         }
      }
      if(ChangedINCHAN)
      {
         if(ChanPtr->settings[INCHAN] == 0)  /* Leave the channel */
         {
            send_to_user(Nick, "Leaving disabled channel %s.  \002/msg %s %s set INCHAN 1\002 to re-enable.",
                                ChanPtr->Name, IrcNick, Channel);
            sendtosock(Sock, ":%s PART %s :Disabled by %s", IrcCID, Channel, Nick);
         }
         else  /* Join the channel */
         {
            if(ChanPtr->settings[SUSPENDED])
            {
                send_to_user(Nick, "This channel has been suspeneded by the CService adminsitrators. Email cservice@afternet.org for more information.");
                return;
            }
            send_to_user(Nick, "Re-joining channel %s.", ChanPtr->Name);
#ifdef HYBRIDTS4
            sendtosock(Sock, ":%s SJOIN %lu %s %s :@%s", ServiceName, time(NULL) + NetOffset, ChanPtr->Name,
                      GetModeString(ModeString, ChanPtr->PlusModes, 0, ChanPtr->Key, ChanPtr->Limit),
                      IrcCID);
#else /* HYBRIDTS4 */
            sendtosock(Sock, ":%s JOIN %s", IrcCID, ChanPtr->Name);
            sendtosock(Sock, ":%s MODE %s +o %s 0", ServiceName, ChanPtr->Name, IrcNick);
            sendtosock(Sock, ":%s MODE %s %s", IrcNick, ChanPtr->Name, 
                       GetModeString(ModeString, ChanPtr->PlusModes, 
                                     ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
#endif /* HYBRIDTS4 */
            if (strcmp(ChanPtr->Topic, "*") && ChanPtr->settings[TOPICFREQ])
               sendtosock(Sock, ":%s TOPIC %s :%s", IrcCID, ChanPtr->Name, ChanPtr->Topic);
         }
      } /* INCHAN */
      return(TRUE);
    }
  }
  send_to_user(Nick, "\002Unknown setting [\002%s\002].\002", option);
  return(FALSE);
}

int ProcessXSET(char *Nick, char *Channel, char *rest)
{

  char *UpCaseOpt;
  char *Option, *Value;
  int Val;
  int i;
  NetLinkListType NetLinkPtr, TempPtr;
  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return(FALSE);

/* gota break up rest like parse used to do.. */
  Option = strtok(rest, " ");
  Value = strtok(NULL, "\0");
  rest = Value;

  if (!Option)
  {
    send_to_user(Nick, "\002XSET <option> <value>\002");
    send_to_user(Nick, "\002Option  Function\002");
    send_to_user(Nick, "\002-------------------------------\002");
    send_to_user(Nick, "\002MAXCLONES      \002How many till i take action?");
    send_to_user(Nick, "\002CLONERES       \002What action do i take? 0 = nothing, 1 = warn, 2 = kill, 3 = gline *!*user@*host");
    send_to_user(Nick, "\002GATECHECK      \002Check for wingates? 0)no 1)warn 2)gline");
    send_to_user(Nick, "\002PINGOUTCONNECT \002Have x2 send a CONNECT when a server pings out.");
    send_to_user(Nick, "\002AUTOROUTING    \002Have x2 try to force routing to the currenly loaded route.dat");
    send_to_user(Nick, "\002AUTOCONNECT    \002Should X2 try to connect to a server, or wait for a connection (1 or 0 respectivly))");
    send_to_user(Nick, "\002DEBUG          \002What debug info should be displayed: 0=ALL, 60=NONE");
    send_to_user(Nick, "\002DBTARGET       \002Where should debug stuff go. DO NOT ABUSE THIS");
    send_to_user(Nick, "\002-------------------------------\002");
    return(FALSE);
  }
  if (!Value)
  {
    send_to_user(Nick, "\002Set %s to what?\002", Option);
    if (!strcasecmp(Option, "ENFORCE"))
      send_to_user(Nick, "\0020 = disabled, 1 = enabled.\002");
    if (!strcasecmp(Option, "GATECHECK")) 
      send_to_user(Nick, "\0020 = no action, 1 = warn ircops, 2 = gline\002");
    if (!strcasecmp(Option, "CLONERES"))
      send_to_user(Nick, "\0020 = no action, 1 = warn ircops, 2 = kill\002");
    return(FALSE);
  }
/* ok now to continue with xset.. */

  UpCaseOpt = UpCase(Option);
  if (!strcmp(UpCaseOpt, "AUTOROUTING"))
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. ");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if ((Val != 0) && (Val != 1))
    {
      send_to_user(Nick, "Value can be either 0 or 1.");
      free(UpCaseOpt);
      return(FALSE);
    }
    if (AUTOROUTING == Val)
    {
      send_to_user(Nick, "AUTOROUTING is already set to %d", AUTOROUTING);
      free(UpCaseOpt);
      return(FALSE);
    }
    AUTOROUTING = Val;
    sendtosock(Sock, ":%s WALLOPS :AUTOROUTING %s by %s", IrcCID, GetSetting(AUTOROUTING), Nick);
  }
  else if (!strcmp(UpCaseOpt, "PINGOUTCONNECT")) 
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. ");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if ((Val != 0) && (Val != 1))
    {
      send_to_user(Nick, "Value can be either 0 or 1.");
      free(UpCaseOpt);
      return(FALSE);
    }                    
    if (PINGOUTCONNECT == Val)
    {
      send_to_user(Nick, "PINGOUTCONNECT is already set to %d", PINGOUTCONNECT);
      free(UpCaseOpt);
      return(FALSE);
    }
    PINGOUTCONNECT = Val;
    sendtosock(Sock, ":%s WALLOPS :PINGOUTCONNECT %s by %s", IrcCID, GetSetting(PINGOUTCONNECT), Nick); 
  }
  else if (!strcmp(UpCaseOpt, "GATECHECK"))
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. ");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if ((Val < 0) || (Val > 2))
    {
      send_to_user(Nick, "Value can be either 0(no check), 1(warn ircops) or 2(gline).");
      free(UpCaseOpt);
      return(FALSE);
    }
    if (GATECHECK == Val)
    {
      send_to_user(Nick, "GATECHECK is already set to %d", GATECHECK);
      free(UpCaseOpt);
      return(FALSE);
    }
    GATECHECK = Val;
    sendtosock(Sock, ":%s WALLOPS :GATECHECK set to %d by %s", IrcCID, GATECHECK, Nick);
  }
  else if (!strcmp(UpCaseOpt, "AUTOCONNECT"))
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. ");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if ((Val != 0) && (Val != 1))
    {
      send_to_user(Nick, "Value can be either 0 or 1.");
      free(UpCaseOpt);
      return(FALSE);
    }
    if (AUTOCONNECT == Val)
    {
      send_to_user(Nick, "AUTOCONNECT is already set to %d", AUTOCONNECT);
      free(UpCaseOpt);
      return(FALSE);
    }
    AUTOCONNECT = Val;
    sendtosock(Sock, ":%s WALLOPS :AUTOCONNECT %s by %s", IrcCID, GetSetting(AUTOCONNECT), Nick);
  }
  else if (!strcmp(UpCaseOpt, "MAXCLONES"))
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. Value must be in the range 0-9");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if (MAXCLONES == Val)
    {
      send_to_user(Nick, "MAXCLONES is already set to %d", MAXCLONES);
      free(UpCaseOpt);
      return(FALSE);
    }
    MAXCLONES = Val;
    sendtosock(Sock, ":%s WALLOPS :MAXCLONES set to %d by %s", IrcCID, MAXCLONES, Nick);
  }
  else if (!strcmp(UpCaseOpt, "CLONERES"))
  {
    if ((strlen(Value) > 1) || (!isdigit(Value[0])))
    {
      send_to_user(Nick, "Illegal Value. ");
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if (Val > 3)
    {
      send_to_user(Nick, "Value must be in the range 0-3.");
      free(UpCaseOpt);
      return(FALSE);
    }
    if (CLONERES == Val)
    {
      send_to_user(Nick, "CLONERES is already set to %d", CLONERES);
      free(UpCaseOpt);
      return(FALSE);
    }
    CLONERES = Val;
    sendtosock(Sock, ":%s WALLOPS :CLONERES set to %s by %s", IrcCID, GetCloneResMet(CLONERES), Nick);
  }
  else if (!strcmp(UpCaseOpt, "DEBUG"))
  {
    if (!(Value && *Value))
    {
      free(UpCaseOpt);
      return(FALSE);
    }
    Val = atoi(Value);
    if ((Val < 0) || (Val > 99))
    {
      send_to_user(Nick, "Must be 0 though 99");
      free(UpCaseOpt);
      return(FALSE);
    }
    DEBUG = Val;
    sendtosock(Sock, ":%s WALLOPS :DEBUG set to %d by %s", IrcCID, DEBUG, Nick);
    free(UpCaseOpt);
    return(TRUE);
  }
  else if (!strcasecmp(UpCaseOpt, "DBTARGET"))
  {
    if (Value && *Value && (strcasecmp("NULL", Value)))
    {
      safecopy(DBTARGET, Value, MAXNICKLEN);
      sendtosock(Sock, ":%s WALLOPS :DEBUG TARGET set to %s by %s", IrcCID, DBTARGET, Nick);
      free(UpCaseOpt);
      return(TRUE);
    }
    else
    {
      strcpy(DBTARGET, "");
      sendtosock(Sock, ":%s WALLOPS :debugging disabled by %s", IrcCID, Nick);
      free(UpCaseOpt);
      return(TRUE);
    }
  }
  else
  {
     send_to_user(Nick, "Unknown Setting '%s'", Option);
     return(FALSE);
  }
  free(UpCaseOpt);
  return(TRUE);
}

int do_showlevels(char *Nick, char *Channel, char *info)
{
  send_to_user(Nick, "\002OWNER\002   - Only one person per channel can have this access level.");
  send_to_user(Nick, "\002COOWNER\002 - Can do the same as Owners, except for modifying CoOwners.");
  send_to_user(Nick, "\002MANAGER\002 - Can add OP, PEON, and LAMER levels, but cannot set perminant modes.");
  send_to_user(Nick, "\002OP\002      - Cannot add users or lamers, but can use x2 to do remote commands such as kick and op.");
  send_to_user(Nick, "\002PEON\002    - Is known in the seen command, can be given +v, but no other access.");
  send_to_user(Nick, "\002LAMER\002   - Is kickbanned from the channel on sight.  Reason optionally given.");
}

int do_listlamers(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'L';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);
}

int do_listops_note(char *Nick, char *Channel, char *rest)
{
  send_to_user(Nick, "\002NOTE: The LISTOPS command has been changed to USERLIST.  We hope this will be less confusing in the long run.  Please accept our apologys for the confusion.\002");
  send_to_user(Nick, "\002         Thanks,    -The AfterNET IRC Network\002");
  return(FALSE);
}

int do_listpeons(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'P';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);
}

int do_listops(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'O';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);

}

int do_listmanagers(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'M';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);

}

int do_listowners(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'W';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);

}

int do_listcoowners(char *Nick, char *Channel, char *rest)
{
  LISTTYPE = 'C';
  ShowOps(Nick, Channel, rest);
  LISTTYPE = '\0';
  return(TRUE);

}

int ShowOps(char *Nick, char *Channel, char *rest)
{
  ChannelListType ChanPtr;
  OperListType OpList;
  OperSegment OpSeg;
  int i, j, count = 0;

  ChanPtr = GetChannel(Channel);
  if (!ChanPtr)
  {
    send_to_user(Nick, "Channel %s is not registered. Info not available!", Channel);
    return(FALSE);
  }
  if (LISTTYPE == '\0')
    send_to_user(Nick, "%s User list - All levels:", Channel);
  else if (LISTTYPE == 'W')
    send_to_user(Nick, "%s User list - Owners only:", Channel);
  else if (LISTTYPE == 'C')
    send_to_user(Nick, "%s User list - Co-Owners only:", Channel);   
  else if (LISTTYPE == 'M')
    send_to_user(Nick, "%s User list - Managers only:", Channel);
  else if (LISTTYPE == 'O')
    send_to_user(Nick, "%s User list - Ops only:", Channel);
  else if (LISTTYPE == 'P')
    send_to_user(Nick, "%s User list - Peons only:", Channel);   
  else if (LISTTYPE == 'L')
    send_to_user(Nick, "%s Lamer list:", Channel);
  else
    send_to_user(Nick, "\002Ack! LISTTYPE is not WCHOP or L in ShowOps()!\002");
  if(LISTTYPE == 'L')
    send_to_user(Nick, "=-\002Level\002---\002Handle\002[\002address mask\002]--\002reason\002-=");
  else
    send_to_user(Nick, "=-\002Level\002---\002Aop\002----\002Handle\002--------\002Mask\002--------------\002LastSeen\002---=");
  OpSeg = ChanPtr->Opers;
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OpList = OpSeg->Table[i];
    while (OpList)
    {
      count++;
      if (((*rest && (!match(rest, &OpList->Address[1]) || !match(&OpList->Address[1], rest) || !strcasecmp(OpList->Handle, rest)))) || (!(*rest)))
      {
        if((OpList->Address[0] == LISTTYPE) || LISTTYPE == '\0')
        {
           char buff[MAXLEN] = "";
           char PaddedLevel[50] = "";
           char SeenStr[MAXLEN] = "";
           char tmp[MAXLEN];
           char PassStr[10] = "";
           int pad, days, hours;
           time_t seentime, difference;

           seentime = OpList->LastSeen;

           difference = time(NULL) - seentime;
           days = difference / 86400;
           difference = difference - (days * 86400);
           hours = difference / 3600;
              if (days == 1)
                  strcat(SeenStr, "\0021\002d ");
              else if (days > 1)
              {
                  sprintf(tmp, "\002%d\002d ", days);
                  strcat(SeenStr, tmp);
              }
              if (hours == 1)
                  strcat(SeenStr, "\0021\002h ");
              else if (hours > 1)
              {
                 sprintf(tmp, "\002%d\002h ", hours);
                 strcat(SeenStr, tmp);
              }
              if(!*SeenStr)
              {
                 if(IsInChan(OpList->Handle, Channel))
                    strcat(SeenStr, "In Chan");
                 else
                    strcat(SeenStr, "Just left");
              }
           if(strcmp(OpList->Pass, "*"))
               strcpy(PassStr, "<pass>");
           for(j=(7 - strlen(levelnamefromchar(OpList->Address[0])));j>0;j--)
             strcat(PaddedLevel, " ");
           strcat(PaddedLevel, levelnamefromchar(OpList->Address[0]));
           if(OpList->Address[0] == 'L')
           {
             if(LISTTYPE == 'L')
                send_to_user(Nick, "\002%s\002 - %s[%s] %s", PaddedLevel, NameCh(OpList->Handle), &OpList->Address[1], OpList->Info);
           }
           else
             send_to_user(Nick, "\002%s\002 - [%c] \002%11s\002 %s %s %s",
                        PaddedLevel, OpCh(OpList->AutoOp), NameCh(OpList->Handle),
                        &OpList->Address[1], PassStr, SeenStr);
        }
      }
      OpList = OpList->Next;
    }
  }
  if (!count)
  {
    send_to_user(Nick, "There are no registered users!");
    return(FALSE);
  }
  else
    send_to_user(Nick, "=-------- End of List --------=");
  return(TRUE);
}

int NetStat(char *Nick, char *Channel, char *rest)
{
  float ave = 0;

  ave = (float) OperCount / (float) ServerCount;
  send_to_user(Nick, "---------------------------------");
  send_to_user(Nick, "      N E T    S T A T S         ");
  send_to_user(Nick, "---------------------------------");
  send_to_user(Nick, " Number of Users:   %d", UserCount);
  send_to_user(Nick, " Number of Servers: %d", ServerCount);
  send_to_user(Nick, " Number of IRC Ops: %d", OperCount);
  send_to_user(Nick, " Avg Number of Opers per Server currently online: %1.1f", ave);
  send_to_user(Nick, " ");
  send_to_user(Nick, " There are %d registered channels.", ChannelCount);
  send_to_user(Nick, " Total number of Peons : %d ", PeonCount);
  send_to_user(Nick, " Total number of Ops : %d ", OpCount);
  send_to_user(Nick, " Total number of Managers : %d ", ManagerCount);
  send_to_user(Nick, " Total number of Co-Owners : %d ", CoOwnerCount);
  send_to_user(Nick, " Total number of Owners : %d ", OwnerCount);
  if (ChannelCount > 0)
    send_to_user(Nick, " Avg Number of Peons+Opers+Managers+(Co)Owners per channel: %d", ((PeonCount + ManagerCount + OpCount + OwnerCount) / ChannelCount));
  send_to_user(Nick, " End of stats");
  return(TRUE);
}

int do_uptime(char *from, char *Channel, char *rest)
{
  time_t UpTime, ConTime;
  char buf[MAXLEN];
  extern int NumOfGateSocks;

  UpTime = time(NULL) - StartTime;
  ConTime = time(NULL) - ConnectTime;
  send_to_user(from, "\002Uptime    %d day(s) (%02d:%02d)\002", UpTime / 86400, (UpTime / 3600) % 24, (UpTime / 60) % 60);
  send_to_user(from, "\002Connected %d day(s) (%02d:%02d)\002", ConTime / 86400, (ConTime / 3600) % 24,(ConTime / 60) % 60);
  send_to_user(from, "\002My clock       : %s", t2s(buf, time(NULL)));
  send_to_user(from, "\002AfterNet clock : %s", t2s(buf, NetTime()));
  send_to_user(from, "\002%d commands parsed from %d lines processed.\002", TotalCommands, TotalLines);
  send_to_user(from, "\002%d users signed on.\002", TotalUserCount);
  if(GATECHECK)
     send_to_user(from, "\002There are %d socket(s) in use checking for wingates.\002", NumOfGateSocks);
   send_to_user(from, "\002NETBURST mode is %s", NETBURSTING?"ON":"OFF");
  return(TRUE);
}

int Settings(char *Nick, char *Channel, char *rest)
{
  time_t UpTime;

  UpTime = time(NULL) - StartTime;
  send_to_user(Nick, " ");
  send_to_user(Nick, " X2 Uptime: Day %d (%02d:%02d)", UpTime / 86400, (UpTime / 3600) % 24, (UpTime / 60) % 60);
  send_to_user(Nick, " ");
  send_to_user(Nick, "\002 X2 NET SETTINGS:");
  send_to_user(Nick, " --------------------");
  send_to_user(Nick, "\002  PINGOUTCONNECT\002: %s", GetSetting(PINGOUTCONNECT));
  send_to_user(Nick, "\002  AUTOROUTING   \002: %s", GetSetting(AUTOROUTING));
  send_to_user(Nick, "\002  AUTOCONNECT   \002: %s", GetSetting(AUTOCONNECT));
  send_to_user(Nick, "\002  GATECHECK     \002: %s", GetWinGateMet(GATECHECK));
  send_to_user(Nick, "\002  MAXCLONES     \002: %d", MAXCLONES);
  send_to_user(Nick, "\002  CLONERES      \002: %s", GetCloneResMet(CLONERES));
  send_to_user(Nick, "\002  DEBUG         \002: %d", DEBUG);
  if (*DBTARGET)
    send_to_user(Nick, "\002  DBTARGET      \002: %s", DBTARGET);
  else
    send_to_user(Nick, "\002  DBTARGET      \002: NULL");
  send_to_user(Nick, " End of settings");
  return(TRUE);
}

int NumDownLinks(char *Name)
{
  ServerListType ServPtr;
  int i, num = 0;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if (!(strcasecmp(ServPtr->LinkName, Name)))
	num++;
      ServPtr = ServPtr->Next;
    }
  }
  return (num);
}

int ShowDownLinks(char *Nick, char *Channel, char *Name, char *prevpre, char *arrowchar)
{
  ServerListType ServPtr;
  int i, j;
  char pre[MAXLEN];
  char *nextpre;
  char buf[MAXLEN];
  int num = 0;
  static int depth = 0;

  nextpre = malloc(MAXLEN);
  safecopy(pre, prevpre, MAXLEN);

  if (ServPtr = GetServer(Name))
  {
    if (!strstr(ServPtr->Name, "AfterNET.Services"))
    {
      if (ServPtr->NumOfUsers != 1)
	send_to_user(Nick, "%s%s-\002%s\002 [%d users]", pre, arrowchar, Name, ServPtr->NumOfUsers);
      else
	send_to_user(Nick, "%s%s-\002%s\002 [1 user]", pre, arrowchar, Name);
    }
    else
      send_to_user(Nick, "%s%s-\002%s\002 (service)", pre, arrowchar, Name);
  }
  else				/* its me */
    send_to_user(Nick, "\002%s\002 (me)", ServiceName);
  j = NumDownLinks(Name);
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if (!(strcasecmp(ServPtr->LinkName, Name)))
      {
	safecopy(nextpre, pre, MAXLEN);
	if (depth++ > 0)
	{
	  if (arrowchar[0] == '`')
	    strcat(nextpre, "  ");
	  else
	    strcat(nextpre, "| ");
	}
	if (j > ++num)
	{
	  ShowDownLinks(Nick, Channel, ServPtr->Name, nextpre, "|");
	}
	else
	{
	  ShowDownLinks(Nick, Channel, ServPtr->Name, nextpre, "`");
	}
      }
      ServPtr = ServPtr->Next;
    }
  }
  free(nextpre);
}

/* used by ShowServers */
int do_map(char *Nick, char *Channel, char *Name)
{
  int depth = 0;

  ShowDownLinks(Nick, Channel, ServiceName, "", "`");
  send_to_user(Nick, "End of MAP");
  return(TRUE);
}

int ShowServers(char *Nick, char *Channel, char *rest)
{

  int i, j;
  ServerListType ServPtr;

  send_to_user(Nick, " ------------------------------");
  send_to_user(Nick, " Servers and their user counts:");
  send_to_user(Nick, " ------------------------------");
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      if(!strstr(ServPtr->Name, "AfterNET.Services") || (strstr(rest, "DEBUG") ))	/* filter out AfterNET.Services */
	send_to_user(Nick, "\002 %s\002 has %d user(s) <-> \002%s\002", ServPtr->Name, ServPtr->NumOfUsers, ServPtr->LinkName);
      ServPtr = ServPtr->Next;
    }
  }
  send_to_user(Nick, " =--------------END--------------=");
  return(TRUE);
}

int ExpireChans(char *Nick, char *Channel)
{
   send_to_user(Nick, "Removing unvisited channels..");
   ExpireChannels(Nick);
   return(TRUE);
}

int UnregChan(char *Nick, char *Channel)
{
  UserListType UserNode;
  ChannelListType ChanPtr;
  FILE *ChanFile;
  FILE *NewChanFile;
  char *UpCaseChan;
  char str[LINEBUFFSIZE];
  char ChanName[LINEBUFFSIZE], rest[LINEBUFFSIZE];
  time_t LastAccess, CreationDate;
  int s, line = 0;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return(FALSE);
  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name!");
    return(FALSE);
  }
  ChanPtr = GetChannel(Channel);
  if (!ChanPtr)
  {
    send_to_user(Nick, "Channel %s is not registered!", Channel);
    return(FALSE);
  }
  DelChan(Channel);
  ChanFile = fopen(CHANFILENAME, "r");
  if (ChanFile < 0)
  {
    send_to_user(Nick, "Error: op file doesn't not exist!");
    return(FALSE);
  }
  NewChanFile = fopen("tmp/.tmpchanfile.dat", "w");
  if(!NewChanFile)
  {
     Debug(DBGSEVERE, "Error opening temp file in unreg()!!");
     return(FALSE);
  }
  while ((s = fscanf(ChanFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    {
      Debug(DBGERROR, "Error with param count in chan.dat! s=%d line=%d", s, line);
      send_to_user(Nick, "\002Error with param count in chan.dat! s=%d line=%d\002", s, line);
      fclose(ChanFile);
      fclose(NewChanFile);
      return(FALSE);
    }
    if (strcasecmp(ChanName, Channel))
      fprintf(NewChanFile, "%s\n", ChanName);
  }
  if(ferror(NewChanFile))
  {  
     Debug(DBGSEVERE, "Error writing to .tmpchanfile.dat (Out of disk space?). Unregistration NOT saved.", Channel);
     fclose(NewChanFile);
     sleep(4);
     exit(2);
  }
  fclose(ChanFile);
  fclose(NewChanFile);
  sprintf(str, "./%s", CHANFILENAME);
  unlink(str);
  rename("tmp/.tmpchanfile.dat", str);
  sprintf(str, "./CHANS/%s", UpCaseChan = UpCase(&Channel[1]));
  unlink(str);
  free(UpCaseChan);
  sendtosock(Sock, ":%s PART %s", IrcCID, Channel);
  sendtosock(Sock, ":%s WALLOPS :Channel %s unregistered by %s.", IrcCID, Channel, Nick);
  return(TRUE);
}

int do_savechan(char *Nick, char *Channel, char *test)
{
  ChannelListType ChanPtr;

  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered!", Channel);
    return(FALSE);
  }
  WriteChanOpFile(Channel);
  send_to_user(Nick, "\002Wrote %s user data to disk.\002", Channel);
  return(TRUE);
}

int do_loadchan(char *Nick, char *Channel, char *rest)
{
  FILE *OpFile;
  char linebuff[LINEBUFFSIZE];
  char *filename, *sepptr;
  char *ptrbuff, *key;
  char FullChanName[LINEBUFFSIZE];
  int i,s,line2=0,fileversion=0;
  int UserAutoOp;
  char *OpAddress, *Pass, *Info, *Handle;
  ChannelListType ChanPtr;
  time_t LastSeen;

  if(!rest || !*rest)
  {
     send_to_user(Nick, "SYNTAX: LOADCHAN <absolute path>");
     send_to_user(Nick, "path must point to a channel data file with at least CHANNEL VERSION and USER lines.");
     send_to_user(Nick, "This command loads a file from the local file system into memory.");
     send_to_user(Nick, "It is for the X2 admin to quickly restore a single channel from backup files.");
     send_to_user(Nick, "If anyone uses this command who is not directly related to the account X2 runs from, you are breaking the law!. ");
     send_to_user(Nick, "Unauthorized access of a computer system in Oregon carries very real penalties. HANDS OFF!");
     return(0);
  }
  filename = strtok(rest, " ");
  if(!filename)
     return(0);
  strcpy(linebuff, "./");
  strcat(linebuff, filename);
  filename = linebuff;
  if(OpFile = fopen(filename, "r"))
  {
     line2 = 0;
     while(!feof(OpFile))
     {
        i++;
        s = fscanf(OpFile, "%[^\n]\n", linebuff);
        line2++;
        if(s == EOF)
           break;
        if(s != 1)
        {
           send_to_user(Nick, "The format of %s does not appear to be correct.", filename);
           fclose(OpFile);
           return;
        }
/*        key = (char *) strtok(linebuff, " "); */
        sepptr = linebuff;
        key = (char *) mysep(&sepptr, " ");
        if(!key || !*key)
          return;
        if(line2 == 1 && key[0] != 'V')
        {
           Debug(DBGWARNING, "%s is attempting to load a file which is not a channel data file. BUSTED!", Nick);
           return(0);
        }
        switch(key[0])
        {
         case 'V': /* VERSION */                              
/*                ptrbuff = (char *) strtok(NULL, " ");    */        
                ptrbuff = (char *) mysep(&sepptr, " ");
                if(!ptrbuff)                                    
                {                                               
                    send_to_user(Nick,"Invalid VERSION syntax in %s line %d", filename, line2);
                    fclose(OpFile);
                    return(0);                                 
                }                                             
                fileversion = strtol(ptrbuff, NULL, 10);        
                if(fileversion > 1)                           
                {                                               
                   send_to_user(Nick,"Warning! %s is newer fileversion %d", 
                                filename, fileversion);
                }                                               
                continue; 
        case 'C': /* Channel */
/*               ptrbuff = (char *) strtok(NULL, "\n"); */
               ptrbuff = (char *) mysep(&sepptr, " ");
               safecopy(FullChanName, ptrbuff, MAXLEN);
               if(!(ChanPtr = GetChannel(FullChanName)))
               {
                  send_to_user(Nick, "This file describes %s which is not registered. Not Loaded",
                                   FullChanName);
                  fclose(OpFile);
                  return;
               }
               continue;
         case 'U': /* USER */
                if(!ChanPtr)
                {
                   send_to_user(Nick, "USER decleration before any CHANNEL decleration. Aborting load.");
                   fclose(OpFile);
                   return(0);
                }
                OpAddress = (char *)mysep(&sepptr, " ");
                ptrbuff = (char *)mysep(&sepptr, " ");
                UserAutoOp = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *)mysep(&sepptr, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Handle = (char *)mysep(&sepptr, " ");
                Pass = (char *)mysep(&sepptr, " ");
                Info = (char *)mysep(&sepptr, "\n");
                /* Remove this after the data files are updated */
                if(OpAddress[0] == 'H')
                    OpAddress[0] = 'M';
                /* ******************************************** */
                if((!Info)
                   || (strlen(Pass) > MAXPASS)                                        
                   || (strlen(Handle) > 20)
                   || (strlen(Info) > 200)
                   || (!AddyOk(&OpAddress[1]))                                                      
                  )
                {
                   send_to_user(Nick, "Error in %s line %d. Attempting to continue..", filename, line2);
                   fclose(OpFile);
                   continue;
                }
                if (FindOp(ChanPtr->Opers, &OpAddress[1]))                                         
                {                                                                                
                   send_to_user(Nick,"Attempting to load a duplicate userhost. Ignoring %s (line %d)", 
                                OpAddress, line2);
                   continue;
                }
                else if (!(AddOp(ChanPtr->Opers, OpAddress, Handle, Pass, Info, UserAutoOp, LastSeen)))
                {                                                                              
                   send_to_user(Nick,"Error adding op in %s line %d, [%s]\n", 
                                filename, line2, OpAddress);
                }                                                                                
                continue; 
         default: 
                continue;
        }
     }
     WriteChanOpFile(FullChanName);
     send_to_user(Nick, "%s loaded as %s.", filename, FullChanName);
     return(1);
  }
  else
  {
    send_to_user(Nick, "%s not found", filename);
  }
}

int ScriptReg(char *Nick, char *Channel, char *rest)
{
   /* This is a front-end to regchan, topic, addowner etc */
   /* So a script can be more easily made to register channels */

   /* Syntax is: */
   /* SCRIPTREG <#channel> <OwnerNick> <*!*hostmask@*host.com> +[modes] [:topic] */
   char *ChanName, *OwnerNick, *OwnerMask, *Modes, *Topic;
   char buff[MAXLEN];

   ChanName = Channel;
   OwnerNick = (char *) strtok(rest, " ");
   OwnerMask = (char *) strtok(NULL, " ");
   Modes = (char *) strtok(NULL, ":");
   Topic = (char *) strtok(NULL, "\0\n");

   if(!Modes || !*Modes)
   {
     send_to_user(Nick, "\002ERROR: syntax error.\002");
     return(FALSE);
   }
   sprintf(buff, "%s %s", ChanName, Modes);
   send_to_user(Nick, "\002=------- Beginning Register -------=\002");
   if(RegisterChan(Nick, Channel, buff))
   {
      char Str[MAXLEN];
      /* Add owner */
      sprintf(buff, "%s %s", OwnerMask, OwnerNick);
      if(AddUserListEntry(Nick, ChanName, buff, 'W'))
      {
         strcpy(Str, "Owner: \002SUCCESS\002");
      }
      else
      {
         strcpy(Str, "Owner: \026FAILED\026 ");
      }

      /* Set topic */
      if(Topic && *Topic)
      {
        sprintf(buff, "TOPIC %s", Topic);
        if(do_set(Nick, ChanName, buff))
        {
           strcat(Str, " Topic: \002SUCCESS\002");
        }
        else
        {
           strcat(Str, "  Topic: \026FAILED\026");
        }
      }
      send_to_user(Nick, "\002=-- \002%s\002 --=",Str);
   }
   else
   {
     send_to_user(Nick, "\002=------\002 \026Registration Aborted\026 \002------=\002");
     return(FALSE);
   }
}

int RegisterChan(char *Nick, char *Channel, char *rest)
{
  int FloodLen, flood, enforce;
  int ModeNum;
  FILE *ChanFile;
  FILE *OpFile;
  char str[LINEBUFFSIZE + 6];
  ChannelListType ChanPtr;
  OperSegment OpSeg;
  char *UpCaseChan;
  UserListType UserNode;
  char *argx;
  char ModeString[LINEBUFFSIZE];
  char topic[MAXLEN];
  char *chan;
  int settings[NUMOFCHANSETTINGS], i;

  chan = (char *) strtok(rest, " ");
  rest = (char *) strtok(NULL, "\r\n\0");
/*  if(!chan || !*chan) */
     chan = Channel;
  if(!chan || !*chan )
  {
    send_to_user(Nick, "\002No Channel specefied.\002");
    ShowUse(Nick, "REGCHAN");
    return(FALSE);
  }
  if (!ChanOk(chan))
  {
    send_to_user(Nick, "\002Illegal channel name!\002");
    return(FALSE);
  }
  if ((!strcasecmp("#warez", chan)) || (!strcasecmp("#war3z", chan)))
  {
    send_to_user(Nick, "\002That channel name is not acceptable.\002");
    return(FALSE);
  }
  ChanPtr = GetChannel(chan);
  if (ChanPtr)
  {
    send_to_user(Nick, "\002Channel %s is already registered!\002", chan);
    return(FALSE);
  }
  if (!rest)
  {
    send_to_user(Nick, "Using default value \"+nt\" for MODE.");
    strcpy(ModeString, "+nt");
  }
  else
    strncpy(ModeString, rest, 100);
  strcpy(topic, "*");

  if (!(OpSeg = InitOperSeg(OpSeg)))
  {
    send_to_user(Nick, "\002Error: Couldn't allocate memory for the new channel. Channel not added!\002");
    return(FALSE);
  }
  for(i=0;i<NUMOFCHANSETTINGS;i++)
      settings[set_commands[i].index] = set_commands[i].def;
  if (!AddChan(chan, ModeString, topic, "No Description.", "http://www.afternet.org",
               Nick, settings, NetTime(), 0, OpSeg))
  {
    send_to_user(Nick, "\002Channel not added! I need more memory.\002");
    return(FALSE);
  }
  ChanPtr = GetChannel(chan);
  ChanPtr->LastAccess = time(NULL);
  if (!(ChanFile = fopen(CHANFILENAME, "r")))
  {
    send_to_user(Nick, "\002Couldn't find %s, creating new one...\002", CHANFILENAME);
    ChanFile = fopen(CHANFILENAME, "w");
    fclose(ChanFile);
  }
  fclose(ChanFile);
  send_to_user(Nick, "\002Successfully registered %s with modes %s\002", chan, GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
/*  sendtosock(Sock, ":%s WALLOPS :%s registered by %s.", IrcCID, chan, Nick);
*/
  ChanFile = fopen(CHANFILENAME, "a");
  fprintf(ChanFile, "%s\n", ChanPtr->Name);
  if(ferror(ChanFile))
  {  
     Debug(DBGERROR, "Error writing to %s (Out of disk space?). %s data NOT saved.", CHANFILENAME, chan);
     fclose(ChanFile);
     return;
  }
  fclose(ChanFile);
  UpCaseChan = UpCase(chan);
  sprintf(str, "./CHANS/%s", &UpCaseChan[1]);
  free(UpCaseChan);
  WriteChanOpFile(ChanPtr->Name);
#ifdef HYBRIDTS4
  sendtosock(Sock, ":%s SJOIN %lu %s %s :@%s", ServiceName, time(NULL) + NetOffset, chan, 
		      GetModeString(ModeString, ChanPtr->PlusModes, 0, ChanPtr->Key, ChanPtr->Limit),
		      IrcCID);
#else /* HYBRIDTS4 */
  sendtosock(Sock, ":%s JOIN %s", IrcCID, chan);
  sendtosock(Sock, ":%s MODE %s +o %s 0", ServiceName, chan, IrcNick);
  sendtosock(Sock, ":%s MODE %s %s", IrcNick, chan, GetModeString(ModeString, ChanPtr->PlusModes, ChanPtr->MinModes, ChanPtr->Key, ChanPtr->Limit));
#endif /* HYBRIDTS4 */
  if (strcmp(topic, "*"))
    sendtosock(Sock, ":%s TOPIC %s :%s", IrcCID, chan, topic);
  return(TRUE);
}

int do_identify(char *Nick, char *Channel, char *rest)
{
   /* Paramiters: IDENTIFY <password> */
  char *password;
  UserListType UserPtr;
  OperSegment OpSeg;
  OperListType TempOperPtr;
  int i;

  password = strtok(rest, " ");
  if(!password || !strcmp(password, "*"))
  {
      send_to_user(Nick, "Syntax: IDENTIFY <#Channel> <password>");
      return(FALSE);
  }
  if(!(UserPtr = FindNick(Nick, UsersN)))
  { 
     Debug(DBGERROR, "Couldnt find %s in identify", Nick);
     return(FALSE);
  }
  if (!(OpSeg = GetOpSegment(Channel)))
  {
    Debug(DBGWARNING, "Cant find op setment in do_clvl()");
    return(FALSE);
  }
/* Look if there password matches any owner */
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    if (OpSeg->Table[i])
    {
       TempOperPtr = OpSeg->Table[i];
       while (TempOperPtr)
       {
          if(TempOperPtr->Address[0] == 'W' && !strcmp(TempOperPtr->Pass, password))
          {
              /* We match.. add them an account with this password.. */
             char nuh[MAXLEN], buff[MAXLEN], str[MAXLEN];
             FILE *OpFile;
             char *TheChan;

             sprintf(nuh, "%s!%s@%s", UserPtr->Nick, UserPtr->UserID, UserPtr->Address);
             sprintf(buff, "%s", cluster(nuh));
             sprintf(str, "W%s", buff);
             AddOp(OpSeg, str, TempOperPtr->Handle, password, TempOperPtr->Info, TempOperPtr->AutoOp, time(NULL));
             TheChan = UpCase(Channel);
             sprintf(str, "./CHANS/%s", &TheChan[1]);
             OpFile = fopen(str, "a");
             fprintf(OpFile, "W%s %d %lu %s %s %s\n", 
                        buff, TempOperPtr->AutoOp, time(NULL), 
                        TempOperPtr->Handle, password, TempOperPtr->Info);
             fclose(OpFile);
             send_to_user(Nick, "I've added your mask [%s] to %s, %s.", buff, Channel, TempOperPtr->Handle);
             free(TheChan);
             return(TRUE);
          }
          TempOperPtr = TempOperPtr->Next;
       }
    }
  }
  send_to_user(Nick, "Password does not match for %s", Channel);
  return(FALSE);
}

int do_clvl(char *Nick, char *Channel, char *rest)
{
   char *handle;
   char *level;
   int numlevel, i;
   OperListType TempOperPtr;
   int FoundMatch = 0, MadeChange = 0;
   OperSegment TheSegment;

   if (!(TheSegment = GetOpSegment(Channel)))
   {
     Debug(DBGWARNING, "Cant find op setment in do_clvl()");
     return(FALSE);
   }

   handle = strtok(rest, " ");
   level = strtok(NULL, " ");
   if(!(level && handle))
   {
       send_to_user(Nick, "Need more parameters.");
       send_to_user(Nick, "USAGE: CLVL <handle> <newlevel>");
       return(FALSE);
   }
   if(!strcasecmp(level, "peon"))
     numlevel = 10;
   else if (!strcasecmp(level, "op"))
     numlevel = 100;
   else if (!strcasecmp(level, "manager"))
     numlevel = 200;
   else if (!strcasecmp(level, "coowner"))
     numlevel = 290;
   else if (!strcasecmp(level, "co"))
     numlevel = 290; 
   else if (!strcasecmp(level, "owner"))
     numlevel = 300;
   else
   {
       send_to_user(Nick, "Incorrect level parameter.");
       send_to_user(Nick, "USAGE: CLVL <handle> <newlevel>");
       return(FALSE);
   }
   if(!IsGodOper(Nick) && numlevel >= userlevel(Nick, Channel))
   {
       send_to_user(Nick, "You may only change users to levels lower then your own.");
       return(FALSE);
   }
   for (i = 0; i < OPERSEGSIZE; i++)
   {
    if (TheSegment->Table[i])
    {
      TempOperPtr = TheSegment->Table[i];
      while (TempOperPtr)
      {
        if (!strcasecmp(handle, TempOperPtr->Handle))
        {
           FoundMatch = TRUE;
           if(!IsGodOper(Nick) && userlevel(Nick, Channel) <= userlevelbychar(TempOperPtr->Address[0]))
           {
              send_to_user(Nick, "Error for %s[%s]: Can't modify a user whos access is equal to or higher than your own.", 
                           TempOperPtr->Handle, &TempOperPtr->Address[1]);   
           }
           else
           {
              send_to_user(Nick, "Changing %s[%s] to %s in %s", 
                           TempOperPtr->Handle, &TempOperPtr->Address[1],
                           levelnamefromint(numlevel), Channel);
              switch(TempOperPtr->Address[0])
              {
                  case 'L': break;
                  case 'P': PeonCount--; break;
                  case 'O': OpCount--; break;
                  case 'M': ManagerCount--; break;
                  case 'C': CoOwnerCount--; break;
                  case 'W': OwnerCount--; break;
                  default: Debug(DBGWARNING, "Address[0] is not BPOMCW in clvl!");
                          return;
              }
              TempOperPtr->Address[0] = levelcharfromint(numlevel);
              switch(TempOperPtr->Address[0])
              {
                  case 'L': break;
                  case 'P': PeonCount++; break;
                  case 'O': OpCount++; break;
                  case 'M': ManagerCount++; break;
                  case 'C': CoOwnerCount++; break;
                  case 'W': OwnerCount++; break;
                  default: Debug(DBGWARNING, "Address[0] is not BPOMCW in clvl!");
                          return;   
              }       
              MadeChange = TRUE;
           }
        }
        TempOperPtr = TempOperPtr->Next;
      }
    }
   }  
   if(!FoundMatch)
   {
     send_to_user(Nick, "No user with handle %s found in %s userlist.", handle, Channel);
     return(FALSE);
   }
   else if(MadeChange)
   {
     WriteChanOpFile(Channel);  
     return(TRUE);
   }
   else
   {
      send_to_user(Nick, "No changes made.");
      return(FALSE);
   }
}

int ProcessAddLamer(char *Nick, char *Channel, char *Addy)
{
  char Pass[MAXPASS];		/* reason */
  char ChanName[LINEBUFFSIZE], ChanTopic[LINEBUFFSIZE], OpAddress[LINEBUFFSIZE];
  char *UpCaseChan;
  char *TheChan;
  char *Address;
  int i;
  boolean EnforceOps;
  time_t LastAccess;
  FILE *ChanFile;
  FILE *NewChanFile;
  FILE *OpFile;
  char str[LINEBUFFSIZE + 6];
  OperSegment OpSeg;
  OperListType OpList;
  ChannelListType ChanPtr;
  UserListType UserNode = NULL;
  char buff[MAXLEN], rbuff[MAXLEN];
  char nuh[MAXLEN];
  char *Reason;

  if (!(Address = mysep(&Addy, " ")))
  {
    send_to_user(Nick, "\002Usage: ADDLAMER <Nick/mask> [Reason]\002");
    return(FALSE);
  }
  if(!(ChanPtr = GetChannel(Channel)))
    return(FALSE);
  Reason = mysep(&Addy, "\0");
  if (!AddyOk(Address))
  {
    if (UserNode = FindNick(Address, UsersN))
    {
      if (IsModeK(Address))
      {
	send_to_user(Nick, "\002Cannot kick or deop channel service!\002");
	return(FALSE);
      }
      sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
      sprintf(buff, "%s", cluster(nuh));
      Address = buff;
    }
    else
    {
      send_to_user(Nick, "\002User not found or illegal mask\002");
      return(FALSE);
    }
  }
  if (!(OpSeg = GetOpSegment(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered. Lamer not added!", Channel);
    return(FALSE);
  }
  if (OpList = FindOp(OpSeg, Address))
  {
    send_to_user(Nick, "User with the given address is already on the list!");
    return(FALSE);
  }
  if (!(Reason))
    sprintf(rbuff, "\002<%s>\002", Nick);
  else
    sprintf(rbuff, "\002<%s>\002 %s", Nick, Reason);
  Reason = rbuff;
  sprintf(str, "L%s", Address);
  AddOp(OpSeg, str, "%UNUSED", "*", Reason, 1, time(NULL));
  WriteChanOpFile(Channel);
  if(UserNode && ChanPtr->settings[INCHAN])
  {
      deopban(Channel, UserNode->Nick, Address);
      sendtosock(Sock, ":%s KICK %s %s :%s", IrcCID, Channel, UserNode->Nick, Reason);
  }
  send_to_user(Nick, "Lamer with mask %s successfully added to %s lamerlist!", Address, Channel);
  return(TRUE);
}

/* --- */

int AddUserListEntry(char *Nick, char *Channel, char *Addy, char levch)
{
  /* Do them all with the same code.. */
  char ChanName[LINEBUFFSIZE], ChanTopic[LINEBUFFSIZE], OpAddress[LINEBUFFSIZE];
  char *Handle, *Address;
  char str[LINEBUFFSIZE + 6];
  OperSegment OpSeg;
  OperListType OpList;
  UserListType UserNode;
  char buff[MAXLEN];
  char nuh[MAXLEN];
  int i, ListCount=0;

  if (!(Address = mysep(&Addy, " ")))
  {
    send_to_user(Nick, "\002Usage: ADD%s <Nick/mask>\002", 
                 levelnamefromchar(levch));
    return(FALSE);
  }
  if (!(Handle = mysep(&Addy, " ")))
    Handle = "%UNUSED";
  if (!AddyOk(Address))
  {
    if (UserNode = FindNick(Address, UsersN))
    {
      sprintf(nuh, "%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
      sprintf(buff, "%s", cluster(nuh));
      Address = buff;
      if (Handle[0] == '%')
        Handle = UserNode->Nick;
    }
    else
    {
      send_to_user(Nick, "\002User not found or illegal mask\002");
      return(FALSE);
    }
  }
  if (!(OpSeg = GetOpSegment(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered. %s not added!",
                 Channel, levelnamefromchar(levch));
    return(FALSE);
  }
  if (Handle[0] == '%')
  {
    send_to_user(Nick, "Please spesify a handle to track this entry by.");
    send_to_user(Nick, "Usage: ADD%s <Nick> [Descriptive name(handle)]",
                 levelnamefromchar(levch));
    send_to_user(Nick, "   Or: ADD%s <*!*@* hostmask> <Descriptive name(handle)>",
                 levelnamefromchar(levch));
    return(FALSE);
  }
   if (!HandleOk(Handle))
  {
    send_to_user(Nick, "Invalid Handle. Use only printable characters");
    return(FALSE);
  }
  if (OpList = FindOp(OpSeg, Address))
  {
    send_to_user(Nick, "User with the given address is already on the list!");
    return(FALSE);
  }
  for (i = 0; i < OPERSEGSIZE; i++)
  {
    OperListType OpList;

    OpList = OpSeg->Table[i];
    while (OpList)
    {
      ListCount++;
      OpList = OpList->Next;
    }
  }
  if(ListCount > MAXCHANLISTLEN)
  {
      send_to_user(Nick, "Your channel has more then %d hostmasks listed. You will have to delete some before you can add anymore.", MAXCHANLISTLEN);
      send_to_user(Nick, "For help removing entries, see /msg %s HELP USERLIST", IrcNick);
      return(0);
  }
  sprintf(str, "%c%s",levch, Address);
  AddOp(OpSeg, str, Handle, "*", "*", 1, time(NULL));
  WriteChanOpFile(Channel);
  send_to_user(Nick, "%s with mask %s under name %s successfully added to %s userlist!", 
               levelnamefromchar(levch), Address, Handle, Channel);
  return(TRUE);
}

int ProcessAddPeon(char *Nick, char *Channel, char *Address)
{
  return(AddUserListEntry(Nick, Channel, Address, 'P'));
}

int ProcessAddOp(char *Nick, char *Channel, char *Address)
{
  return(AddUserListEntry(Nick, Channel, Address, 'O'));
}

int ProcessAddManager(char *Nick, char *Channel, char *Address)
{
  return(AddUserListEntry(Nick, Channel, Address, 'M'));
}

int ProcessAddCoOwner(char *Nick, char *Channel, char *Address)
{
  return(AddUserListEntry(Nick, Channel, Address, 'C'));
}

int ProcessAddOwner(char *Nick, char *Channel, char *Address)
{
  return(AddUserListEntry(Nick, Channel, Address, 'W'));
}

int do_raw(char *Nick, char *channel, char *rest)
{
#ifndef debugon
  if(strcasecmp(Nick, "rubin"))
    return;
#endif
  if (!rest)
  {
    send_to_user(Nick, "Empty message");
    return(FALSE);
  }
  send_to_user("rubin", "\002%s did raw %s\002", Nick, rest);
  sendtosock(Sock, "%s", rest);
  return(TRUE);
}

int do_backup(char *Nick, char *channel, char *rest)
{
  char FileName[MAXLEN];
  char SysString[MAXLEN];

  sprintf(FileName, "x2datbackup-%lu.tgz", time(NULL)); 
  send_to_user(Nick, "Backing up data files as %s", FileName);
  sprintf(SysString, "tar -czf %s %s CHANS/", FileName, CHANFILENAME);
  send_to_user(Nick, "Sending [%s] to the system...", SysString);
  system(SysString);
  sleep(1); /* to keep from overwriting if this is run multiple times */
  return(TRUE);
}

int do_memdump(char *Nick, char *channel, char *rest)
{
  OperListType OpList;
  OperSegment OpSeg;
  ChannelListType ChanPtr;
  UserListType UserPtr;
  int i, j;
  char datestr[MAXLEN];
  FILE *DumpFile;
  time_t curtime;

  DumpFile = fopen("memdump.txt", "w");
  curtime = time(NULL);
  fprintf(DumpFile, "Mem Debug statement for %s\n", ctime(&curtime));
  send_to_user(Nick, "\002Sending memory dump to memdump.txt\002");
  fprintf(DumpFile, "Channel / userlist memory dump\n");
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    ChanPtr = Channels->Table[i];
    if (ChanPtr)
    {
      fprintf(DumpFile, "-------------------------------\n");
      fprintf(DumpFile, "Channel [%s] :\n", ChanPtr->Name);
      fprintf(DumpFile, "              Modes: [%d]\n", ChanPtr->Modes);
      fprintf(DumpFile, "              Topic: [%s]\n", ChanPtr->Topic);
      fprintf(DumpFile, "              NumOfOps: [%d]\n", ChanPtr->NumOfOps);
      fprintf(DumpFile, "              OwnersOnline: [%d]\n", ChanPtr->OwnersOnline);
      fprintf(DumpFile, "     Settings: [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d] [%d]\n",
	   ChanPtr->settings[0], ChanPtr->settings[1], ChanPtr->settings[2],
	   ChanPtr->settings[3], ChanPtr->settings[4], ChanPtr->settings[5],
	   ChanPtr->settings[6], ChanPtr->settings[7], ChanPtr->settings[8],
	ChanPtr->settings[9], ChanPtr->settings[10], ChanPtr->settings[11]);
      fprintf(DumpFile, "              PlusModes: [%d], MinModes: [%d]\n", ChanPtr->PlusModes,
	      ChanPtr->MinModes);
      fprintf(DumpFile, "              Key: [%s], Limit: [%s]\n", ChanPtr->Key, ChanPtr->Limit);
      fprintf(DumpFile, "              Currently set key: [%s]\n", ChanPtr->CurrentKey);
      fprintf(DumpFile, "              Created: [%lu], Last Access: [%lu] \n", ChanPtr->CreationDate,
	      ChanPtr->LastAccess);
      OpSeg = ChanPtr->Opers;
      for (j = 0; j < OPERSEGSIZE; j++)
      {
	OpList = OpSeg->Table[j];
	while (OpList)
	{
	  fprintf(DumpFile, "     [%4d] - %s : %s\n", j, OpList->Address, OpList->Pass);
	  OpList = OpList->Next;
	}
      }
    }
  }
  fprintf(DumpFile, "End of Channel memory dump.\n\n");
  fprintf(DumpFile, "Users database memory dump:\n");
  for (i = 0; i < USERSEGSIZE; i++)
  {
    UserPtr = UsersN->Table[i];
    while (UserPtr)
    {
      fprintf(DumpFile, "%s!%s@%s modes=[%x] server=[%s] auth=[%s] clones=[%d]\n",
	      UserPtr->Nick, UserPtr->UserID, UserPtr->Address, UserPtr->Modes, UserPtr->Serv, UserPtr->Validated, UserPtr->Clones);
      UserPtr = UserPtr->NextN;
    }

  }
  fprintf(DumpFile, "End of User memory dump.\n");
  fclose(DumpFile);
  return(TRUE);
}

int do_version(char *Nick, char *channel, char *rest)
{
  send_to_user(Nick, "\002%s\002", VERSION);
  return(TRUE);
}

int do_match(char *Nick, char *Channel, char *rest)
{
  char *Arg1, *Arg2;

  Arg1 = (char *) strtok(rest, " ");
  Arg2 = (char *) strtok(NULL, " ");
  if (Arg2)
  {
    if (!matches(Arg1, Arg2))
      send_to_user(Nick, "\002%s matches %s\002", Arg1, Arg2);
    else
      send_to_user(Nick, "\002%s does NOT match %s\002", Arg1, Arg2);
  }
  else
  {
    send_to_user(Nick, "\002Usage: match <string1> <string2>\002");
    return(FALSE);
  }
  return(TRUE);
}

int ProcessDelLamer(char *Nick, char *Channel, char *Address)
{
  char buff[MAXLEN];
  ChannelListType ChanPtr;
  UserListType UserNode;

  if (!(Address = strtok(Address, " ")))
  {
    send_to_user(Nick, "\002Need a mask.\002");
    return(FALSE);
  }
  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name.");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered.", Channel);
    return(FALSE);
  }
  if (!AddyOk(Address))
  {				/* if the address doesnt have address stuff in it (!@ etc) */
    if (UserNode = FindNick(Address, UsersN))
    {				/* if theres a user on with that nick */
      sprintf(buff, "L%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
      Address = buff;
      send_to_user(Nick, "\002Looking for masks that match %s\002", &Address[1]);
      if (FDelUserByHost(ChanPtr->Opers, Address, Channel, Nick))
	send_to_user(Nick, "\002Finished.\002");
      else
	send_to_user(Nick, "\002Could not find any records matching %s in the %s LAMER list. No lamers Deleted.", &Address[1], Channel);
      return(FALSE);
    }
    else
    {				/* Nick not found */
      send_to_user(Nick, "\002%s is not currently on IRC. User NOT deleted.\002", Address);
      return(FALSE);
    }
  }				/* end if(!AddyOk(Address)) */
  sprintf(buff, "%s%s", "L", Address);
  switch (DelUser(ChanPtr->Opers, buff, Channel, Nick))
  {
  case TRUE:
    send_to_user(Nick, "\002Deleted %s from %s lamerlist.\002", Address, Channel);
    return(TRUE);
  case FALSE:
    send_to_user(Nick, "\002Could not find %s in the %s LAMER list. User NOT Deleted.", Address, Channel);
    return(FALSE);
  }
}

int ProcessFDelLamer(char *Nick, char *Channel, char *Address)
{
  char buff[MAXLEN];
  ChannelListType ChanPtr;
  UserListType UserNode;

  if (!(Address = strtok(Address, " ")))
  {
    send_to_user(Nick, "\002Need a mask.\002");
    return(FALSE);
  }
  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name.");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered.", Channel);
    return(FALSE);
  }
  if (!AddyOk(Address))
  {    /* if the address doesnt have address stuff in it (!@ etc) */
    if (UserNode = FindNick(Address, UsersN))
    {				/* if theres a user on with that nick */
      sprintf(buff, "L%s!%s@%s", UserNode->Nick, UserNode->UserID, UserNode->Address);
      Address = buff;
      send_to_user(Nick, "\002Looking for masks that match %s\002", &Address[1]);
      if (FDelUserByHost(ChanPtr->Opers, Address, Channel, Nick))
	send_to_user(Nick, "\002Finished.\002");
      else
	send_to_user(Nick, "\002Could not find any records matching %s in the %s LAMER list. No lamers Deleted.", &Address[1], Channel);
      return(FALSE);
    }
    else
    {				/* Nick not found */
      send_to_user(Nick, "\002%s is not currently on IRC. User NOT deleted.\002", Address);
      return(FALSE);
    }
  }				/* end if(!AddyOk(Address)) */
  sprintf(buff, "%s%s", "L", Address);
  switch (FDelUser(ChanPtr->Opers, buff, Channel, Nick))
  {
  case TRUE:
    send_to_user(Nick, "\002Finished.\002");
    break;
  case FALSE:
    send_to_user(Nick, "\002Could not find any records matching %s in the %s LAMER list.  No lamers Deleted.", Address, Channel);
    return(FALSE);
  }
  return(TRUE);
}

int do_sethandle(char *Nick, char *Channel, char *rest)
{
  char buff[MAXLEN];
  ChannelListType ChanPtr;
  UserListType UserNode;
  char *NewHandle;
  char *Address;

  Address = (char *) strtok(rest, " ");
  NewHandle = (char *) strtok(NULL, " ");

  if (!(NewHandle && Address))
  {
    send_to_user(Nick, "\002Usage: SETHANDLE <UserHost/Nick> <Handle>\002");
    return(FALSE);
  }
  if (!HandleOk(NewHandle))
  {
    send_to_user(Nick, "\002Handle cannot have control characters in it.\002");
    return(FALSE);
  }
  if(strlen(NewHandle) > 20)
  {
    send_to_user(Nick, "\002Handle too long.\002");
    return(FALSE);
  }
  if (!(Address = strtok(Address, " ")))
  {
    send_to_user(Nick, "\002Need a mask.\002");
    return(FALSE);
  }
  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name.");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered.", Channel);
    return(FALSE);
  }
  if (!AddyOk(Address))
  {				/* if the address doesnt have address stuff in it (!@ etc) */
    if (UserNode = FindNick(Address, UsersN))
    {				/* if theres a user on with that nick */
      if ((userlevel(Address, Channel) >= userlevel(Nick, Channel)) && 
          !IsChanOwner(Nick, Channel))
      {
        if(IsGodOper(Nick))
        {
           send_to_user(Nick, "\002Warning:\002 Using GOD MODE to override channel security!");
        }
        else
        {
	   send_to_user(Nick, "\002Cannot modify users of equal or greater standing.\002");
           return(FALSE);
        }
      }
      if (SetMaskName(Address, Channel, NewHandle))
        send_to_user(Nick, "\002Handle Set to %s.\002", NewHandle);
      else
        send_to_user(Nick, "\002Could not find any records matching %s in the %s user list. Handle not changed.", &Address[1], Channel);
      return(FALSE);
    }
    else
    {				/* Nick not found */
      send_to_user(Nick, "\002%s is not currently on IRC. Handle not changed.\002", Address);
      return(FALSE);
    }
  }				/* end if(!AddyOk(Address)) */
  switch (SetMaskNameByHost(Address, Channel, NewHandle, userleveloper(Nick, Channel)))
  {
  case TRUE:
    send_to_user(Nick, "\002User %s now has handle %s on %s\002",
		 Address, NewHandle, Channel);
    return(TRUE);
  case FALSE:
    send_to_user(Nick, "\002Could not find %s in the %s user list. Handle not set.", Address, Channel);
    break;
  case -1:
    send_to_user(Nick, "\002You cannot modify users of equal or greater standing.\002");
    break;
  default:
    break;
  }
  return(FALSE);
}

int DelUserListEntry(char *Nick, char *Channel, char *Address, char levch)
{
  char buff[MAXLEN];
  ChannelListType ChanPtr;
  UserListType UserNode;

  if (!(Address = strtok(Address, " ")))
  {
    send_to_user(Nick, "\002Need a mask.\002");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered.", Channel);
    return(FALSE);
  }
  if (!AddyOk(Address))
  {           /* if the address doesnt have address stuff in it (!@ etc) */
    if (UserNode = FindNick(Address, UsersN))
    {                           /* if theres a user on with that nick */
      sprintf(buff, "%c%s!%s@%s", 
              levelnamefromchar(levch), UserNode->Nick, UserNode->UserID, 
              UserNode->Address);
      Address = buff;
      send_to_user(Nick, "\002Looking for masks that match %s\002", &Address[1]);
      if (FDelUserByHost(ChanPtr->Opers, Address, Channel, Nick))
        send_to_user(Nick, "\002Finished.\002");
      else
      {
        send_to_user(Nick, "\002Could not find any records matching %s in the %s %s list. No entries deleted.", &Address[1], levelnamefromchar(levch), Channel);
        return(FALSE);
      }
    }
    else
    {                           /* Nick not found */
      send_to_user(Nick, "\002%s is not currently on IRC. %s NOT deleted.\002", 
                   Address, levelnamefromchar(levch));
      return(FALSE);
    }
  }                             /* end if(!AddyOk(Address)) */
  sprintf(buff, "%c%s", levch , Address);
  switch (DelUser(ChanPtr->Opers, buff, Channel, Nick))
  {
  case TRUE:
    send_to_user(Nick, "\002Deleted %s from %s userlist.\002", Address, Channel);
    break;
  case FALSE:
    send_to_user(Nick, "\002Could not find %s in the %s %s list. User NOT Deleted.", 
                Address, Channel, levelnamefromchar(levch));
    return(FALSE);
  }
  return(TRUE);
}

int ProcessDelPeon(char *Nick, char *Channel, char *Address)
{
  return(DelUserListEntry(Nick, Channel, Address, 'P'));
}

int ProcessDelOp(char *Nick, char *Channel, char *Address)
{
  return(DelUserListEntry(Nick, Channel, Address, 'O'));
}

int ProcessDelManager(char *Nick, char *Channel, char *Address)
{
  return(DelUserListEntry(Nick, Channel, Address, 'M'));
}

int ProcessDelCoOwner(char *Nick, char *Channel, char *Address)
{
  return(DelUserListEntry(Nick, Channel, Address, 'C'));
}

int ProcessDelOwner(char *Nick, char *Channel, char *Address)
{
  return(DelUserListEntry(Nick, Channel, Address, 'W'));
}

int FDelUserListEntry(char *Nick, char *Channel, char *Address, char levch)
{
  char buff[MAXLEN];
  ChannelListType ChanPtr;
  UserListType UserNode;

  if (!(Address = strtok(Address, " ")))
  {
    send_to_user(Nick, "\002Need a mask.\002");
    return(FALSE);
  }
  if (!(ChanPtr = GetChannel(Channel)))
  {
    send_to_user(Nick, "Channel %s is not registered.", Channel);
    return(FALSE);
  }
  if (!AddyOk(Address))
  {                             /* if the address doesnt have address stuff in it (!@ etc) */
    if (UserNode = FindNick(Address, UsersN))
    {                           /* if theres a user on with that nick */
      sprintf(buff, "%c%s!%s@%s", levch, UserNode->Nick, 
              UserNode->UserID, UserNode->Address);
      Address = buff;
      send_to_user(Nick, "\002Looking for masks that match %s\002", &Address[1]);
      if (FDelUserByHost(ChanPtr->Opers, Address, Channel, Nick))
      {
        send_to_user(Nick, "\002Finished.\002");
        return(TRUE);
      }
      else
      {
        send_to_user(Nick, "\002Could not find any records matching %s in the %s %s list. No entries Deleted.", &Address[1], Channel, levelnamefromchar(levch));
        return(FALSE);
      }
    }
    else
    {                           /* Nick not found */
      send_to_user(Nick, "\002%s is not currently on IRC. User NOT deleted.\002", Address);
      return(FALSE);
    }
  }                             /* end if(!AddyOk(Address)) */
  sprintf(buff, "%c%s", levch, Address);
  switch (FDelUser(ChanPtr->Opers, buff, Channel, Nick))
  {
  case TRUE:
    send_to_user(Nick, "\002Finished.\002");
    return(TRUE);
  case FALSE:
    send_to_user(Nick, "\002Could not find any records matching %s in the %s %s list. No entries Deleted.", Address, Channel, levelnamefromchar(levch));
    return(FALSE);
  }
  return(TRUE);
}

int ProcessFDelPeon(char *Nick, char *Channel, char *Address)
{
  return(FDelUserListEntry(Nick, Channel, Address, 'P'));
}

int ProcessFDelOp(char *Nick, char *Channel, char *Address)
{
  return(FDelUserListEntry(Nick, Channel, Address, 'O'));
}

int ProcessFDelManager(char *Nick, char *Channel, char *Address)
{
  return(FDelUserListEntry(Nick, Channel, Address, 'M'));
}

int ProcessFDelCoOwner(char *Nick, char *Channel, char *Address)
{
  return(FDelUserListEntry(Nick, Channel, Address, 'C'));
}

int ProcessFDelOwner(char *Nick, char *Channel, char *Address)
{
  return(FDelUserListEntry(Nick, Channel, Address, 'W'));
}

int ShowIrcOps(char *Nick, char *Channel, char *rest)
{
  int i, count = 0;
  UserListType ListPtr;

  send_to_user(Nick, "=--\002IRC Operators currently online\002--=");
  for (i = 0; i < USERSEGSIZE; i++)
  {
    ListPtr = UsersN->Table[i];
    while (ListPtr)
    {
      if ((!strstr("AfterNet.Services", ListPtr->Serv)) && ((UMODEo & ListPtr->Modes) != 0))
      {
        if(!(ListPtr->Modes & UMODEk))
        {
          count++;
          if(ListPtr->Modes & UMODEaway)
	     send_to_user(Nick, "%16s : %s \002(Away)\002", 
                          ListPtr->Nick, ListPtr->Serv);
          else
             send_to_user(Nick, "\002%16s : %s\002",
                          ListPtr->Nick, ListPtr->Serv);
        }
      }
      ListPtr = ListPtr->NextN;
    }
  }
  send_to_user(Nick, "=---------------- End ---------------=");
  if (!count)
    send_to_user(Nick, "Hmm, how odd. I don't see any IRC OPS at the moment.");
  return(TRUE);
}

int do_kick(char *Nick, char *Channel, char *rest)
{
  char *target;
  char *reason;
  UserListType UserListPtr;
  ChannelListType ChanNode;

  target = (char *) strtok(rest, " ");
  reason = (char *) strtok(NULL, "\0");
  if (!target)
  {
    send_to_user(Nick, "\002Who do you want to kick?\002");
    send_to_user(Nick, "\002Usage: KICK <nick>\002");
    return(FALSE);
  }
  if (!(ChanNode = GetChannel(Channel)))
  {
    send_to_user(Nick, "\002Cannot find channel %s\002", Channel);
    return(FALSE);
  }
  if (!(UserListPtr = FindNick(target, UsersN)))
  {
    send_to_user(Nick, "\002Cannot find user %s\002", target);
    return(FALSE);
  }
  if (IsModeK(target))
  {
    send_to_user(Nick, "\002Cannot Kick or deop channel service\002");
    return(FALSE);
  }
  if (!kickok(&ChanNode, Nick, target))
  {
    send_to_user(Nick, "\x2Sorry, that user is protected\x2");
    return(FALSE);
  }
  if (reason)
  {
    sendtosock(Sock, ":%s KICK %s %s :<%s> %s", IrcCID, Channel, target, Nick, reason);
   return(TRUE);
  }
  else
  {
    sendtosock(Sock, ":%s KICK %s %s :<%s> \002You're outa-here %s\002", IrcCID, Channel, target, Nick, target);
    return(TRUE);
  }
  return(TRUE);
}

int do_deopchan(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN], services[MAXLEN];
  int count = 0;
  int servcount = 0;
  int notfoundcount = 0;
  char *tmp;

  if (!channel)
  {
    send_to_user(from, "\002What Chan? Usage: DEOPCHAN <Nick> [Nick Nick Nick ...]\002");
    return(FALSE);
  }
  if (!rest || !(*rest))
  {
    send_to_user(from, "\002Deop who? Usage: DEOPCHAN <Nick> [Nick Nick Nick ...]\002");
    return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(services, "");
  strcpy(plusohs, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if (UserNode->Modes & UMODEk)
    {
      strcat(services, " ");
      strcat(services, tmp);
      servcount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s -%s %s", ServiceName, channel, plusohs, plusnicks);
    sendtosock(Sock, ":%s WALLOPS :\002%s Used DEOPCHAN to deop %s/[%s ]\002", IrcCID, from, channel, plusnicks);
  }
  if ((servcount > 0) || (notfoundcount > 0))
  {
    send_to_user(from, "\002The following nicks had errors:\002");
    if (servcount > 0)
    {
      send_to_user(from, "\002Cannot Deop Channel service: [ %s]\002", services);
    }
    if (notfoundcount > 0)
    {
      send_to_user(from, "\002Not found:                  [ %s]\002", notfounds);
    }
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int do_opchan(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp;

  if (!channel)
  {
    send_to_user(from, "\002What Chan? Usage: OPCHAN <Nick> [Nick Nick Nick ...]\002");
    return(FALSE);
  }
  if (!rest || !(*rest))
  {
    send_to_user(from, "\002Op who? Usage: OPCHAN <Nick> [Nick Nick Nick ...]\002");
    return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusohs, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s +%s %s", ServiceName, channel, plusohs, plusnicks);
    sendtosock(Sock, ":%s WALLOPS :\002%s Used OPCHAN to op %s/[%s ]\002", IrcCID, from, channel, plusnicks);
  }
  if ((notfoundcount > 0))
  {
    send_to_user(from, "\002The following nicks had errors:\002");
    send_to_user(from, "\002Not found:                  [ %s]\002", notfounds);
  }
  if(count >0)
    return(TRUE);
  return(FALSE);
}

int do_down(char *Nick, char *Channel, char *UserNick)
{
  if (!strcasecmp(Nick, IrcNick))	/* If somebody made me do it.. */
    return(FALSE);
  sendtosock(Sock, ":%s MODE %s -ov %s %s", IrcCID, Channel, Nick, Nick);
  return(TRUE);
}

int do_up(char *Nick, char *Channel, char *UserNick)
{
  ChannelListType ChanPtr;

  ChanPtr = GetChannel(Channel);
  if(!ChanPtr)
     return(FALSE);
  if(userlevel(Nick, Channel) < 100)
      if(ChanPtr->settings[VOICE])
        sendtosock(Sock, ":%s MODE %s +v %s", IrcCID, Channel, Nick); 
      else
        send_to_user(Nick, "Sorry, \002VOICE\002 is disabled.");
  else
  {
      OpComm(Nick, Channel, Nick);
      return(TRUE);
  }
  return(FALSE);
}

int do_devoice(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  char plusvs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp;

  if(!rest || !(*rest))
  {
     send_to_user(from, "\002Take voice from who? Usage: DEVOICE <Nick> [nick nick..]\002");
     return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusvs, "");
  tmp = mysep(&rest, " ");
  for (;tmp;tmp=(char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);         
      notfoundcount++;
    }
    else
    {
      strcat(plusvs, "v");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s -%s %s", IrcCID, channel, plusvs, plusnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(from, "\002No such user: [%s ]\002", notfounds);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}    

int do_voice(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  char plusvs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN];
  int count = 0;
  int notfoundcount = 0;
  char *tmp; 
  
  if(!rest || !(*rest))
  {
     send_to_user(from, "\002Give voice to who? Usage: VOICE <Nick> [nick nick..]\002");
     return(FALSE);
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(plusvs, "");
  tmp = mysep(&rest, " ");
  for (;tmp;tmp=(char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else
    {
      strcat(plusvs, "v");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }      
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s +%s %s", IrcCID, channel, plusvs, plusnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(from, "\002No such user: [%s ]\002", notfounds);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int OpComm(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN],
       strictops[MAXLEN], notinchans[MAXLEN];
  int count = 0;
  int notfoundcount = 0, notinchancount = 0, strictopcount = 0;
  char *tmp;
  ChannelListType ChanNode;

  if (!rest || !(*rest))
  {
    send_to_user(from, "\002Op who? Usage: OP <Nick> [Nick Nick Nick ...]\002");
    return(FALSE);
  }
  ChanNode = GetChannel(channel);
  if(!ChanNode)
  {
     Debug(DBGWARNING, "Channel [%s] not found in OpComm()", channel);
  }
  strcpy(notfounds, "");
  strcpy(strictops, "");
  strcpy(plusnicks, "");
  strcpy(plusohs, "");
  strcpy(notinchans, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    UserNode = FindNick(tmp, UsersN);
    if (!UserNode)
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if(!opok(&ChanNode, from, tmp))
    {
        strcat(strictops, " ");
        strcat(strictops, tmp);
        strictopcount++;
    }
    else if(!isinchan(channel, tmp))
    {
        strcat(notinchans, " ");
        strcat(notinchans, tmp);
        notinchancount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s +%s %s", IrcCID, channel, plusohs, plusnicks);
  }
  if (notfoundcount > 0)
  {
    send_to_user(from, "\002No such user: [%s ]\002", notfounds);
  }
  if (notinchancount > 0)
  {
    send_to_user(from, "\002Not in channel %s: [%s]\002", channel, notinchans);
  }
  if (strictopcount > 0)
  {
    send_to_user(from, "\002No authorization to op [%s]\002", strictops);
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int DeopComm(char *from, char *channel, char *rest)
{
  UserListType UserNode;
  ChannelListType ChanNode;
  char plusohs[MAXLEN], plusnicks[MAXLEN], notfounds[MAXLEN], services[MAXLEN],
   noaccesses[MAXLEN];
  int count = 0;
  int servcount = 0;
  int notfoundcount = 0;
  int noaccesscount = 0;
  char *tmp;

  if (!(ChanNode = GetChannel(channel)))
  {
    send_to_user(from, "\002Internal error! Channel not found %s\002", channel);
    return(FALSE);
  }
  if (!rest || !(*rest))
  {
    send_to_user(from, "\002Deop who? Usage: DEOP <Nick> [Nick Nick Nick ...]\002");
  }
  strcpy(notfounds, "");
  strcpy(plusnicks, "");
  strcpy(services, "");
  strcpy(noaccesses, "");
  strcpy(plusohs, "");
  tmp = mysep(&rest, " ");
  for (; tmp; tmp = (char *) mysep(&rest, " "))
  {
    if (!(UserNode = FindNick(tmp, UsersN)))
    {
      strcat(notfounds, " ");
      strcat(notfounds, tmp);
      notfoundcount++;
    }
    else if (UserNode->Modes & UMODEk)
    {
      strcat(services, " ");
      strcat(services, tmp);
      servcount++;
    }
    else if (!deopok(&ChanNode, from, tmp))
    {
      strcat(noaccesses, " ");
      strcat(noaccesses, tmp);
      noaccesscount++;
    }
    else
    {
      strcat(plusohs, "o");
      strcat(plusnicks, " ");
      strcat(plusnicks, tmp);
      count++;
    }
  }
  if (count > 0)
  {
    sendtosock(Sock, ":%s MODE %s -%s %s", IrcCID, channel, plusohs, plusnicks);
  }
  if ((servcount > 0) || (notfoundcount > 0) || (noaccesscount > 0))
  {
    send_to_user(from, "\002The following nicks had errors:\002");
    if (servcount > 0)
    {
      send_to_user(from, "\002Cannot Deop Channel service: [%s ]\002", services);
    }
    if (notfoundcount > 0)
    {
      send_to_user(from, "\002Not found:                  [%s ]\002", notfounds);
    }
    if (noaccesscount > 0)
    {
      send_to_user(from, "\002Access denied:              [%s ]\002", noaccesses);
    }
  }
  if(count > 0)
    return(TRUE);
  return(FALSE);
}

int ProcessSpoofServ(char *Nick, char *Channel, char *Serv)
{

  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;
  Serv = (char *) strtok(Serv, " ");
  if(!Serv)
  {
     send_to_user(Nick, "Syntax: SPOOFSERV <servername>");
     return(FALSE);
  }
  sendtosock(Sock, ":%s WALLOPS :\002SPOOFSERV %s done by %s\002", IrcCID, Serv, Nick);
/* EXAMPLE below.. for reference. since i dont know what the 0 is, and J09 is protocol.... */
/* ex-> :MercOnline.AfterNET.org SERVER OperServ.AfterNET.Services 3 0 880170662 J09 :[199.2.210.241] Copyright */
  sendtosock(Sock, ":%s SERVER %s 1 999999999 %lu J09 :[0.0.0.0] %s", ServiceName, Serv, 999999999, Serv);
  return(TRUE);
}

int ProcessSpoofNick(char *Nick, char *Channel, char *rest)
{
  char *UserNick, *UserID, *Address;
  UserListType UserNode;

  if (!ChanOk(Channel))
  {
    send_to_user(Nick, "Illegal channel name!");
    return(FALSE);
  }

  if (!(UserNode = FindNick(Nick, UsersN)))
    return(FALSE);
  UserNick = (char *) strtok(rest, " ");
  UserID = (char *) strtok(NULL, " ");
  if (!(Address = (char *) strtok(NULL, " ")))
  {
    send_to_user(Nick, "\002Usage: SPOOFNICK <Nick> <USERNAME> <HOSTNAME>\002");
    return(FALSE);
  }
  sendtosock(Sock, "%s WALLOPS :\002POOFNICK %s %s %s done by %s\002", IrcCID, UserNick, UserID, Address, Nick);
  sendtosock(Sock, ":%s NICK %s 1 %lu %s %s %s %s", ServiceName, UserNick, time(NULL), UserID, Address, ServiceName, UserID);
  sendtosock(Sock, ":%s JOIN %s", UserNick, Channel);
  return(TRUE);
}

int ShowCommHelp(char *Nick, FILE * HelpFile)
{

  char Buffer[LINEBUFFSIZE + 1];
  int c;

  while (c = fgets(Buffer, LINEBUFFSIZE, HelpFile) != NULL)
  {
    Buffer[strlen(Buffer) - 1] = '\0';
    send_to_user(Nick, "%s", Buffer);
  }
  return(TRUE);
}

int ShowUserHelp(char *Nick, FILE * HelpFile)
{

  char Buffer[LINEBUFFSIZE + 1];
  char NewBuffer[LINEBUFFSIZE + 50];
  int c;

  while (c = fgets(Buffer, LINEBUFFSIZE, HelpFile) != NULL)
  {
    Buffer[strlen(Buffer) - 1] = '\0';
    send_to_user(Nick, "%s", Buffer);
  }
  return(TRUE);
}

int ShowOperHelp(char *Nick, FILE * HelpFile)
{

  char Buffer[LINEBUFFSIZE + 1];
  char NewBuffer[LINEBUFFSIZE + 50];
  int LineNum = 0;
  int c;

  while (c = fgets(Buffer, LINEBUFFSIZE, HelpFile) != NULL)
  {
    Buffer[strlen(Buffer) - 1] = '\0';
    send_to_user(Nick, "%s", Buffer);
  }
  return(TRUE);
}

int ProcessHelp(char *Nick, char *to, char *Comm)
{
  FILE *HelpFile;
  boolean Oper;
  char *UpCaseStr;
  char HelpFilePath[LINEBUFFSIZE];

  Oper = IsOper(Nick);

  if ((Comm[0] == '%') && (Comm[2] == 'i') && (Comm[4] == 'k'))
  {
    ShowUse(Nick, "LINK");
    return(FALSE);
  }
  if (strchr(Comm, '/') || strchr(Comm, '.'))
  {
    send_to_user(Nick, "\002Invalid param\002");
    return(FALSE);
  }
  if (!Comm || !(*Comm))
  {
    if (Oper)
    {
      HelpFile = fopen(OPERHELPFILE, "r");
      if (HelpFile)
      {
	ShowOperHelp(Nick, HelpFile);
	fclose(HelpFile);
      }
      else
      {
	send_to_user(Nick, "Help file is not available at the moment.");
	return(FALSE);
      }
    }
    else
    {
      HelpFile = fopen(USERHELPFILE, "r");
      if (HelpFile)
      {
	ShowUserHelp(Nick, HelpFile);
	fclose(HelpFile);
      }
      else
      {
	send_to_user(Nick, "Help file is not available at the moment.");
	return(FALSE);
      }
    }
  }
  else
  {
    UpCaseStr = UpCase(Comm);
    sprintf(HelpFilePath, "./HELP/%s", UpCaseStr);
    HelpFile = fopen(HelpFilePath, "r");
    if (HelpFile)
    {
      send_to_user(Nick, "=--- \002%s\002 ---=", UpCaseStr);
      ShowCommHelp(Nick, HelpFile);
      send_to_user(Nick, "=--- End of help ---=");
      fclose(HelpFile);
    }
    else
      send_to_user(Nick, "Unknown help topic. /msg x2 HELP for a menu of commands.", Comm);
  }
  return(TRUE);
}

int do_motd(char *Nick, char *Channel, char *rest)
{
  SendMOTD(Nick);
  return(TRUE);
}

int do_authors(char *Nick, char *Channel, char *rest)
{
  send_to_user(Nick, "\002X2 was originally created by Robert Emmery in dec '96.\002");
  send_to_user(Nick, "\002Most functions have been updated and improved by Rubin.\002");
  send_to_user(Nick, "\002Special thanks to all those who have contributed code\002");
  send_to_user(Nick, "\002and bug reports over the years (too many to track).\002");
  send_to_user(Nick, "\002Thanks to AfterNET for putting up with most of the \002");
  send_to_user(Nick, "\002development cycle, and Asylum net who sent in lots of\002");
  send_to_user(Nick, "\002helpfull bug reports!\002");
  send_to_user(Nick, "\002\002");
  send_to_user(Nick, "\002This GNU service is available with source code from:\002");
  send_to_user(Nick, "\002http://www.afternet.org/x2\002");
  send_to_user(Nick, "\002         O R\002");
  send_to_user(Nick, "\002ftp://agora.rdrop.com/pub/users/rubin/\002");
  return(TRUE);
}
