/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <signal.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "config.h"
#include "init.h"
#include "chan.h"
#include "user.h"
#include "parse.h"
#include "misc.h"
#include "sock.h"
#include "server.h"

boolean DaemonStarted;
boolean AUTOROUTING, AUTOCONNECT, ENFORCEOPS, PINGOUTCONNECT, GATECHECK,
        FLOODCONTROL, NETLINK, NETTALK, PUBLIC, NETBURSTING;
unsigned char CLONERES, MAXCLONES, MAXCHAN, DEBUG;
char ServiceName[MAXLEN],
     Uplink[MAXLEN], ServAddress[16], Password[20], 
     IrcNick[MAXNICKLEN], 
#ifdef HYBRIDTS4
     IrcCID[MAXCIDLEN], 
#endif
     CommandMark, HomeDir[MAXLEN];

u_short Port, SERVERPORT;
int Sock, NewSock;
struct sockaddr_in ServAddr, CliAddr;
fd_set fdvar, wfdvar;
int MaxSock;

boolean SockConnected;
boolean ServerConnected;
boolean SentBurst;
boolean Authenticated;
boolean RoutsOk;
boolean GotPong;
boolean NOFORK = FALSE;

time_t LinkTimeStamp, ServerTimeStamp, ConnectTime, DiscTime, StartTime,
       PingStamp, PongStamp, FloodStamp;
long int NetOffset;
time_t bantimer;

char LastNickAdded[MAXNICKLEN];
char LastRequestNick[MAXNICKLEN];

ChanSegment Channels;
GlineListType *Glines;
GlineListType *SuperGlines;
ServSegment Servers, Routs;
UserSegment UsersN, UsersA;
#ifdef HYBRIDTS4
UserSegment UsersC;
#endif /* HYBRIDTS4 */

int ChannelCount, ServerCount, RoutCount, UserCount, OperCount,
    PeonCount, OpCount, ManagerCount, CoOwnerCount, OwnerCount, LamerCount;

NetMessageType NetMessage;
UpLinkList *CurUpLink;
UpLinkList *UpLinks = NULL;
extern set_list set_commands[NUMOFCHANSETTINGS];

/* Start of the code section */

void StartDaemon(int Sock)
{
  register int childpid, fd;
  FILE *ProcFile;

#ifndef debugon
  if(!NOFORK)
  {
   childpid = fork();
   if (childpid < 0)
      exit(1);
   else if (childpid > 0)
      exit(0);
   close(0);			/* stdin */
   close(1);			/* stdout */
   close(2);			/* stderr */
   setsid();
  }
#endif
  if(ProcFile = fopen(ProcFilePath, "w"))
     fprintf(ProcFile, "%d", getpid());
  fclose(ProcFile);
  DaemonStarted = TRUE;
}

void AddUpLinkServer(char *ServIp, char *ServPass, char *ServName, int port)
{
   UpLinkList *UpLinkPtr;

   if(!(UpLinkPtr = malloc(sizeof(*UpLinkPtr))))
   {
      Debug(DBGSEVERE, "Out of memory adding server to uplink list");
      return;
   }
   strcpy(UpLinkPtr->ServName, ServName);
   strcpy(UpLinkPtr->ServIp, ServIp);
   strcpy(UpLinkPtr->ServPass, ServPass);
   UpLinkPtr->Port = port;
   UpLinkPtr->Next = UpLinks;
   UpLinks = UpLinkPtr;
}

int ReadConfig(char *Filename)
{

  FILE *ConfFile;
  char StrBuffr[MAXCONFSTRBUFF];
  char ch;
  int PassLen;
  UpLinks = NULL;

  if ((ConfFile = fopen(Filename, "r")) == NULL)
  {
    printf("Can't open file \"%s\"!\n", Filename);
    return (0);
  }
  while (fscanf(ConfFile, "%s", StrBuffr) != EOF)
  {
    if (StrBuffr[0] == '#')
    {
      while ((ch = fgetc(ConfFile)) != '\n')
	;
      continue;
    }
    if (!strcmp(StrBuffr, "SERVER"))
    {
      char *ServIp, *ServPass, *ServName, *ServPort;
      int IntPort = 0;

      fscanf(ConfFile, "%s", StrBuffr);
      ServIp = (char *) strtok(StrBuffr, ":");
      ServPass = (char *) strtok(NULL, ":");
      ServName = (char *) strtok(NULL, ":");
      ServPort = (char *) strtok(NULL, ":");
      if(!ServPort)
      {
         printf("Incorrect SERVER line syntax\n");
         exit(1);
      }
      if((IntPort = atoi(ServPort)) >= 0)
      {
        AddUpLinkServer(ServIp, ServPass, ServName, IntPort);
        printf("Adding Uplink server %s ip %s on port %s\n", ServName, ServIp, ServPort);
      }
      else
      {
        printf("Invalid port (%d) in SERVER line.  \n", IntPort);
        exit(1);
      }
    }
    else if (!strcmp(StrBuffr, "PORT"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      Port = atol(StrBuffr);
    }
    else if (!strcmp(StrBuffr, "SERVICE"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(ServiceName, StrBuffr);
    }
    else if (!strcmp(StrBuffr, "PASSWORD"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(Password, StrBuffr);
    }
    else if (!strcmp(StrBuffr, "IRCNICK"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(IrcNick, StrBuffr);
      printf("Nick: %s\n", IrcNick);
    }
    else if (!strcmp(StrBuffr, "HOME"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      strcpy(HomeDir, StrBuffr);
      printf("Home Directory: %s\n", HomeDir);
    }
    else if (!strcmp(StrBuffr, "PINGOUTCONNECT"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      PINGOUTCONNECT = atoi(StrBuffr);
      printf("AutoConnecting pingouts: %s\n", GetSetting(PINGOUTCONNECT));
    }
    else if (!strcmp(StrBuffr, "GATECHECK"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      GATECHECK = atoi(StrBuffr);
      printf("AutoConnecting pingouts: %s\n", GetSetting(GATECHECK));
    }
    else if (!strcmp(StrBuffr, "AUTOROUTING"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      AUTOROUTING = atoi(StrBuffr);
      printf("Automated Routing: %s\n", GetSetting(AUTOROUTING));
    }
    else if (!strcmp(StrBuffr, "AUTOCONNECT"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      AUTOCONNECT = atoi(StrBuffr);
      printf("Automated Reconnection: %s\n", GetSetting(AUTOCONNECT));
    }
    else if (!strcmp(StrBuffr, "ENFORCEOPS"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      ENFORCEOPS = atoi(StrBuffr);
      printf("Enforcing Ops: %s\n", GetSetting(ENFORCEOPS));
    }
    else if (!strcmp(StrBuffr, "MAXCLONES"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      MAXCLONES = atoi(StrBuffr);
    }
    else if (!strcmp(StrBuffr, "CLONERES"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      CLONERES = atoi(StrBuffr);
      printf("Clone Resolution Method: %s\n", GetCloneResMet(CLONERES));
      if (CLONERES)
	printf("Clone Limit: %d\n", MAXCLONES);
    }
    else if (!strcmp(StrBuffr, "LISTEN"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      SERVERPORT = atoi(StrBuffr);
      printf("I will listen on port %d if not connected\n", SERVERPORT);
    }
    else if (!strcmp(StrBuffr, "CMDCHAR"))
    {
      fscanf(ConfFile, "%s", StrBuffr);
      CommandMark = *StrBuffr;
      printf("My command charactor is %c\n", CommandMark);
    }
    else
    {
      printf("unknown config setting \"%s\"\n", StrBuffr);
      fclose(ConfFile);
      return (0);
    }
  }
  fclose(ConfFile);
  return (1);
}

int ReadGlineFile(char type, char *filename)
{
  FILE *TheFile;
  GlineListType *gptr;

  char address[MAXLEN], reason[MAXLEN];
  time_t gtime;
  int s, i;
  char buff[MAXLEN];

  if (!(TheFile = fopen(filename, "r")))
  {
    Debug(DBGWARNING, "No gline file found.");
    return (1);
  }
  while ((s = fscanf(TheFile, "%s%lu %[^\n]\n", address, &gtime, reason)) != EOF)
  {
    if (s != 3)
    {
      Debug(DBGWARNING, "Corruption in gline file!");
      return (2);
    }
    if(type == 'G')
      ProcessAddGline(address, gtime - NetTime(), reason);
    else if(type == 'S')
      ProcessAddSGline(address, gtime - NetTime(), reason);
  }
  fclose(TheFile);
  return(0);
}

int ReadChanFile()
{
  FILE *TheFile;
  FILE *OpFile;
  char linebuff[LINEBUFFSIZE], *key;
  char *ptrbuff, *lb;
  char ChanName[LINEBUFFSIZE], ChanTopic[LINEBUFFSIZE], ModeString[LINEBUFFSIZE], *OpAddress;
  char URL[LINEBUFFSIZE], Description[LINEBUFFSIZE], Reggedby[MAXNICKLEN+1];
  int i, s, line = 0, line2, dupcount, fileversion=0;
  int UserAutoOp;
  int settings[NUMOFCHANSETTINGS];
  boolean EnforceOps;
  time_t LastAccess, CreationDate;
  time_t LastSeen;
  OperSegment NewOperSeg;
  char ChanFileName[LINEBUFFSIZE + 6];
  char *UpCaseChan;
  char *Pass;
  char *Info;
  char FullChanName[LINEBUFFSIZE];
  char *Handle;

  Debug(DBGNOTICE, "Processing chan.dat\n");
  if (!(TheFile = fopen(CHANFILENAME, "r")))
  {
    printf("No chan.dat.. creating..");
    TheFile = fopen(CHANFILENAME, "w");
    fclose(TheFile);
    return (1);
  }
  while ((s = fscanf(TheFile, "%s\n", ChanName)) != EOF)
  {
    line++;
    if (s != 1)
    {
      printf("Parameter error in chan.dat line %d. only %d nybles read.\n", line, s);
      return (0);
    }
    if (!(NewOperSeg = InitOperSeg(NewOperSeg)))
    {
      printf("Error initializing oper seg in %s, line %d\n", ChanName, line);
      return (0);
    }
    UpCaseChan = UpCase(ChanName);
    sprintf(ChanFileName, "./CHANS/%s", &UpCaseChan[1]);
    free(UpCaseChan);
    if (!(OpFile = fopen(ChanFileName, "r")))
    {
      printf("Error opening OpFile for %s. chan.dat line %d\n", ChanName, line);
      return (0);
    }
    line2 = 0;
    dupcount = 0;
    strcpy(ChanTopic, "*");
    strcpy(ModeString, "");
    strcpy(Description, "*");
    strcpy(URL, "*");
    strcpy(Reggedby, "UNKNOWN");
    while (!feof(OpFile))
    {
      i++;
      s = fscanf(OpFile, "%[^\n]\n", linebuff);
      line2++;
      if (s == EOF)
	break;
      if (s != 1)
      {
         printf("Corruption in CHANS/%s, line %d.", ChanName, line2);
         return(0);
      }
      lb = linebuff;
      key = (char *) mysep(&lb, " ");
      if(!key || !*key)
      {
         printf("Cant really happen..\n");
         return(0);
      }
      switch(key[0])
      {
         case 'V': /* VERSION */
                ptrbuff = (char *) mysep(&lb, " ");
		if(!ptrbuff)
                {
                    printf("Invalid VERSION syntax in CHANS/%s line %d\n", ChanName, line2);
                    return(0);
                }
                fileversion = strtol(ptrbuff, NULL, 10);
                if(fileversion > 1)
                {
                   printf("Warning! Reading CHANS/%s which is fileversion %d\n", ChanName, fileversion);
                }
                continue;
         case 'C': /* Channel */
                ptrbuff = (char *) mysep(&lb, "\n");
		safecopy(FullChanName, ptrbuff, MAXLEN);
                if(strcasecmp(FullChanName, ChanName))
                {
                  printf("Warning! CHANS/%s claims to be %s. Quitting..\n", ChanName, ptrbuff);
                  return(0);
                }
                continue;
         case 'T': /* Topic */
                ptrbuff = (char *)mysep(&lb, "\n"); 
		safecopy(ChanTopic, ptrbuff, MAXTOPICLEN);
                continue;
         case 'M': /* ModeString */
                ptrbuff = (char *)mysep(&lb, "\n");
		safecopy(ModeString, ptrbuff, MAXLEN);
                continue;
         case 'D': /* Description */
		ptrbuff = (char *)mysep(&lb, "\n");
		safecopy(Description, ptrbuff, MAXLEN);
                continue;
         case 'S': /* Settings */
                for(i=0;ptrbuff = (char *) mysep(&lb, " ");i++)
                {
                   if(*ptrbuff)
                     settings[i] = strtol(ptrbuff, NULL, 10);
                }
		if(i < NUMOFCHANSETTINGS)
                {    /* If settings have been added,
			set them to the defined default
			values */
                   for(;i<NUMOFCHANSETTINGS;i++)
                      settings[set_commands[i].index] = set_commands[i].def;
                }
                continue;
         case 'L': /* LastVisted */
                ptrbuff = (char *)mysep(&lb, " ");
                LastAccess = strtol(ptrbuff, NULL, 10);
                continue;
         case 'R': /* Regged by */
                ptrbuff = (char *)mysep(&lb, " ");
		safecopy(Reggedby, ptrbuff, MAXNICKLEN);
                ptrbuff = (char *)mysep(&lb, " ");
                CreationDate = strtol(ptrbuff, NULL, 10);
                continue;
         case 'W': /* Website (URL) */
		ptrbuff = (char *)mysep(&lb, " ");
		safecopy(URL, ptrbuff, MAXLEN);
		continue;
         case 'U': /* USER */
                OpAddress = (char *)mysep(&lb, " ");
                ptrbuff = (char *)mysep(&lb, " ");
                UserAutoOp = strtol(ptrbuff, NULL, 10);
                ptrbuff = (char *)mysep(&lb, " ");
                LastSeen = strtol(ptrbuff, NULL, 10);
                Handle = (char *)mysep(&lb, " ");
                Pass = (char *)mysep(&lb, " ");
                Info = (char *)mysep(&lb, "\n");
                /* ******************************************** */
		if(!Info)
                {
                    printf("Invalid USER syntax in CHANS/%s, line %d\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Pass) > MAXPASS)
                {
                    printf("Error in CHANS/%s, line %d. Password too long.\n", ChanName, line2);
                    return(0);
                }
                if(strlen(Handle) > 20)
                {
                    printf("Error in CHANS/%s, line %d. Handle too long.\n", ChanName, line2);
                    return(0); 
                }
                if(strlen(Info) > 200)
                {
                    printf("Error in CHANS/%s, line %d. Infoline too long.\n", ChanName, line2);
                    return(0);
                }    
                if (!AddyOk(&OpAddress[1]))
                {
                    printf("Invalid OpAddress in CHANS/%s, line %d. [%s]\n", ChanName, line2, OpAddress);
                    return (0);
                }
                if (FindOp(NewOperSeg, &OpAddress[1]))
                {
                   printf("Fixing duplicate user in %s line %d, [%s].\n", ChanName, line2, OpAddress);
	           dupcount++;	   /* we'll actualy track and repair this once were done with the file. */
                }
                else if (!(AddOp(NewOperSeg, OpAddress, Handle, Pass, Info, UserAutoOp, LastSeen)))
                {
	           printf("Error adding op in %s line %d, [%s]\n", ChanName, line2, OpAddress);
	           return (0);
                }
		continue;
         default: printf("Unknown key %s in CHANS/%s\n", key, ChanName);
                  return(0);
      }
    }
    fclose(OpFile);
    /*NumOfOps = i; */
    if (!AddChan(ChanName, ModeString, ChanTopic,
                 Description, URL, Reggedby,
		 settings, CreationDate, LastAccess, NewOperSeg))
    {
      printf("Error with AddChan in %s line %d\n", ChanName, line2);
      return (0);
    }
    if (dupcount > 0)
    {
      sprintf(FullChanName, "%s", ChanName);
      printf("Writing %s userlist...\n", FullChanName);
      WriteChanOpFile(FullChanName);
    }
  }
  fclose(TheFile);
  Debug(DBGNOTICE, ":Finished reading channel data files\n");
  return (1);
}

int ReadRoutFile(char *RoutFileName)
{

  FILE *TheFile;
  char ServName[LINEBUFFSIZE];
  char ServLink[LINEBUFFSIZE];

  if (!(TheFile = fopen(RoutFileName, "r")))
    return (2);

  while ((fscanf(TheFile, "%s%s", ServName, ServLink)) != EOF)
  {
    if (!AddRout(ServName, ServLink))
    {
       fclose(TheFile);
      return (0);
    }
  }
  fclose(TheFile);
  return (1);
}

main(int argc, char *argv[])
{
  int i, c, s;
  int logfile;
  int FirstLoop = 1;
  char *cfgfile;
  char tmp[MAXLEN], buf[MAXLEN];
  NetLinkListType TempNetLinkPtr;
  FILE *ProcFile;
  static struct timeval polltime;
  time_t TopicTimer_4, TopicTimer_3, TopicTimer_2, TopicTimer_1, TopicTimer_0;
  struct stat StatStruct;
  char SocketBuffer[MAXRECVLEN+10];

  if (argc < 2)
  {
    cfgfile = DefaultConfigFile;
  }
  else if (argc > 2) 
  {
     printf("try -h for help\n\n");
     exit(1);
  }
  if(argc == 2)
  {
    if( !strcmp(argv[1], "-h"))
    {
       printf("%s\n", VERSION);
       printf("By Rubin@afternet.org\n");
       printf("\nUsage: %s [-h|] [-r]\n\n", argv[0]);
       printf("-n = Don't fork\n");
       printf("-h = this help\n");
       exit(0);
    }
    else if (!strcmp(argv[1], "-n"))
    {
       NOFORK = TRUE;
    }
  }
  if (!ReadConfig(cfgfile))
  {
    printf(":configuration file error! cant open %s. Aborting.\n", cfgfile);
    exit(1);
  }
  if (chdir(HomeDir) < 0)
  {
    printf(":Couldn't cd to %s\nAborting!\n", HomeDir);
    exit(1);
  }
  if (!InitGlobals())
  {
    fprintf(stderr, ":initialization error! I need more memory! -- Aborting..\n");
    exit(1);
  }
  if (!ReadChanFile())
  {
    printf(":error encountered while reading channel file! - possibly missing a file in CHANS directory.\n");
    exit(1);
  }
  if (!ReadRoutFile(ROUTFILENAME))
  {
    printf(":error encuontered while reading routing file!\n");
    KillChanList();
    exit(1);
  }
  if (!(s = ReadGlineFile('G', GLINEFILENAME)))
  {
    if(s == 1)
    {
       printf(":gline.dat not found.");
    }
    else if(s == 2)
    {
      printf(":error encountered while reading gline data file!\n");
      printf(":Continuing anyway..\n");
    }
  }
  if (!(s = ReadGlineFile('S', SUPERGLINEFILENAME)))
  {
    if(s == 1)
    {  
       printf(":supergline.dat not found.");
    }
    else if(s == 2)
    { 
      printf(":error encountered while reading supergline data file!\n");
      printf(":Continuing anyway..\n");
    }                                                           
  } 
  stat(MESSAGEFILENAME, &StatStruct);
  NetMessage.Timestamp = StatStruct.st_mtime;
  Debug(DBGINFO, "NetMessage file %s has timestamp %lu", MESSAGEFILENAME, NetMessage.Timestamp);
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);

  SockConnected = FALSE;
  ServerConnected = FALSE;
  DaemonStarted = FALSE;
  SentBurst = FALSE;
  Authenticated = FALSE;

  /* Prime us so we start with the first server this round.. 
   * This is kinda kludgy but makes the clode below have to do less tricks 
   * each loop..  This way it thinks it just had a bad connection (ConnectTime is less then 30 seconds ago)
   * and it will cycle the server and connect.. */
  if(!UpLinks)
  {
     printf("No servers defined. Edit x2.conf.\n");
     exit(1);
  }
  for(CurUpLink = UpLinks;CurUpLink->Next;CurUpLink = CurUpLink->Next);

  /* Initalize our timer system. crude but does the job */
  StartTime = bantimer = TopicTimer_4 = TopicTimer_3 = TopicTimer_2 = TopicTimer_1 = TopicTimer_0 = time(NULL);

  logfile = open(AccountingLog, O_WRONLY | O_CREAT | O_APPEND);
  if (logfile != -1)
  {
    strcpy(tmp, ctime(&StartTime));
    tmp[strlen(tmp) - 1] = '\0';
    sprintf(buf, ":%s:%s Started\n", tmp, IrcNick);
    write(logfile, buf, strlen(buf));
    close(logfile);
  }
  printf("-=[*](%s Is Engaged)[*]=-\n", IrcNick);
  StartDaemon(Sock);
  for (;;)   /* Main loop */
  {
    if (!SockConnected)
    {
      int trycount;

      DiscTime = time(NULL);
      if(!FirstLoop)
        sleep(7); /* Dont want to flood us if theres an error.. */
      else
        FirstLoop = FALSE;
      close(Sock);
      trycount = 0;
      Debug(DBGINFO, "Starting connection phase.");
      ServerConnected = FALSE;
      NETBURSTING = TRUE;
      SentBurst = FALSE;
      Authenticated = FALSE;
      KillServerList();
      KillUserList();
      OperCount = 0;
      if(AUTOCONNECT)		/* If we should try to connect */
      {
        if(time(NULL) - ConnectTime < 30) /* If we get dumped in under 30 seconds, try a new server */
        {
           CurUpLink = CurUpLink->Next;
           if(!CurUpLink)
              CurUpLink = UpLinks;
        }
        while(!(TryToConnect(CurUpLink)))
        {
            CurUpLink = CurUpLink->Next;
            if(!CurUpLink)
            {
               CurUpLink = UpLinks;
               if(trycount++ > 2)
                 break; /* Try the loop twice, then just listen.. */
            }
        }
        if(SockConnected)
        {
          Authenticate(CurUpLink);
          Authenticated = TRUE;
          SentBurst = TRUE;
        }
        else
	{
	  Debug(DBGINFO, "Unable to obtain connection. Listning on port %d", SERVERPORT);
	  Sock = my_init_server();
	}
      }				/* End if autoconnect */
      else
      {
	/* Stop and wait for a connection. */
        Debug(DBGINFO, "Listning for connection on port %d\n", SERVERPORT);
	Sock = my_init_server();
      }
    }
    if(!SockConnected)
       continue;
    SelectSockets();		/* Setup s elect() for our socket(s) */
    polltime.tv_sec = (unsigned long) PINGFREQ;
    polltime.tv_usec = (unsigned long) 1;
    c = select(MaxSock + 1, &fdvar, &wfdvar, NULL, &polltime);
      if(c == -1)
      {
#ifdef debugon
perror("SELECT");
#endif
          //Debug(DBGSEVERE, "select returned -1. Errno:%d", errno);
          //exit(1);  profiling causes "Interrupted System Call" here.. 
      }
      if(c > 0) /* Incomming data */
      {
        if(FD_ISSET(Sock, &fdvar) )
        {
           if(ReadSocket(Sock, SocketBuffer) == -1)
           {
              Debug(DBGWARNING, "Read from socket returned -1. errno: %d", errno);
              SockConnected = FALSE;
              continue;
           }
           else
              PongStamp = time(NULL);
        }
      }
      else
      {
          /* These things happen only if select times out */
          if(c == 0)
          {
             Debug(DBGINFO, "select() timeout.");
          }
          if (((time(NULL) - PongStamp) > PINGFREQ) && (GotPong))
          {
            sendtosock(Sock, "PING :%s", Uplink);
            GotPong = false;
            PingStamp = time(NULL);
          }

      }
    /* These also check for timeout, so call them
     * even if select just timesout */
      GateCheck_CheckSend(wfdvar);
      GateCheck_CheckRead(fdvar);
    /* Handle time stuff here: */
      if (ServerConnected)
      {
/*
*	if (((time(NULL) - PongStamp) > PINGFREQ) && (GotPong))
*	{
*	  sendtosock(Sock, "PING :%s", Uplink);
*	  GotPong = false;
*	  PingStamp = time(NULL);
*	}
*/
	if (((time(NULL) - PingStamp) > PINGTIMEOUT) && (!GotPong))
	{
	  SockConnected = false;
	  continue;
	}
	if ((time(NULL) - bantimer) >= 120)
	{			/* Once every 120 seconds check for expired bans.. */
	  bantimer = time(NULL);
	  CheckBanTimeouts();
	  CheckGlineTimeouts();
          CheckSuperGlineTimeouts();
	  NETBURSTING = FALSE;	/* after 2 min start parsing on_join completely */
	}
	if ((time(NULL) - TopicTimer_1) >= 86400)
	{
	  TopicTimer_1 = TopicTimer_2 = TopicTimer_3 = TopicTimer_4 = time(NULL);
	  SetTopics(1);		/* One a day and less */
          ExpireChannels(NULL); /* This seems like a good place */
	}
	else if ((time(NULL) - TopicTimer_2) >= 43200)
	{
	  TopicTimer_2 = TopicTimer_3 = TopicTimer_4 = time(NULL);
	  SetTopics(2);		/*  12 hours and less */
	}
	else if ((time(NULL) - TopicTimer_3) >= 14400)
	{
	  TopicTimer_3 = TopicTimer_4 = time(NULL);
	  SetTopics(3);		/*  4 hours and less */
	}
	else if ((time(NULL) - TopicTimer_4) >= 3600)
	{			/* Once an hour fix hourly topics.. */
	  TopicTimer_4 = time(NULL);
	  SetTopics(4);		/* 1 hour */
	}
      }
  }
}

/* EOF */
