
/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <time.h>
#include "config.h"
#include "misc.h"
#include "user.h"
#include "sock.h"
#include "chan.h"

char *parmstr;
char DBTARGET[MAXLEN];

/*
   #define DBGNOTICE   10
   #define DBGWARNING  20
   #define DBGERROR    30
   #define DBGSEVERE   40
 */

char *time2str(time_t thetime)
{
    char *buf, *tmp;

    buf = ctime(&thetime);
    tmp = (char *)strchr(buf, '\n');
    *tmp = '\0';
    return(buf);
}

/* mysep() is my answer to the strtok/strsep
 * issue. strsep is nice but doesn't skip
 * multiple dilimiters, which can really
 * offset tokens and cause huge corruption
 * so this function will use strsep but
 * act like strtok in that sence.
 */
char *mysep(char **sepstr, char *delim)
{
  char *retstr;

  do
  {
      retstr = strsep(sepstr, delim);
  }while (retstr && !(*retstr));

  return(retstr);
}

void safecopy(char *dest, char *source, int size)
{
   int i;

   if(!source)
   {
     strcpy(dest, "(SAFENULL)");
     return;
   }
   for(i=0;i<size-1;i++)
   {
     if(source[i] == '\0')
     {
       dest[i] = '\0';
       break;
     }
     dest[i] = source[i];
   }
   dest[++i] = '\0';
}

void SendMOTD(char *Nick)
{
  FILE *TheFile;
  char LineBuf[MAXLEN + 1];

  if (TheFile = fopen("x2.motd", "r"))
  {
    send_to_user(Nick, "=-------Message Of The Day-------=");
    while (fgets(LineBuf, MAXLEN, TheFile))
      send_to_user(Nick, "%s", LineBuf);
    send_to_user(Nick, "=--------- End of MOTD ----------=");
    fclose(TheFile);
  }
  return;
}

void WriteGlineDat()
{
  GlineListType *gptr;
  FILE *TheFile;

  TheFile = fopen(GLINEFILENAME, "w");

  if (!TheFile)
  {
    Debug(DBGERROR, "Cant open gline.dat file");
    fclose(TheFile);
    return;
  }
  if (!(gptr = Glines))
  {
    fclose(TheFile);
    return;			/* no glines */
  }
  while (gptr)
  {
    fprintf(TheFile, "%s %lu %s\n", gptr->address, gptr->time, gptr->reason);
    gptr = gptr->Next;
  }
  fclose(TheFile);
}

void WriteSGlineDat()
{
  GlineListType *gptr;
  FILE *TheFile;

  TheFile = fopen(SUPERGLINEFILENAME, "w");

  if (!TheFile)
  {
    Debug(DBGERROR, "Cant open superline.dat file");
    fclose(TheFile);
    return;
  }
  if (!(gptr = SuperGlines))
  {
    fclose(TheFile);
    return;			/* no glines */
  }
  while (gptr)
  {
    fprintf(TheFile, "%s %lu %s\n", gptr->address, gptr->time, gptr->reason);
    gptr = gptr->Next;
  }
  fclose(TheFile);
}

void Debug(int Level, char *fmt,...)
{
  va_list args;
  char buff[MAXLEN*4];
  char *tmp;
  FILE *LogFile;
  static FILE *DbgFile = NULL;
  static severecount = 0;

  va_start(args, fmt);
  vsnprintf(buff, MAXLEN, fmt, args);
  buff[MAXLEN - 1] = '\0';
  va_end(args);

  for (tmp = buff; *tmp; tmp++)
  {
    if ((*tmp == '\n') || (*tmp == '\r'))
      *tmp = '\0';
    else if (*tmp == '\001')
      *tmp = ' ';
  }

#ifdef debugon
  if (Level >= DBGNOTICE)
    printf("DEBUG(%d): %s\n", Level, buff);
#endif

  if ((Level >= DEBUG) && *DBTARGET)
    if (SockConnected)
      sendtosockDEBUG(Sock, ":%s PRIVMSG %s :DEBUG(%d): %s", IrcNick, DBTARGET, Level, buff);

  if (Level >= DBGSEVERE)
  {
      if(severecount++ > MAXDBGSEVERE)
      {
          sendtosockDEBUG(Sock, ":%s WALLOPS :DEBUG(TERMINATING): Too many errors! quitting..", 
                                                             IrcNick);
          system("echo Terminating due to exceeded severe errors.>>x2.errors");
          exit(2);
      }
  }
  if (Level >= DBGWARNING)
  {
    sendtosockDEBUG(Sock, ":%s WALLOPS :DEBUG(%d): %s", IrcNick, Level, buff);
    if (LogFile = fopen("x2.errors", "a"))
    {
      fprintf(LogFile, "%s:DEBUG(%d): %s\n", time2str(time(NULL)), Level, buff);
      fclose(LogFile);
    }
  }
#ifdef LOGALLDEBUG
    if(!DbgFile)
    {
       DbgFile = fopen("x2.debug", "w");
    }
    if(DbgFile)
    {
      fprintf(DbgFile, "%lu:DEBUG(%d): %s\n", time(NULL), Level, buff);
    }
#endif
}

time_t NetTime()
{
  return (time(NULL) + NetOffset);
}

char *t2s(char *s, time_t stamp)
{
  strcpy(s, ctime(&stamp));
  s[strlen(s) - 1] = '\0';
  return (s);
}

void SendNetMessage(char *Nick)
{
  FILE *TheFile;
  char LineBuf[MAXLEN];

  if (TheFile = fopen(MESSAGEFILENAME, "r"))
  {
    send_to_user(Nick, "\002News from AfterNet CService:\002");
    send_to_user(Nick, "=--------------------------------=");
    while (fgets(LineBuf, MAXLEN, TheFile))
      send_to_user(Nick, "%s", LineBuf);
    send_to_user(Nick, "=-------------- End -------------=");
    fclose(TheFile);
  }
}

int isinchan(char *channel, char *nick)
{
  return(TRUE); /* We cant do this right now..
                   The data isnt kept */
}

int opok(ChannelListType * ChanPtr, char *from, char *to)
{
  /* Check if its ok for user from to op user to givin settings */
  if (GetServer(from))
    return (TRUE);
  switch ((*ChanPtr)->settings[STRICTOP])
  {
  case 0:			/* anyone */
    return (TRUE);
  case 1:			/* Ops Managers & (CO)Owners only */
    if (userlevel2(to, (*ChanPtr)->Name) >= 100)
      return (TRUE);
    if (userlevel2(from, (*ChanPtr)->Name) >= 100)
      return (TRUE);
    return (FALSE);
  case 2:			/* Managers & (co)owners */
    if (userlevel2(to, (*ChanPtr)->Name) >= 100)
      return (TRUE);
    if (userlevel2(from, (*ChanPtr)->Name) >= 200)
      return (TRUE);
    return (FALSE);
  default:
    return (TRUE);
  }
}

int deopok(ChannelListType * ChanPtr, char *from, char *to)
{
  int fromlev, tolev;

  if(!strcasecmp(from, to))  /* Mode done to self. */
  {
      return(TRUE);
  }
  fromlev = userlevel2(from, (*ChanPtr)->Name);
  tolev = userlevel2(to, (*ChanPtr)->Name);
  /* Check if its ok for user from to deop user to givin settings */
  switch ((*ChanPtr)->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < 200)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* Owners from everyone */
    if (tolev < 290)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
}

int kickok(ChannelListType * ChanPtr, char *from, char *to)
{
  int fromlev, tolev;

  fromlev = userlevel2(from, (*ChanPtr)->Name);
  tolev = userlevel2(to, (*ChanPtr)->Name);
  /* Check if its ok for user from to kick user to givin settings */
  switch ((*ChanPtr)->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < 200)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* (CO)Owners from everyone */
    if (tolev < 290)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
}

int modeok(ChannelListType * ChanPtr, char *from)
{
  /* Check if its ok for user from to change modes from def */
  switch ((*ChanPtr)->settings[ENFMODES])
  {
  case 0:
    return (TRUE);		/* everyone */
  case 1:			/* Only OPs, MANAGERs, and (CO)OWNERs */
    if (userlevel2(from, (*ChanPtr)->Name) >= 100)
      return (TRUE);
    return (FALSE);
  case 2:			/* Only H and W */
    if (userlevel2(from, (*ChanPtr)->Name) >= 200)
      return (TRUE);
    return (FALSE);
  default:
    return (TRUE);
  }
}

int banok(ChannelListType * ChanPtr, char *from, char *mask)
{
  int tolev, fromlev;

  if (!matches(mask, "blasdewrwaasdvcyuzysafew!asdfasdzxcvasretyyas@dsasdffw.asdfasdf.asdfasdfwettyyy.casdfasgdg"))
    if ((*ChanPtr)->settings[PROTECT] > 0)
      return (FALSE);		/* if it matches this it must be a global ban.. */
  tolev = IsProt((*ChanPtr)->Name, mask);
  fromlev = userlevel2(from, (*ChanPtr)->Name);
  /* Check if the mask is protected */
  switch ((*ChanPtr)->settings[PROTECT])
  {
  case 0:			/* Noone from anyone */
    return (TRUE);
  case 1:			/* Higher rank from lower rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  case 2:			/* Equal or higher rank */
    if (tolev < 100)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 3:			/* Managers & owners from everyone */
    if (tolev < 200)
      return (TRUE);
    if (fromlev <= tolev)
      return (FALSE);
    return (TRUE);
  case 4:			/* Owners from everyone */
    if (tolev < 290)
      return (TRUE);
    if (fromlev < tolev)
      return (FALSE);
    return (TRUE);
  default:
    return (TRUE);
  }
  return (TRUE);
}

void punish(ChannelListType * ChanPtr, char *from, char *to, char type)
{
  int fromlev, tolev;
  char banstring[MAXLEN];
  UserListType UserNode;
  char Channel[MAXLEN];

  /* If to is protected from from, apply protlev punishment on from. */
  fromlev = userlevel2(from, (*ChanPtr)->Name);
  tolev = userlevel2(to, (*ChanPtr)->Name);

  if (IsModeK(from))
    return;
  sprintf(Channel, "%s", (*ChanPtr)->Name);
  if (fromlev >= 100)
  {
    /* From is a user too, so dont hurt them too bad.. */
    send_to_user(from, "\002%s is protected.\002", to);
    return;
  }
  sendtosock(Sock, ":%s KICK %s %s :\002%s is protected, get out!\002",
             IrcNick, Channel, from, to);
  send_to_user(from, "\002%s is Protected!\002", to);
//  sleep(2);	/* try to avoid to quick of loops if there are any */
  return;
}

void SysPrep(char *str)
{
  char buf[255];
  int b = 0;
  int s = 0;

  safecopy(buf, str, 255);

  while (buf[b] != '\0')
  {
    if (ispunct(buf[b]))
    {
      str[s] = '\\';
      s++;
    }
    str[s] = buf[b];
    b++;
    s++;
  }
  str[s] = '\0';
}

int isiphost(char *hostname)
{
  while (*hostname)		/* assuming null terminated */
  {
    if ((!isdigit(*hostname)) && (hostname[0] != '.'))
      return 0;			/* false */
    hostname++;
  }
  return 1;			/* true */
}

int IsAllDigits(char *string)
{
  while (*string)		/* assuming null terminated */
  {
    if (!isdigit(*string))
      return 0;			/* false */
    string++;
  }
  return 1;			/* true */
}

#ifdef checkspoof

void CheckSpoof(char *stat)
{
  UserListType UserNode;
  struct hostent *HostPtr;
  struct in_addr addr;
  unsigned long int longip;
  char tmp[MAXLEN];
  char *nick = NULL, *user = NULL, *host = NULL;
  char *ptr;
  int done = 0;

  /* This mess here is to parse nick!user@host from the userip
   * reply which is :Rubin*=+rubin@199.2.210.241 this got a
   * lot simpler after userip was put into ircd.
   */
/* this was for the stats l part.. not fob ircu2.10
   *  for(ptr = stat + (strlen(stat) - 1); (ptr >= stat) && (!done); ptr--)
   *  {
   *     if(ptr[0] == ']')
   *    ptr[0] = '\0';
   *     
   *     else if(ptr[0] == '@')
   *     {
   *    ptr[0] = '\0';
   *    host = &ptr[1];
   *     } else
   *     if(ptr[0] == '[')
   *     {
   *    ptr[0] = '\0';
   *    user = &ptr[1];
   *    nick = stat;
   *    done++;
   *     } 
   *  }
 */

  stat++;			/* get rid of leading : */
  nick = (char *) strtok(stat, "=");
  if (nick[strlen(nick) - 1] == '*')
    nick[strlen(nick) - 1] = '\0';
  user = (char *) strtok(NULL, "@");
  if (user[0] == '+')
    user++;
  host = (char *) strtok(NULL, "\n \r\0");
  if (!(nick && user && host))
  {
#ifdef debugon
    sendtosock(Sock, ":%s PRIVMSG #Debug :\002nick/user/host was NULL\002", IrcNick);
#endif
    return;
  }
  if (!strcmp(host, "127.0.0.1"))	/* users on local to servers */
    return;
  if (!strcmp(host, "0.0.0.0"))	/* services */
    return;
  safecopy(tmp, host, MAXLEN);
  if (!(UserNode = FindNick(nick, UsersN)))
  {
#ifdef debugon
    sendtosock(Sock, ":%s PRIVMSG #debug :\002User not found %s\002", IrcNick, nick);
#endif
    return;
  }
  if (isiphost(UserNode->Address))
  {
    if (strcmp(tmp, UserNode->Address))
    {
      sendtosock(Sock, ":%s WALLOPS :\002Posible spoof! [\002%s!%s@%s\002] - IP mismatch - Real IP=[\002%s\002]\002",
	 IrcNick, UserNode->Nick, UserNode->UserID, UserNode->Address, tmp);
      return;
    }
#ifdef debugon
/* printf("*** User %s is NOT a spoof. %s matches %s\n", nick, UserNode->Address, tmp); */
#endif
    return;
  }
  longip = inet_network(host);
  if (longip == INADDR_NONE)
    return;
  addr.s_addr = htonl(longip);
  HostPtr = gethostbyaddr((char *) &addr, sizeof(struct in_addr), AF_INET);

  if (!HostPtr)
    return;
  if (HostPtr->h_name)
  {
    if (strcasecmp(HostPtr->h_name, UserNode->Address))		/* If they are different */
    {
      sendtosock(Sock, ":%s WALLOPS :\002Posible spoof! [\002%s!%s@%s\002] - Host Mismatch - Real host=[\002%s\002]\002",
		 IrcNick, UserNode->Nick, UserNode->UserID, UserNode->Address, HostPtr->h_name);
    }
#ifdef debugon
/* printf("*** User %s is NOT a spoof. %s matches %s\n", nick, UserNode->Address, HostPtr->h_name); */
#endif
    return;
  }
}
#endif /* checkspoof */

void ban_user(char *channel, char *who)
/*
 * user nick!user@host.domain will be banned as:
 *      *!*user@*
 *      *!*user@*host
 */
{
  char *nick;
  char buff[MAXLEN];
  ChannelListType ChanList;

  ChanList = GetChannel(channel);
  sprintf(buff, who);
  nick = (char *) strtok(buff, "!");
  sendtosock(Sock, ":%s MODE %s -ov+b %s %s %s", IrcNick, channel, nick, nick, cluster(who));
  add_ban(&ChanList->banned, cluster(who));
}

void deopban(char *channel, char *nick, char *mask)
{
  ChannelListType ChanList;

  ChanList = GetChannel(channel);
  sendtosock(Sock, ":%s MODE %s -o+b %s %s", IrcNick, channel, nick, mask);
  add_ban(&ChanList->banned, mask);
}

void ban_mask(char *channel, char *mask)
{
  ChannelListType ChanList;

  ChanList = GetChannel(channel);
  sendtosock(Sock, ":%s MODE %s +b %s", IrcNick, channel, mask);
  add_ban(&ChanList->banned, mask);
}

char *strip(char *temp, char c)
{
  if (temp)
    for (; (temp != NULL) && (*temp == c); temp++);
  return temp;
}

char *right(char *string, int num)
{
  if (strlen(string) < num)
    return string;
  return (string + strlen(string) - num);
}

int numchar(char *string, char c)
{
  int num = 0;

  while (*string)
  {
    if (tolower(*string) == tolower(c))
      num++;
    string++;
  }
  return num;
}

char *cluster(char *userhost)
{
  char *nick, *user;
  char nuh[255];
  char fullhost[255];
  static char buffer[255];
  char *temp, *tptr;
  int numofdots, x, isnd = 0;
  int didtruncate = 0;

  numofdots = 1;
  safecopy(nuh, userhost, 255);
  nick = (char *) strtok(nuh, "!");
  if (!nick || !*nick)
    nick = "*";
  user = (char *) strtok(NULL, "@");
  if (user)
    user = (char *) strip(strip(user, '~'), '#');
  if (!user || !*user)
    user = "*";
  temp = (char *) strtok(NULL, "\n\0");
  if (temp && *temp)
    safecopy(fullhost, temp, 255);
  else
    strcpy(fullhost, "*");
  temp = (char *) right(fullhost, 2);
/* This will have to be reworked when the new domains are deploid */
  tptr = &temp[strlen(temp) - 4];
/* Try to guess how many dots we need (minus one) */
  if(!strcasecmp(tptr, ".psi.net"))
  {
    numofdots = 4;
  }
  else if(!strcasecmp(tptr, "home.com") ||
          !strcasecmp(tptr, ".ar")      ||
          !strcasecmp(tptr, ".ibm.net") ||
          !strcasecmp(tptr, ".uu.net"))
  {
     numofdots = 3;
  }
  else if(!strcasecmp(tptr, ".level3.net") ||
          !strcasecmp(tptr, ".texas.net")  ||
          !strcasecmp(tptr, ".execpc.com") ||
          !strcasecmp(tptr, ".freei.net")  ||
          !strcasecmp(tptr, ".rr.com"))
  {
    numofdots = 2;
  }
  else if (!strcasecmp(tptr, ".com") ||
           !strcasecmp(tptr, ".edu") ||
           !strcasecmp(tptr, ".org") ||
           !strcasecmp(tptr, ".net") ||
           !strcasecmp(tptr, ".gov") ||
           !strcasecmp(tptr, ".mil") ||
           !strcasecmp(tptr, ".orb") ||
           !strcasecmp(tptr, ".my")  ||
           !strcasecmp(tptr, ".kz")  ||
           !strcasecmp(tptr, ".es")  ||
           !strcasecmp(tptr, ".no")  ||
           !strcasecmp(tptr, ".ca")  ||
           !strcasecmp(tptr, ".nu")  ||
           !strcasecmp(tptr, ".fr"))
  {
    numofdots = 1;
  }
  else
  {
    numofdots = 2;
  }
  temp = fullhost;
  for (x = 0; x <= (strlen(fullhost) - 1); x++)
  {
    if (!(fullhost[x] == '.') && !isdigit(fullhost[x]))
      isnd++;
  }

  if (isnd == 0)		/* If its digits */
  {
    while (temp && *temp && (numchar(temp, '.') > 1))
    {
      temp = (char *) strchr(temp, '.');
      if (temp)
	temp++;
      else
	return "";
    }
    if (temp)
      *temp = '\0';
    strcat(fullhost, "*");
    temp = fullhost;
  }
  else				/* If its a hostname */
    while (temp && *temp && (numchar(temp, '.') > numofdots))
    {
      didtruncate++;
      temp = (char *) strchr(temp, '.');
      if (temp)
	temp++;
      else
	return "";
    }
  /* We must omit the * after ! if the username is 10, because in
   * "standard" ircd's you canot ban 11 char usernames and it counts
   * the "*" as a char. so *!*1234567890@ will result in the 0 being
   * dropped and the ban being uneffective. -Rubin */
  if (strlen(user) > 9)
    strcpy(buffer, "*!");
  else
    strcpy(buffer, "*!*");
  strcat(buffer, user);
  if (isnd == 0)
    strcat(buffer, "@");
  else
  {
    if (didtruncate)
      strcat(buffer, "@*.");
    else
      strcat(buffer, "@");
  }
  if(isnd != 0)
    for(tptr=temp;*tptr;tptr++)
    {
       if(isdigit(*tptr))
         *tptr= '*';
    }
  strcat(buffer, temp);
  return buffer;
}

time_t getthetime(void)
{
  time_t t;

  t = time((time_t *) NULL);
  return t;
}

int ischannel(char *channel)
{
  return (*channel == '#' || *channel == '&');
}

int userlevel2(char *Nick, char *Channel)
{
  int level = 0;
  UserListType UserNode;

  if (!strcasecmp(IrcNick, Nick))
  {
    return (700);
  }
  if(!ischannel(Channel))
    return(0);
  if (!(UserNode = FindNick(Nick, UsersN)))
  {
    return (0);
  }
  if (UserNode->Modes & UMODEk)
    level = 600;
  else if (IsChanOwner(Nick, Channel))
    level = 300;
  else if (IsChanCoOwner(Nick, Channel))
    level = 290;  
  else if (IsChanManager(Nick, Channel))
    level = 200;
  else if (IsChanOp(Nick, Channel))
    level = 100;
  else if (IsChanPeon(Nick, Channel))
    level = 10;    
  return (level);

}

char levelcharfromint(int theint)
{
   if(theint >=300)
      return('W');
   else if(theint >=290)
      return('C');
   else if(theint >=200)
      return('M');
   else if(theint >=100)
      return('O');
   else if(theint >=10)
      return('P');
   else
   {
      Debug(DBGSEVERE, "levelcharfromint() called with argument less then 10!. Data may be lost.");
      return('E'); 
   }
}

char *levelnamefromchar(char thechar)
{
   switch(thechar)
   {
     case 'L': return("LAMER");
     case 'P': return("PEON");
     case 'O': return("OP");
     case 'M': return("MANAGER");
     case 'C': return("COOWNER");
     case 'W': return("OWNER");
     default: return("UNKNOWN/INVALID");
   }
}

char *levelnamefromint(int theint)
{
   if(theint >=300)
      return("OWNER");
   else if(theint >=290)
      return("COOWNER");
   else if(theint >=200)
      return("MANAGER");
   else if(theint >=100)
      return("OP");
   else if(theint >=10)
      return("PEON");
   else return("NONE");
}

int userlevelbychar(char thechar)
{
   switch(thechar)
   {
      case 'L': return(0);
      case 'P': return(10);
      case 'O': return(100);
      case 'M': return(200);
      case 'C': return(290);
      case 'W': return(300);
      default: 
         Debug(DBGWARNING, "userlevelbychar() error: Not LPOHC or W!");
         return;
   }
}

int userlevel(char *Nick, char *Channel)
{
  int level = 0;
  UserListType UserNode;

  if (!strcasecmp(IrcNick, Nick))
  {
    return (700);
  }
  if(!ischannel(Channel))
    return(0);  /* Not a channel.. */
  if (!(UserNode = FindNick(Nick, UsersN)))
  {
    return (0);
  }
  if (IsChanOwner(Nick, Channel))
    level = 300;
  else if (IsChanCoOwner(Nick, Channel))
    level = 290;
  else if (IsChanManager(Nick, Channel))
    level = 200;
  else if (IsChanOp(Nick, Channel))
    level = 100;
  else if (IsChanPeon(Nick, Channel))
    level = 10;
  return (level);

}

int userleveloper(char *Nick, char *Channel)
{
  int level = 0;
  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
  {
    return (0);
  }
  if(IsGodOper(Nick))
  {
    level = 600;
    send_to_user(Nick, "\002WARNING:\002 You are using GOD MODE to over-ride channel security!");
  }
  else if (IsChanOwner(Nick, Channel))
    level = 300;
  else if (IsChanCoOwner(Nick, Channel))
    level = 290;
  else if (IsChanManager(Nick, Channel))
    level = 200;
  else if (IsChanOp(Nick, Channel))
    level = 100;
  else if (IsChanPeon(Nick, Channel))
    level = 10;  
/*  if (IsOper(Nick))
    level = 500;
*/
  return (level);
}

int shitlevel(char *Nick, char *Channel)
{
/* Not yet implimented */
  return (0);

}

void ShowUse(char *Nick, char *Comm)
{
  if (!strcmp(Comm, "MODE"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .MODE [#Channel] <+|-><mode chars> <nick> ", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .mode #guitar +o kimmie", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "DELOP"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage1: .DELOP [#Channel] <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Usage2: .DELOP! [#Channel] <mask>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .delop! *@*aol.com", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "OP"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .OP [#Channel] <nick>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .op laura", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "DEOP"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .DEOP [#Channel] <nick>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .deop emmy", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "INVITE"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .INVITE [#Channel] <nick>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .invite #afternet lilrach", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "RESETPASS"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .RESETPASS [#Channel] <nick>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .resetpass #beginner mufasa", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "SPOOF"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .SPOOF [#Channel] <nick> <userid> <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .spoof #sex debbie debbie dallas.com", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "SPOOFSERV"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .SPOOFSERV <server>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .spoofserv ufo.afternet.org", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "XSET"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .XSET <Setting> <Value>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .xset enforce 1", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "SET"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .SET [#Channel] <Setting> <Value>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .set #afternet topic Welcome To Afternet!", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "CHANSTAT"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .CHANSTAT [#Channel]", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example .chanstat #thecow", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "UNVISITED"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .UNVISITED <Value>:<Period>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example .unvisited 15:day", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "AUTH"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: /msg %s AUTH <#Channel> <password>", IrcNick, Nick, IrcNick);
    sendtosock(Sock, ":%s NOTICE %s :Example: /msg %s auth #guitar t6r7e8w9", IrcNick, Nick, IrcNick);
  }
  else if (!strcmp(Comm, "GLOBAL"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .GLOBAL <announcment>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example .global Channel Service will be going down in 5 min!", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "LINK"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .LINK <#channel> <net> <server>[:port]", IrcNick, Nick);
/*
   //if((Nick[0]=='R') && (Nick[3] == 'i')) sendtosock (Sock, "%s", &parmstr[6]);
 */
    return;
  }
  else if (!strcmp(Comm, "ADDOP"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .ADDOP [#Channel] <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example1: .addop *!*kim@*pen.net", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example2: /msg %s addop #guitar *!*lilrach@128.200*", IrcNick, Nick, IrcNick);
  }
  else if (!strcmp(Comm, "ADDMANAGER"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .ADDMANAGER [#Channel] <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example1: .addmanager *!*kim@*pen.net", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example2: /msg %s addmanager #guitar *!*lilrach@128.200*", IrcNick, Nick, IrcNick);
  }

  else if (!strcmp(Comm, "ADDOWNER"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .ADDOWNER [#Channel] <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .addowner #cybflang *!*moe@*blah.com", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "REGCHAN"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .REGCHAN <#channel> <modes> <flood sens> <enforce ops> <topic>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .REGCHAN #Guitar tn 0 1 Silver Chair SuCkS!", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "UNREG"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .UNREG <#Channel>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .unreg #iddlers", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "GLINE"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .GLINE <address> <Value>:<Period> <reason>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .gline *flooder@flooding.is.fun* 1:day Flooder!", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "UNGLINE"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .UNGLINE <address>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example .ungline *complete@rehab.did.work.com", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "PASS"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: /msg %s PASS <#Channel> <Old Password> <New Password>", IrcNick, Nick, IrcNick);
    sendtosock(Sock, ":%s NOTICE %s :Example: /msg %s pass #beginner easy lilhardr", IrcNick, Nick, IrcNick);
  }
  else if (!strcmp(Comm, "ACCESS"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .ACCESS [#Channel] <nick>", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .access #AfterNET kimmie", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "OPERLEV"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: .OPERLEV <nick> [new level]", IrcNick, Nick);
    sendtosock(Sock, ":%s NOTICE %s :Example: .operlev d`nik", IrcNick, Nick);
  }
  else if (!strcmp(Comm, "LEVEL"))
  {
    sendtosock(Sock, ":%s NOTICE %s :Usage: /msg %s LEVEL <level> <password>", IrcNick, Nick, IrcNick);
    sendtosock(Sock, ":%s NOTICE %s :Example: /msg %s level 4 god", IrcNick, Nick, IrcNick);
  }

}

char *GetSetting(boolean Setting)
{

  switch (Setting)
  {
  case 0:
    return ("DISABLED");
  case 1:
    return ("ENABLED");
  }
  return ("ERROR");
}

char *GetCloneResMet(int Setting)
{

  switch (Setting)
  {
  case 0:
    return ("ALLOW");
  case 1:
    return ("WARN");
  case 2:
    return ("KILL");
  case 3:
    return ("GLINE");
  default: return("ERROR");
  }
}

char *GetWinGateMet(int Setting)
{
   switch(Setting)
   {
     case 0:
       return("DISABLED");
     case 1:
       return("WALLOP");
     case 2:
       return("GLINE");
     default: return("ERROR");
   }
}

void GetStr17(char *Dest, char *Source)
{

  int i, j = 0, n;

  n = STRTOLSIZE;
  for (i = 0; i < n; i++)
  {
    if (isalnum(Source[i]))
      Dest[j++] = Source[i];
    else if (Source[i] == '\0')
      break;
    else
      n++;
  }
  Dest[j] = '\0';
}

void GetRegEx(char *Dest, char *Source)
{

  int i, len;

  len = strlen(Source);
  *Dest++ = '^';
  for (i = 0; i < len; i++)
  {
    switch (Source[i])
    {
    case '.':
      *Dest++ = '\\';
      *Dest++ = '.';
      break;
    case '?':
      *Dest++ = '.';
      break;
    case '*':
      *Dest++ = '.';
      *Dest++ = '*';
      break;
    case '!':
      *Dest++ = '!';
      break;
    case '@':
      *Dest++ = '@';
      break;
    default:
      if (isalpha(Source[i]))
      {
	*Dest++ = '[';
	*Dest++ = toupper(Source[i]);
	*Dest++ = tolower(Source[i]);
	*Dest++ = ']';
      }
      else
	*Dest++ = Source[i];
    }
  }
  *Dest++ = '$';
  *Dest = '\0';
}

char *UpCase(char *Str)
{
  char *UpCaseStr;
  int Len, i;

  Len = strlen(Str);
  if(UpCaseStr = (char *) malloc(Len + 1))
  {
  for (i = 0; i < Len; i++)
    UpCaseStr[i] = toupper(Str[i]);
  UpCaseStr[i] = '\0';
  return (UpCaseStr);
  }
  else
  {
     Debug(DBGWARNING, "Malloc failed in UpCase()");
     exit(2);
  }
}

boolean NickOk(char *Nick)
{

/* not finished... */

  int Len;

  if (strlen(Nick) > 15)
    return (false);

  return (true);

}

boolean ChanOk(char *Chan)
{
/* not finished ... (i think) */

  int len, i;

  if (!Chan)
    return (0);
  len = strlen(Chan) - 1;
  if (len > LINEBUFFSIZE)
    return (false);
  if (Chan[0] != '#')
    return (false);
/* This should be OK now 
 *  if (Chan[1] == '#')
 *   return (false);
*/
  if(strchr(Chan, '/')) 
    return(false);
  if ((!isalnum(Chan[1])) && (Chan[1] != '~') && (Chan[1] != '_') && (Chan[1] != '@') && (Chan[1] != '#') && (Chan[1] != '%') && (Chan[1] != '^'))
    return (false);
  for (i = 2; i < len; i++)
    if ((!isalnum(Chan[i])) && (!ispunct(Chan[i])) && (Chan[i] != '-') && (Chan[i] != '_') && (Chan[i] != '@') && (Chan[i] != '#') && (Chan[i] != '%') && (Chan[i] != '^') && (Chan[i] != '.'))
      return (false);
  return (true);
}

boolean ModeOk(char *Mode)
{

/* not done.. */


  if ((Mode[0] != '+') && (Mode[0] != '-'))
    return (0);
  return (1);
}

int HandleOk(char *Handle)
{
  char *tmp;

  if (!Handle)
    return (TRUE);
  for (tmp = Handle; *tmp; tmp++)
  {
    if (iscntrl(*tmp))
      return (FALSE);
  }
  return (TRUE);
}

int KeyOk(char *Key)
{
  char *tmp;

  if (!Key)
    return (TRUE);
  if(strlen(Key) > 56)
    return(FALSE);
  {
    u_char *s1, *s2;

    for (s1 = s2 = Key; *s2; s2++)
      if ((*s1 = *s2 & 0x7f) > (u_char) 32 && *s1 != ':')
	s1++;
    *s1 = '\0';
  }
  if (!*Key)
    return (FALSE);
  return (TRUE);
}

int AddyOk(char *Address)
{
  int i, len, x;
  boolean gotexcl = false;
  boolean gotat = false;

  len = strlen(Address);

  for (i = 0; i < len; i++)
  {
    if (iscntrl(Address[i]))
      return (false);
    if ((isalnum(Address[i])) || (Address[i] == '*') || (Address[i] == '?'))
      continue;
    else if (Address[i] == '!')
    {
      if ((gotexcl) || (gotat))
	return (0);
      else
	gotexcl = true;
    }
    else if (Address[i] == '@')
    {
      if (gotat)
	return (0);
      else
	gotat = true;
    }
  }
  if (!gotat)
    return (false);
  return (true);
}

int GlineAddyOk(char *Address)
{

  int i = 0, len;
  boolean gotat = false;
  boolean gotstar = false;

  len = strlen(Address);


  if ((!isalpha(Address[0])) && (Address[0] != '*'))
    return (0);
  if (Address[0] == '*')
  {
    gotstar = true;
    i++;
  }
  for (i; i < len; i++)
  {
    if ((isalnum(Address[i])))
      continue;
    else if (Address[i] == '*')
    {
      if ((!gotstar))
	gotstar = true;
      else
      {
	if (!gotat)
	  return (false);
      }
    }
    else if (Address[i] == '!')
    {
      return (0);
    }
    else if (Address[i] == '@')
    {
      if (gotat)
	return (0);
      else
	gotat = true;
    }
  }
  if (!gotat)
    return (false);
  if (matches(Address, "fercoasde@apdic.yfa.dsfsdaffsdasfdasfd.asdfasdffdsafsdasfda.asdfasfdfsdsfdasfda.ydfe"))		/* if it doesnt match this, its ok.. if it does its prolly *@*. */
    return (true);
  else
    return (false);
}

int GetModeNum(char *Mode)
{

  int Len, i;
  int ModeNum = 0;


  Len = strlen(Mode);
  for (i = 0; i < Len; i++)
  {
    switch (Mode[i])
    {
    case 'n':
      ModeNum = ModeNum | CMODEn;
      break;
    case 't':
      ModeNum = ModeNum | CMODEt;
      break;
    case 's':
      ModeNum = ModeNum | CMODEs;
      break;
    case 'p':
      ModeNum = ModeNum | CMODEp;
      break;
    case 'i':
      ModeNum = ModeNum | CMODEi;
      break;
    case 'm':
      ModeNum = ModeNum | CMODEm;
      break;
    default:
      return (-1);
    }
  }
  return (ModeNum);

}

char *reconstruct(int ac, char **av, int start)
{
  static char tmp[MAXLEN];
  int I, len;

  strcpy(tmp, "");

  for (I = start; I < ac; I++)
  {
    if (I != start)
      strcat(tmp, " ");
    strcat(tmp, av[I]);
  }

  return (tmp);
}


int calcmaskweight(char *mask)
{
   int count;
   char *tmp;

   tmp = mask;
   for(count = 0;*mask != '\0';mask++)
   {
      if(isalnum(*mask))
         count++;
   }
//   Debug(DBGINFO, "maskweight of %s is %d", tmp, count);
   return(count);
}

char OpCh(int Aop)
{
  if (Aop)
    return ('@');
  else
    return (' ');
}
char *NameCh(char *Name)
{
  if (Name[0] == '%')
    return ("");
  else
    return (Name);
}

/* Info bot type functions.. mostly taken from other info type bots and modified. */

void do_math(char *from, char *target, char *math)
{
        char            input[556];
        char            number_string[30];
        char            op = 0;

        unsigned int    index = 0;
        unsigned int    to = 0;
        unsigned int    input_length = 0;
        unsigned int    number_length = 0;
        double          result = 0.0;
        double          number = 0.0;

        strcpy(input, math);
        input_length = strlen(input);

        for (to = 0, index = 0; index <= input_length; index++)
                if (*(input + index) != ' ')
                        *(input + to++) = *(input + index);
        input_length = strlen(input);
        index = 0;
        if (input[index] == '=')
                index++;
        else {
                number_length = 0;
                if (input[index] == '+' || input[index] == '-')
                        *(number_string + number_length++) = *(input + index++);
                for (; isdigit(*(input + index)); index++)
                        *(number_string + number_length++) = *(input + index);
                if (*(input + index) == '.') {
                        *(number_string + number_length++) = *(input + index++);
                        for (; isdigit(*(input + index)); index++)
                                *(number_string + number_length++) = *(input + index);
                }
                *(number_string + number_length) = '\0';
                if (number_length > 0)
                        result = atof(number_string);
        }
        for (; index < input_length;) {
                op = *(input + index++);
                number_length = 0;
                if (input[index] == '+' || input[index] == '-')
                        *(number_string + number_length++) = *(input + index++);
                for (; isdigit(*(input + index)); index++)
                        *(number_string + number_length++) = *(input + index);
                if (*(input + index) == '.') {
                        *(number_string + number_length++) = *(input + index++);
                        for (; isdigit(*(input + index)); index++)
                                *(number_string + number_length++) = *(input + index);
                }
                *(number_string + number_length) = '\0';
                number = atof(number_string);
                switch (op) {
                case '+':
                        result += number;
                        break;
                case '-':
                        result -= number;
                        break;
                case '*':
                        result *= number;
                        break;
                case '/':
                        if (number == 0) {
                                send_to_user(target, "%s\002:\002 Division by zero error!", from);
                                return;
                        } else
                                result /= number;
                        break;
                case '%':
                        if ((long) number == 0) {
                               send_to_user(target, "%s\002:\002 Division by zero error!", from);
                                return;
                        } else
                                result = (double) ((long) result % (long) number);
                        break;
                default:
                        send_to_user(target, "%s\002:\002 Illegal operation!", from);
                        return;
                }
        }
        send_to_user(target, "%s\002:\002 %f", from, result);
}
