/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "sock.h"
#include "chan.h"
#include "user.h"
#include "server.h"
#include "misc.h"
#include "commands.h"

extern unsigned long int TotalCommands, TotalLines, TotalUserCount;
extern time_t bantimer;

int ProcessCommand(char *from, char *to, char *rest)
{
   ChannelListType ChanPtr = NULL;
   char *Chan = "";
   char *Command;
   int public = 0;

   if(to[0] == '#')
   { /* A public command to a channel */
      if(rest[0] == CommandMark)
      {
         rest++;
         ChanPtr = GetChannel(to);     
         if(!ChanPtr)
         {
            Debug(DBGWARNING, "Recieved channel chatter when i dont have that channel in my DB");
            return(0);
         }
/*
         if(!ChanPtr->settings[PUBLICCMDS])
            return(0); 
*/
      }
      else
        return(0);  /* Msg to channel. ignore.. */
      public = TRUE;
   }
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");
      ChanPtr = GetChannel(Chan);
      if(!ChanPtr)
      {
         send_to_user(from, "Command failed: I'm not in channel %s", Chan);
         return(0);
      }
   }
   else if(public)
   {
      Chan = to;
   }
   Command = (char *) strtok(rest, " ");
   if(!Command)
   {
      Debug(DBGINFO, "Empty command. ignoring.");
      return(0);
   }
   rest = (char *) strtok(NULL, "\0\r\n");
/* OK well make the old way work too.. */
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");

/*
      ChanPtr = GetChannel(Chan);

      if(!ChanPtr)
      {
         send_to_user(from, "Command failed: I'm not in channel %s", Chan);
         return(0);
      }
*/
   }
   on_msg(from, Chan, to, public, Command, rest);
   return(0);
}

int ProcessCTCP(char *from, char *to, char *rest)
{
    char *command;

    command = (char *) strtok(rest, " ");
    rest = (char *) strtok(NULL, "\0\r\n");
    if(!command)
    {
       return(0); /* empty ctcp */
    }
    if(!strcasecmp(command, "VERSION"))
    {
       sendtosock(Sock, ":%s NOTICE %s :%cVERSION %s%c", IrcNick, from, CTCP, VERSION, CTCP);
       return(0);
    }
    else if(!strcasecmp(command, "PING"))
    {
       if(!rest)
         return(0); /* empty ping request */
       sendtosock(Sock, ":%s NOTICE %s :%cPING %s%c", IrcNick, from, CTCP, rest, CTCP);
       return(0);
    }
    else if(!strcasecmp(command, "CLIENTINFO"))
    {
      sendtosock(Sock, ":%s NOTICE %s :%cCLIENTINFO VERSION PING CLIENTINFO TIME%c", IrcNick, from, CTCP, CTCP);
      return(0);
    }
    else if(!strcasecmp(command, "TIME"))
    {
      sendtosock(Sock, ":%s NOTICE %s :%cTIME %s%c", IrcNick, from, CTCP, time2str(time(NULL)), CTCP);
      return(0);
    }
}

int ProcessMsg(char *from, char *to, char *rest)
{
   char *target,    /* The commands intended target */
        *channel;   /* Where they typed the command */

   /* Was it done in a channel or to me personaly? */
   /* Just a thought.. is it possible to get NICK PRIVMSG #channel1,#channel2 etc? */
   if(rest[0] == ':')
      rest++;
   /* Check for CTCP */
   if(rest[0] == CTCP)
   {
      if(rest[strlen(rest)-1] != CTCP)
      {  /* No terminator on CTCP, ignore it */
        Debug(DBGINFO, "Ignoring unterminated CTCP");
        return(0);
      }
      rest++;
      rest[strlen(rest)-1] = '\0';
      ProcessCTCP(from, to, rest);
      return(0);
   }
   else
   {
      ProcessCommand(from, to, rest);
      return(0);
   }
}

int ProcessLine(char *LineBuff)
{
  char *Arg[MAXLEN];
  char rest[MAXLINELEN];
  static char RemotePass[MAXPASS]; /* should be this or just '10'? */
  int i, num;
  FILE *LogFile;

  /* Give us a reasonable length limit.. */
  if(strlen(LineBuff) > MAXLINELEN)
  {
     LineBuff[MAXLINELEN-1] = '\0';
  }
  if(LogFile = fopen("x2.log", "w"))
  {
      fprintf(LogFile, "%s : %s\n", time2str(time(NULL)), LineBuff);
  }
  fclose(LogFile);
  Debug(DBGNOTICE, "<- %s", LineBuff);
  TotalLines++;
  Arg[0] = (char *) strtok(LineBuff, " ");
  Arg[1] = (char *) strtok(NULL, " ");

  if(!Arg[1])
  {
     Debug(DBGERROR, "No second argument from socket. Should this happen?");
     /*exit(1);*/  /* Temporary. eventualy this should just return.. */
     return(0);
  }
  if(Arg[0][0] != ':')
  {
    /* If a server or nick message, process below, else do it here.. */

      if(!strcmp(Arg[0], "PING"))
      {    /* Respond with a ping reply */
#ifdef HYBRIDTS4
        sendtosock(Sock, "PONG %s", Arg[1]); /* Already has the : */
#else  /* HYBRIDTS4 */
        sendtosock(Sock, "PONG :%s", Arg[1]); /* Needs the : */
#endif  /* HYBRIDTS4 */
        return (0);
      }
      else if (!strcmp(Arg[0], "ERROR"))
      {    /* Error handling should be implimented here.. */
        if(!(Arg[2] = (char *) strtok(NULL, "\0")))
           Arg[2] = "";
        Debug(DBGSEVERE, "ERROR from socket! :%s %s %s", Arg[0], Arg[1], Arg[2]);
        SockConnected = FALSE;
        return (0);
      }
      else if (!strcmp(Arg[0], "PASS"))
      {    /* store the password for compareing later when we get SERVER */
#ifndef HYBRIDTS4  /* Hybrid doesnt send the : before the password.. */
         Arg[1]++;
#endif
         Debug(DBGNOTICE, "Got PASS from server: %s", Arg[1]);
         safecopy(RemotePass, Arg[1], 10);
         return(0);
      }
#ifdef HYBRIDTS4
      else if (!strcmp(Arg[0], "CAPAB"))
      {
         /* Parse this correctly at some point.. */
         return(0);
      }
      else if (!strcmp(Arg[0], "SVINFO"))
      {  /*   SVINFO <TS_CURRENT> <TS_MIN> <TIMEDELTA-FLAG. :<UTC-TIME> */
         Arg[2] = (char *) strtok(NULL, " ");  /* Current TimeStamp */
         Arg[3] = (char *) strtok(NULL, " ");  /* Min TimeStamp */
         Arg[4] = (char *) strtok(NULL, " ");  /* Delta */
         Arg[5] = (char *) strtok(NULL, " "); /* UTC Time WITH the :*/
             /* Handle this eventually.. */
         return(0);
      }
      else if (!strcmp(Arg[0], "NOTICE"))
      {
        return(0);   /* Connection info */
      }
#endif /* HYBRIDTS4 */
      else if (!strcmp(Arg[0], "SERVER"))
      {    /* Beginning of burst. validate password and name, authenticate and send our burst
            * if we were listning for the link */
         UpLinkList *UpLinkPtr;
    
         Debug(DBGNOTICE, "Got SERVER line from uplink. Processing.");
         Arg[2] = (char *) strtok(NULL, " ");         /* Hops */
         Arg[3] = (char *) strtok(NULL, " ");         /* TS #1 */
         Arg[4] = (char *) strtok(NULL, " ");         /* TS #2 */
         Arg[5] = (char *) strtok(NULL, " ");         /* Protocol */
         Arg[6] = (char *) strtok(NULL, "\r\n\0");      /* Description */
#ifdef HYBRIDTS4
         if(!Arg[2])
#else
         if(!(Arg[6] && Arg[5]))
#endif
         {
            Debug(DBGWARNING, "Invalid SERVER recieved.  Killing link.");
            SockConnected = FALSE;
            return(0);
         }
         for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
         {
             Debug(DBGINFO, "Comparing incomming connection (%s) to our (%s)", Arg[1], UpLinkPtr->ServName);
             if(!strcasecmp(UpLinkPtr->ServName, Arg[1]))  /* If they match */
             {
                 /* We have a match, check password */
                 if(strcmp(RemotePass, UpLinkPtr->ServPass))
                 {  /* If the passwords dont match.. */
                    Debug(DBGWARNING, "Password for server %s doesnt match. Killing link.", Arg[1]);
                    sendtosock(Sock, ":%s SQUIT %s :Password mismatch", ServiceName, Arg[1]);
                    SockConnected = FALSE;
                    return;
                 }
                 AddServer(Arg[1], ServiceName);
                 NETBURSTING = TRUE;  /* Ignore incomming users for a while*/
                 bantimer = time(NULL);  /* Start ban timer over */
#ifndef HYBRIDTS4
                 NetOffset = atoi(Arg[4]) - time(NULL);  /* Get the network timestamp from the server */
#endif
                 ServerConnected = TRUE;  /* We now have a good link.. */
                 safecopy(Uplink, Arg[1], MAXSERVERLEN);
                 if(!SentBurst) /* If we were listning, we must now return the burst.. */
                 {
                    if(!Authenticated)
                      Authenticate(UpLinkPtr);
//                    SentBurst = TRUE; /* Don't re-send burst unless we loose the connection. */
                 }
                 JoinChans();
#ifdef HYBRIDTS4
                 AddUser(IrcCID, IrcNick, IrcNick, "X2", ServiceName,
                         ServiceName, (UMODEk & UMODEo),
                         NULL, NULL);
                 /* As of yet we havnt decided if afternet hybrid will use K or k for no kill/kick/deop */
                 ProcessUserMode(IrcNick, "+kows");
#else
                 AddUser(IrcNick, IrcNick, "X2", ServiceName,
                         ServiceName, (UMODEk & UMODEo), 
                         NULL, NULL);
                 ProcessUserMode(IrcNick, "+kows");
#endif /* HYBRIDTS4 */
                 return(0);
             }
        }
        Debug(DBGWARNING, "Warning server %s is not in our list. Killing link.", Arg[1]);
        sendtosock(Sock, ":%s SQUIT %s :No C/N conf lines", ServiceName, Arg[1]);
        SockConnected = FALSE;
        sleep(4); /* Lets not flood the server with connect requests TOO fast ;) */
        return(0); /* Servername not in our list */
      }
      else if (!strcasecmp(Arg[0], "SQUIT"))
      {   /* Server quit: Clear all the users and server of this
           * and everyone behind it. */
          /* I think SQUIT's generally happy as Arg[1] so this shouldnt be used.. */
        RemoveServer(Arg[1]);
        if (AUTOROUTING)
          Reroute();
        return;
      }
  }
  else   /* Arg[0][1] is equal to ':' so its a msg from server or user*/
  {
      Arg[0]++; /* get rid of the : */
/* PING */
      if(!strcasecmp(Arg[1], "PING"))
      {
         Arg[2] = (char *) strtok(NULL, " ");
         if(Arg[2])
           sendtosock(Sock, ":%s PONG %s :%s", ServiceName, ServiceName, Arg[2]);
         return(0);
      }
/* JOIN */
      if(!strcasecmp(Arg[1], "JOIN"))
      {
        Arg[2] = (char *) strtok(NULL, " ");
        if(!Arg[2])
        {
           Debug(DBGWARNING, "Invalid JOIN syntax.");
           return(0);
        }
        /* Need to check for comma seperated channels */
        for(i=strlen(Arg[2])+1;i>=0;i--)
        {/* move right to left.. call ProcessJoin when we hit a comma or beginning */
           if(Arg[2][i] == ',')
           {
#ifdef HYBRIDTS4
             ProcessJoin(Arg[0], &Arg[2][i+1], "");
#else
             ProcessJoin(Arg[0], &Arg[2][i+1]);
#endif
             Arg[2][i] = '\0';
           }
        }
#ifdef HYBRIDTS4
        ProcessJoin(Arg[0], Arg[2], "");
#else
        ProcessJoin(Arg[0], Arg[2]);
#endif
        return(0);
      }
/* PART */
      else if(!strcasecmp(Arg[1], "PART"))
      {
        Arg[2] = (char *) strtok(NULL, " ");
        if(!Arg[2])
        {
           Debug(DBGWARNING, "Invalid PART syntax.");
           return(0);
        }
        /* Need to check for comma seperated channels */
        for(i=strlen(Arg[2])+1;i>=0;i--)
        {/* move right to left.. call ProcessPart when we hit a comma or beginning */
           if(Arg[2][i] == ',')
           {
#ifdef HYBRIDTS4
             ProcessPart((char *) NickFromCID(Arg[0]),&Arg[2][i+1]);
#else
             ProcessPart(Arg[0],&Arg[2][i+1]);
#endif
             Arg[2][i] = '\0';
           }
        }
#ifdef HYBRIDTS4
        ProcessPart((char *)NickFromCID(Arg[0]),Arg[2]);
#else
        ProcessPart(Arg[0],Arg[2]);
#endif
        return(0);
      }
/* MODE */
      else if (!strcasecmp(Arg[1], "MODE"))
      {
         Arg[2] = (char *) strtok(NULL, " ");   /* nick/channel */
         Arg[3] = (char *) strtok(NULL, " ");   /* +/- mode   */
         if (Arg[2][0] == '#')
         {
             Arg[4] = (char *) strtok(NULL, " ");
             Arg[5] = (char *) strtok(NULL, "\n\0");
             snprintf(rest, MAXLEN, "%s %s %s %s", Arg[2], Arg[3], Arg[4], Arg[5]);
                        /* "channel +s+nkl stuff stuff" */
#ifdef HYBRIDTS4
             on_mode((char *)NickFromCID(Arg[0]), rest);
#else
             on_mode(Arg[0], rest);
#endif /* HYBRIDTS4 */
         }
         else
             ProcessUserMode(Arg[2], &Arg[3][1]);
         return (0);
      }
/* PONG */
      else if (!strcmp(Arg[1], "PONG"))
      {
        GotPong = true;
        PongStamp = time(NULL);  /* does it matter who sent us the pong? */
        return (0);
      }
/* TOPIC */
      else if (!strcmp(Arg[1], "TOPIC"))
      {
        return (0);			/* Dont need to track topic, yet... */
      }
/* AWAY */
      else if (!strcmp(Arg[1], "AWAY"))
      {
        Arg[2] = (char *)strtok(NULL, " ");  /* :Reason */
#ifdef HYBRIDTS4
        if(Arg[2])
          SetAway((char *)NickFromCID(Arg[0]));
        else
          ClearAway((char *)NickFromCID(Arg[0]));
#else  /* HYBRIDTS4 */
        if(Arg[2])
          SetAway(Arg[0]);
        else
          ClearAway(Arg[0]);
#endif  /* HYBRIDTS4 */
        return(0);
      }
/* INVITE */
      else if (!strcmp(Arg[1], "INVITE"))
      {
        return(0);
      }
/* NOTICE */
      else if (!strcmp(Arg[1], "NOTICE"))
      {
        return (0);			/* Perhapse some day use this..? */
      }
/* SETTIME */
      else if (!strcmp(Arg[1], "SETTIME"))
      {
        Arg[2] = (char *) strtok(NULL, " ");
        if(Arg[2])
        {
           NetOffset = atoi(Arg[2]) - time(NULL);
        }
        return(0);
      }
/* INFO */
      else if (!strcmp(Arg[1], "INFO"))
      {
        return(0);
      }
/* ADMIN */
      else if (!strcmp(Arg[1], "ADMIN"))
      {
        return(0);
      }
/* CONNECT */
      else if (!strcmp(Arg[1], "CONNECT"))
      {
        return(0);
      }  
/* RUBIN */
      else if (!strcmp(Arg[1], "RUBIN"))
      {
        return(0);
      }
/* TIME */
      else if (!strcmp(Arg[1], "TIME"))
      {
        return(0);
      }
/* TRACE */
      else if (!strcmp(Arg[1], "TRACE"))
      {
        return(0);
      }
/* MOTD */
      else if (!strcmp(Arg[1], "MOTD"))
      {
        return(0);
      }
/* RPING */
      else if (!strcmp(Arg[1], "RPING"))
      {
        return(0);
      }
/* SILENCE */
      else if (!strcmp(Arg[1], "SILENCE"))
      {
         return(0);
      }
/* UPING */
      else if (!strcmp(Arg[1], "UPING"))
      {
        return(0);
      }
/* VERSION */
      else if (!strcmp(Arg[1], "VERSION"))
      {
        return(0);
      }
/* WALLOPS */
      else if (!strcmp(Arg[1], "WALLOPS"))
      {
        return(0);
      }
/* WHOIS */
      else if (!strcmp(Arg[1], "WHOIS"))
      {
#ifdef HYBRIDTS4
        Arg[0] = (char *) NickFromCID(Arg[0]);
#else
        Arg[0] = Arg[0];  /* ???? why.. */
#endif
        sendtosock(Sock, ":%s 311 %s %s %s %s * :%c) /MSG %s HELP (%c",
    	       ServiceName, Arg[0], IrcNick, IrcNick, ServiceName, CommandMark, IrcNick, CommandMark);
        sendtosock(Sock, ":%s 312 %s %s %s :%s",
	       ServiceName, Arg[0], IrcNick, ServiceName, VERSION);
        sendtosock(Sock, ":%s 313 %s %s :is an AfterNet IRC Operator\n",
	       ServiceName, Arg[0], IrcNick);
        sendtosock(Sock, ":%s 317 %s %s 0 %d :seconds idle, signon time",
	       ServiceName, Arg[0], IrcNick, ConnectTime);
        sendtosock(Sock, ":%s 318 %s %s :End of /WHOIS list.",
	       ServiceName, Arg[0], IrcNick);
        return (0);
      }
/* STATS */
      else if (!strcmp(Arg[1], "STATS"))
      {
        return (0);
      }
/* GLINE */
      else if (!strcmp(Arg[1], "GLINE"))
      {
        Arg[2] = (char *) strtok(NULL, " ");	/* Target (*) */
        Arg[3] = (char *) strtok(NULL, " ");	/* +/-*@* /*
    						   Arg[4] = (char *) strtok(NULL, " "); /* Time */
        Arg[5] = (char *) strtok(NULL, " ");	/* time? */
        Arg[6] = (char *) strtok(NULL, "\0");	/* Reason */
        if (Arg[3][0] == '+')
        {
          static time_t gwritets=0;
          if (!Arg[4])
        	return (0);		/* Invalid no time.. */
          ProcessAddGline(&Arg[3][1], atoi(Arg[5]), Arg[6]);
          if(time(NULL) > gwritets + 60)
          {
             WriteGlineDat();
             gwritets=time(NULL);
          }
        }
        else if (Arg[3][0] == '-')
          ProcessRemoveGline(&Arg[3][1]);
        return (0);
      }
#ifdef HYBRIDTS4
      else if (!strcmp(Arg[1], "CLIENT"))
      { 
         Arg[2] = (char *) strtok(NULL, " ");       /* ConnID */
         Arg[3] = (char *) strtok(NULL, " ");       /* TimeStamp */
         Arg[4] = (char *) strtok(NULL, " ");       /* Real Nick */
         Arg[5] = (char *) strtok(NULL, " ");       /* Unknown integer (1) */
         Arg[6] = (char *) strtok(NULL, " ");       /* Modes (+io) */
         Arg[7] = (char *) strtok(NULL, " ");       /* UserID */
         Arg[8] = (char *) strtok(NULL, " ");       /* Host */
         Arg[9] = (char *) strtok(NULL, "\n");      /* RealName */
         if(Arg[9] && *Arg[9])
             Arg[9]++;
         AddUser(Arg[2], Arg[4], Arg[7], Arg[9], Arg[8], Arg[0], 0, NULL, NULL);
         ProcessUserMode(Arg[4], Arg[6]);
         return (0);
      }
      else if (!strcmp(Arg[1], "SJOIN"))
      {
        char modestr[MAXLEN], modetargets[MAXLEN], cidtmp[MAXLEN]; 
        int j;

        Arg[2] = (char *) strtok(NULL, " "); /* Timestamp */
        Arg[3] = (char *) strtok(NULL, " "); /* Channel */
        Arg[4] = (char *) strtok(NULL, ":"); /* Modes */
        Arg[5] = (char *) strtok(NULL, "\n"); /* All the users (@.cid @.cid @.cid) */
        if(!Arg[5])
        {
           Debug(DBGWARNING, "Invalid SJOIN syntax. (no 5th argument)");
           return(0);
        }
        /* If there are modes, process them */
	strcpy(modestr, "");
	Arg[4] = (char *) strtok(Arg[4], " "); /* Remove any tailing space */
        if(strcmp(Arg[4], "0"))                /* If its not just a 0, */
	{
		snprintf(rest, MAXLEN, "%s %s", Arg[3], Arg[4]);
		on_mode(Arg[0], rest);
	}
        if(Arg[5][strlen(Arg[5]) - 1] == ' ')
           Arg[5][strlen(Arg[5]) - 1] = '\0';  /* Clip tailing space */
        Debug(DBGINFO, "Processing SJOIN for users [%s] (Arg[0]=%s", Arg[5], Arg[0]);
        for(i=strlen(Arg[5])-1;i>=0;i--) /* Walk word by word though nicks who joined */
        {
              if((Arg[5][i-1] == ' ') || (Arg[5][i-1] == '\0'))    /* If theres a space, stop */
	      {							   /* and deal with the nick  */
                      strcpy(modetargets, "");
                      strcpy(modestr, "");

		      if(isalpha(Arg[5][i]))  /* If its a ban or exclusion */
		      {
                          Debug(DBGINFO, "Found a Ban/ex: %s", &Arg[5][i]);
			     /* Add ban and e mode arsing here.. */ 
		      }
		      else  /* Its a @ % + or . */
		      {
			  Debug(DBGINFO, "Found a user: %s", &Arg[5][i]);
                          for(j=0;;j++)
		          {                /* Copy the @%+ crap off the front */
				  if(Arg[5][i+j] == '.')
				  {
					  break;
				  }
				  switch(Arg[5][i+j])
				  {
					  case '@':
						  strcat(modestr, "o");
						  break;
					  case '%':
						  strcat(modestr, "h");
						  break;
					  case '+':
						  strcat(modestr, "v");
						  break;
					  default:
						  Debug(DBGWARNING, "Unknown SJOIN symbol [%c]", Arg[5][i+j]);
				  }
		          }
			  strcpy(cidtmp, &Arg[5][i+j]);  /* copy the CID with '.' */
			  if(*modestr)   /* If theres stuff in the modes */
			  {
                               for(j=strlen(modestr);j>0;j--)
			       {
				       strcat(modetargets, (char *) NickFromCID(cidtmp));
				       strcat(modetargets, " ");
			       }
			       snprintf(rest, MAXLEN, "%s +%s %s", Arg[3], modestr, modetargets);
			       Debug(DBGINFO, "cidtmp=%s rest=%s", cidtmp, rest);
			       ProcessJoin((char *)NickFromCID(cidtmp), Arg[3], modestr);
			       on_mode(Arg[0], rest); /* Send the mode to the mode parser.. */
			  }
			  else
			  {
				  ProcessJoin((char *)NickFromCID(cidtmp), Arg[3], modestr);
			  }
		      }
		      Arg[5][i-1] = '\0';
              }
        }
        return(0);
      }
#endif
      else if (!strcmp(Arg[1], "SERVER"))
      {
         Arg[2] = (char *) strtok(NULL, " ");   /*   link */
         if(!Arg[2])
         {
           Debug(DBGWARNING, "Incorrect syntax in SERVER line");
           return(0);
         }
         AddServer(Arg[2], Arg[0]);
         return (0);
      }
      else if (!strcmp(Arg[1], "NICK"))
      {
         Arg[2] = (char *) strtok(NULL, " ");   /* nick */
         Arg[3] = (char *) strtok(NULL, " ");
         Arg[4] = (char *) strtok(NULL, " ");
         if (Arg[4] == NULL)
         {
#ifdef HYBRIDTS4
             ProcessNickChange((char *)NickFromCID(Arg[0]), Arg[2]);
#else
             ProcessNickChange(Arg[0], Arg[2]);
#endif
             return (0);
         }
         Arg[5] = (char *) strtok(NULL, " ");   /* userID */
         Arg[6] = (char *) strtok(NULL, " ");   /* address */
         Arg[7] = (char *) strtok(NULL, " ");  /* server */
         Arg[8] = (char *) strtok(NULL, "\n");  /* :RealName */
         if(Arg[8] && *Arg[8])
            Arg[8]++;
         TotalUserCount++;
         if(!(Arg[6]))
         {
             Debug(DBGWARNING, "Incorrect syntax in NICK line");
             return(0);
         }
#ifndef HYBRIDTS4
         AddUser(Arg[2], Arg[5], Arg[6], Arg[8], Arg[0], NOMODES, NULL, NULL);
         if (!NETBURSTING)
           SendMOTD(Arg[2]);
         sendtosock(Sock, ":%s USERIP %s", IrcNick, Arg[2]);
#endif /* HYBRIDTS4 */
         return (0);
      }
      else if (!strcmp(Arg[1], "KICK"))
      {
         Arg[2] = (char *) strtok(NULL, " ");
         Arg[3] = (char *) strtok(NULL, " ");
         if(!Arg[3])
         {
            Debug(DBGWARNING, "Incorrect KICK syntax in line.");
            return(0);
         }
#ifdef HYBRIDTS4
         ProcessKick((char *) NickFromCID(Arg[0]), Arg[3], Arg[2]);
#else /* HYBRIDTS4 */
         ProcessKick(Arg[0], Arg[3], Arg[2]);
#endif /* HYBRIDTS4 */
         return (0);
      }
      else if (!strcmp(Arg[1], "QUIT"))
      {
#ifdef HYBRIDTS4
         DeleteUser((char *) NickFromCID(Arg[0]), false, true);
#else
         DeleteUser(Arg[0], false, true);
#endif
         return (0);
      }
      else if (!strcmp(Arg[1], "KILL"))
      {
         if (Arg[2] = (char *) strtok(NULL, " "))
            DeleteUser(Arg[2], false, true);
         else
            Debug(DBGWARNING, "Incorrect syntax for KILL in line");
         return (0);
      }
      else if (!strcmp(Arg[1], "SQUIT"))
      {
         ServerListType ServerPtr;

         Arg[2] = (char *) strtok(NULL, " "); /* Server name */
         Arg[3] = (char *) strtok(NULL, " "); /* Time stamp */
         Arg[4] = (char *) strtok(NULL, "\n"); /* Reason */  
/*         Debug(DBGERROR, "DEBUG: Processing SQUIT server: %s %s %s", Arg[2], Arg[3], Arg[4]);   
*/
         if (!(Arg[2]))
         {
             Debug(DBGWARNING, "Incorrect syntax for SQUIT in line");
             return (0); 
         }
         if (!strcasecmp(Arg[2], Uplink))
         {
           /* If its us, handle it.. */
           Debug(DBGWARNING, "We were SQUIT, killing link..");
           SockConnected = FALSE;
           return(0);
         }
        if (PINGOUTCONNECT)
        {
             if(!strcasecmp(Arg[4], ":Ping Timeout"))
             {
                ServerPtr = GetServer(Arg[2]);
                if(!ServerPtr)
                {
                   Debug(DBGWARNING, "Cant find server %s to squit!", Arg[2]);
                   return;
                }
/*                Debug(DBGERROR, "Sending connect, server %s pinged off of %s", 
                               ServerPtr->Name, ServerPtr->LinkName);
*/
                sendtosock(Sock, ":%s CONNECT %s 8888 %s", 
                               IrcNick, ServerPtr->Name, ServerPtr->LinkName);
             }
        }  
         RemoveServer(Arg[2]);
         if (AUTOROUTING)
           Reroute();
         return (0);
      }
      else if (!strcmp(Arg[1], "PRIVMSG"))
      {
         Arg[2] = (char *) strtok(NULL, " ");
         Arg[3] = (char *) strtok(NULL, "\0\r\n");
         if(!Arg[3] || !Arg[2])
         {
#ifdef HYBRIDTS4
           Debug(DBGWARNING, "Empty message from %s (%s).", Arg[0], 
                                 (char *)NickFromCID(Arg[0]));
#else
           Debug(DBGWARNING, "Empty message from %s.", Arg[0]);
#endif
           return(0);
         }
         /*        *from*   *to+rest* */
#ifdef HYBRIDTS4
         ProcessMsg((char *) NickFromCID(Arg[0]), Arg[2], Arg[3]);
#else
         ProcessMsg(Arg[0], Arg[2], Arg[3]);
#endif
         return(0);
      }
/* Check for numerics */
  num = atoi(Arg[1]);
  if (num > 0)
  {
    switch (num)
    {
    case 1:			/* welcome */
    case 2:			/* Your host.. */
    case 3:			/* Created... */
    case 4:			/* MYINFO */
    case 5:			/* Map reply */
    case 6:			/* Map More */
    case 7:			/* Map End */
    case 8:			/* Server Notice Mask */
    case 9:			/* StatMem Total */
    case 10:			/* StatMem */

    case 200:			/* Trace Link */
    case 201:			/* Trace Connecting.. */
    case 202:			/* Trace Handshake */
    case 203:			/* Trace Unknown */
    case 204:			/* Trace oper */
    case 205:			/* Trace User */
    case 206:			/* Trace Server */
    case 208:			/* TraceNewType */
    case 209:			/* Trace Class */
    case 211:			/* Stats Link Info */
    case 212:			/* Stats commands */
    case 213:			/* Stats C line */
    case 214:			/* Stats N line */
    case 215:			/* Stats I line */
    case 216:			/* Stats K line */
    case 217:			/* Stats P line */
    case 218:			/* Stats Y line */
    case 219:			/* end of Stats */
    case 220:			/* NULL */
    case 221:			/* UMODEIS */
    case 231:			/* Service Info */
    case 232:			/* End of Service Info */
    case 233:			/* Service */
    case 234:			/* Serv list */
    case 235:			/* Serv list End */
    case 241:			/* Stats L line */
    case 242:			/* Stats Uptime */
    case 243:			/* Stats O line */
    case 244:			/* Stats H line */
    case 246:			/* Stats T lien */
    case 247:			/* Stats G line */
    case 248:			/* Stats U line */
    case 250:			/* Stats Conn */
    case 251:			/* LuserClient */
    case 252:			/* Luser Op */
    case 253:			/* Luser Unknown */
    case 254:			/* Luser Channels */
    case 255:			/* Luser Me */
    case 256:			/* Admin */
    case 257:			/* AdminLOC1 */
    case 258:			/* AdminLOC2 */
    case 259:			/* Admin Email */
    case 261:			/* TraceLog */
    case 262:			/* TracePing */
    case 271:			/* Silent list */
    case 272:			/* End of silent list */
    case 275:			/* Stats D lien */
    case 280:			/* Glist */
    case 281:			/* End of G list */
    case 283:			/* QUAR Nick */
      return (0);

    case 301:			/* away reply */
         /* Ok, so now we send "WHOIS nick nick" when we see people oper */
         /* Then here we see them away .. so after a net join .IRCOPS */
         /* knows who'se here and who'se not */
      Arg[2] = (char *) strtok(NULL, " "); /* me */
      Arg[3] = (char *) strtok(NULL, " "); /* Target */
#ifdef HYBRIDTS4
      SetAway((char *)NickFromCID(Arg[3]));  /* This is untested. I don't know if this param is a CID or not */
#else
      SetAway(Arg[3]);
#endif
      return (0);
    case 302:			/* UserHost reply */
    case 303:			/* ISon reply */
    case 304:			/* RPL_TEXT */
    case 305:			/* No longer away */
    case 306:			/* Now marked as away */
      return (0);
    case 307:			/* userip reply */
      Arg[2] = (char *) strtok(NULL, " ");		/* Dest nick (me) */
      Arg[3] = (char *) strtok(NULL, " ");		/* nick[user@host] */
      if (!Arg[3])
        return (0);
      Arg[4] = (char *) strtok(Arg[3], "*=+");
      Arg[5] = (char *) strtok(NULL, "@");
      Arg[6] = (char *) strtok(NULL, "\0\r\n");
      if(Arg[4] && Arg[6])
      {
        if(!SuperGlineCheck(++Arg[4], Arg[6]))
          BeginGateScan(Arg[4], Arg[6]);
      }
      return (0);
    case 311:			/* WHOIS USER */
    case 312:			/* whois server */
    case 313:			/* whois oper */
    case 314:			/* whowas user */
    case 315:			/* End of WHO */
    case 316:			/* NULL */
    case 317:			/* whois Idle */
    case 318:			/* End of whois */
    case 319:			/* whois chans */
    case 320:			/* NULL */
    case 321:			/* List Start */
    case 322:			/* RPL_LIST */
    case 323:			/* List end */
      return (0);
    case 324:			/* Channel Mode Is... */
      Arg[2] = (char *) strtok(NULL, " ");
      Arg[3] = (char *) strtok(NULL, " ");
      Arg[4] = (char *) strtok(NULL, "\r\n");
      sendtosock(Sock, ":%s MODE %s -%s", IrcNick, Arg[3], &Arg[4][1]);
      return (0);
    case 325:                   /* Unknown */
    case 326:                   /* Unknown */
    case 327:                   /* Unknown */
    case 328:                   /* DalNet Channel URL */
    case 329:			/* Creation Time */
    case 331:			/* No Topic */
    case 332:			/* RPL_TOPIC */
    case 333:			/* Topic who time */
    case 334:			/* List Usage */
    case 341:			/* RPL_Inviting */
    case 351:			/* Version reply */
    case 352:			/* Who reply */
    case 353:			/* Name Reply */
    case 354:			/* Whospcrpl */
    case 361:			/* KILL DONE */
    case 362:			/* CLOSING */
    case 363:			/* CloseEnd */
    case 364:			/* LINKS */
    case 365:			/* End of links */
    case 366:			/* End of Names */
    case 367:			/* Banlist */
    case 368:			/* End of Banlist */
    case 369:			/* End of WhoWas */
    case 371:			/* INFO */
    case 372:			/* MOTD */
    case 373:			/* InfoStart */
    case 374:			/* End of Info */
    case 375:			/* MOTD Start */
    case 376:			/* End of MOTD */
    case 381:			/* Your nowan ircop */
    case 382:			/* Rehashing */
    case 384:			/* My Port Is */
    case 385:			/* Not Oper anymore */
    case 391:			/* Time reply */
      return (0);


    case 401:			/* No Such Nick */
    case 402:			/* No such Server */
    case 403:			/* No such channel */
      return (0);
    case 404:
      Arg[2] = (char *) strtok(NULL, " ");
      Arg[3] = (char *) strtok(NULL, " ");
      Debug(DBGWARNING, "Error 404: Cannot send to channel %s", Arg[3]);
      return (0);
    case 405:			/* Too many Channels */
    case 406:			/* Was no such nick */
    case 407:			/* TooManyTargets */
    case 408:			/* NULL */
    case 409:			/* No Origin */
    case 410:			/* NULL */
    case 411:			/* NO Recipient */
    case 412:			/* No text to send */
    case 413:			/* No top level domain */
    case 414:			/* Wildcard in toplevel domain */
    case 415:			/* NULL */
    case 416:			/* Query too long */
    case 417:			/* NULL */
    case 421:			/* Unknown command */
    case 422:			/* No MOTD */
    case 423:			/* No Admin Info */
    case 424:			/* NULL */
    case 431:			/* No Nick Given */
    case 432:			/* Erroneus nick name */
    case 433:			/* Nick name in use */
    case 434:			/* NULL */
    case 435:			/* NULL */
    case 436:			/* Nick Collision */
    case 437:			/* Baned user nick change error */
    case 438:			/* Nick change to fast */
    case 439:			/* Target changed to fast */
    case 440:			/* NULL */
    case 441:			/* User Not In Channel */
    case 442:			/* You arnt on channel */
    case 443:			/* Already in channel */
    case 444:			/* NULL */
    case 451:			/* Not registered */
    case 452:			/* NULL */
    case 461:			/* Need more params */
    case 462:			/* Already registered */
    case 463:			/* No permmision for host */
    case 464:			/* Pass mismatch */
    case 465:			/* Your banned creep */
    case 466:			/* You will be banned */
    case 467:			/* Chan key already set */
    case 468:			/* Invalid username */
    case 469:			/* NULL */
    case 470:			/* NULL */
    case 471:			/* Chan full */
    case 472:			/* Unknown mode */
    case 473:			/* Invite only chan */
    case 474:			/* Banned from chan */
    case 475:			/* Bad key */
    case 476:			/* Bad Channel Mask */
    case 477:			/* Chan does not support modes */
    case 478:			/* Banlist full */
    case 479:			/* NULL */
    case 480:			/* NULL */
    case 481:			/* No Privlidges (not oper) */
    case 482:			/* Your not channel op */
    case 483:			/* Cant kill a server stupid */
    case 484:			/* cant kill/kick/deop chan service (+k) */
    case 485:			/* NULL */
    case 491:			/* NO Olines for your host */
    case 501:			/* Unknown Mode flag */
    case 502:			/* Cant change other users modes */
    case 511:			/* Silent list full */
    case 512:			/* No such gline */
    case 513:			/* Bad Ping */
    case 514:			/* Cant kill ircop (afternet) */
      return (0);
    default:
      return (0);
    }
  }
 }
  Arg[2] = (char *) strtok(NULL, "\0\r\n");
  if(!Arg[2])
    Arg[2] = "";
  Debug(DBGWARNING, "I dont know how to parse: %s %s %s", Arg[0], Arg[1], Arg[2]);
  return(0);
}


/*
 * Gets data from the socket.. doesnt care what it is. 
 * returns the number of bytes gotten.. the data is put in Buffer
 */
int ReadSocket(int TheSocket, char *Buffer)
{
   int c;
   int i;
   static char remainder[MAXRECVLEN+10] = "";
   char LineString[MAXRECVLEN+10];
   int rsize;
   char *newline;
   char *doneline;
   
   memset(Buffer, '\0', MAXRECVLEN+1);
   rsize = 0;
   if(*remainder)
   {
      rsize = strlen(remainder);
      memcpy(Buffer, remainder, rsize);
   }
   c = recv(TheSocket, &Buffer[rsize], MAXRECVLEN-rsize, 0);
   if(c <= 0)
   {
       Debug(DBGWARNING, "0 bytes from socket. Killing..");
       remainder[0] = '\0';
       SockConnected = FALSE;
       return(0);
   }
   newline = Buffer;
   c = c+rsize;
   Buffer[c] = '\0';
   for(i=0;i<c;i++)
   {
      if(Buffer[i] == '\n')
      {
         Buffer[i] = '\0';
         memcpy(LineString, newline, strlen(newline)+1);
         ProcessLine(LineString);
         doneline = newline;
         newline = &Buffer[i+1]; 
      }
      if(Buffer[i] == '\r')
         Buffer[i] = '\0';
   }
   if(*newline)
   {
      strcpy(remainder, newline);
   }
   else
   {
      memset(remainder, '\0', MAXRECVLEN);
   }
}

