
/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "sock.h"
#include "misc.h"
#include "server.h"
#include "user.h"


void CheckGlineTimeouts()
{
  GlineListType *gptr, *gtmp;

  if (!(gptr = Glines))
  {
    return;			/* No glines */
  }
  while (gptr)
  {
    gtmp = gptr->Next;
    if (gptr->time < NetTime())
    {
#ifdef debuging
      sendtosock(Sock, ":%s WALLOPS :\002Removing expired gline [%s]\002", IrcNick, gptr->address);
#endif
      ProcessRemoveGline(gptr->address);
    }
    gptr = gtmp;
  }
}

void CheckSuperGlineTimeouts()
{
  GlineListType *gptr, *gtmp;

  if (!(gptr = SuperGlines))
  {
    return;			/* No glines */
  }
  while (gptr)
  {
    gtmp = gptr->Next;
    if (gptr->time < NetTime())
    {
#ifdef debuging
      sendtosock(Sock, ":%s WALLOPS :\002Removing expired supergline [%s]\002", IrcNick, gptr->address);
#endif
      ProcessRemoveSGline(gptr->address);
    }
    gptr = gtmp;
  }
}

int SuperGlineCheck(char *Nick, char *IP)
{
   UserListType UserPtr;
   char uhr[MAXLEN];
   GlineListType *gptr;
   char Address[MAXLEN];
   char reason[MAXLEN];

   if(!(UserPtr = FindNick(Nick, UsersN)))
     return(1);
   sprintf(uhr, "%s@%s$%s", UserPtr->UserID, UserPtr->Address, UserPtr->RealName);
   
   gptr = SuperGlines;
   if(!gptr)
     return(FALSE);
   while(gptr)
   {
     if (!matches(gptr->address, uhr))                                           
     {                                                                                   
          sprintf(Address, "*@%s", IP);
          sprintf(reason, "Added automatically from SuperGline(%s)", gptr->reason);
          sendtosock(Sock, ":%s GLINE * +%s 3600 :%s", ServiceName, Address, reason);
          ProcessAddGline(Address, 3600, reason);
          return(TRUE);
     }                                                                                   
     gptr = gptr->Next;
   }
   return(FALSE);
}

GlineListType *GetGline(char *Address)
{
  GlineListType *gptr;

  gptr = Glines;
  if (!gptr)
    return (NULL);
  while (gptr)
  {
    if (!strcasecmp(Address, gptr->address))
    {
      return (gptr);
    }
    gptr = gptr->Next;
  }
  return (NULL);
}

GlineListType *GetSuperGline(char *Address)
{
  GlineListType *gptr;

  gptr = SuperGlines;
  if (!gptr)
    return (NULL);
  while (gptr)
  {
    if (!strcasecmp(Address, gptr->address))
    {
      return (gptr);
    }
    gptr = gptr->Next;
  }
  return (NULL);
}


void ProcessAddGline(char *Address, time_t gtime, char *Reason)
{
  GlineListType *gptr;
  char Addr[MAXLEN];

  if(!strchr(Address, '@'))
    strcpy(Addr, "*@");
  else
    strcpy(Addr, "");
  strncat(Addr, Address, MAXLEN-1);
  if (gptr = GetGline(Addr))
  {
    Debug(DBGNOTICE, "Gline already in my list, updating... ");
    gptr->time = NetTime() + gtime;
    return;
  }
  if (gptr = malloc(sizeof(GlineListType)))
  {
    safecopy(gptr->address, Addr, MAXLEN);
    safecopy(gptr->reason, Reason, MAXLEN);
    gptr->time = (NetTime() + gtime);
    gptr->Next = Glines;
    Glines = gptr;
    return;
  }
  else
  {
    Debug(DBGERROR, "Cant allocate memory for new gline");
    return;
  }
}

void ProcessAddSGline(char *Address, time_t gtime, char *Reason)
{
  GlineListType *gptr;
  char Addr[MAXLEN];
  char *tmp;

  safecopy(Addr, Address, MAXLEN);
  if (gptr = GetSuperGline(Address))
  {
    Debug(DBGNOTICE, "SuperGline already in my supergline list, updating... ");
    gptr->time = NetTime() + gtime;
    return;
  }
  if (gptr = malloc(sizeof(GlineListType)))
  {
    safecopy(gptr->address, Address, MAXLEN);
    safecopy(gptr->reason, Reason, MAXLEN);
    gptr->time = (NetTime() + gtime);
    gptr->Next = SuperGlines;
    SuperGlines = gptr;
    return;
  }
  else
  {
    Debug(DBGERROR, "Cant allocate memory for new supergline");
    return;
  }
}

void ProcessRemoveGline(char *Address)
{
  GlineListType *gptr, *gtmp, *glast;

  gptr = Glines;
  glast = NULL;
  while (gptr)
  {
    if (!strcasecmp(gptr->address, Address))
    {
      if (glast)
	glast->Next = gptr->Next;
      else
	Glines = gptr->Next;
      free(gptr);
      WriteGlineDat();
      return;
    }
    glast = gptr;
    gptr = gptr->Next;
  }
}

void ProcessRemoveSGline(char *Address)
{
  GlineListType *gptr, *gtmp, *glast;

  gptr = SuperGlines;
  glast = NULL;
  while (gptr)
  {
    if (!strcasecmp(gptr->address, Address))
    {
      if (glast)
	glast->Next = gptr->Next;
      else
	SuperGlines = gptr->Next;
      free(gptr);
      WriteSGlineDat();
      return;
    }
    glast = gptr;
    gptr = gptr->Next;
  }
}

ServerListType GetServer(char *Name)
{
  int Pos;
  unsigned long Key;
  char Str17[18];
  ServerListType TempPtr;
  char *UpCaseStr;
  char *UpCaseName;

  UpCaseName = UpCase(Name);
  GetStr17(Str17, UpCaseName);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % SERVERSEGSIZE;

  TempPtr = Servers->Table[Pos];
  while (TempPtr)
  {
    if (!strcmp(UpCaseStr = UpCase(TempPtr->Name), UpCaseName))
    {
      free(UpCaseStr);
      free(UpCaseName);
      return (TempPtr);
    }
    free(UpCaseStr);
    TempPtr = TempPtr->Next;
  }
  free(UpCaseName);
  return (NULL);
}

int AddServer(char *ServName, char *ServLink)
{
  ServerListType NewServ, ServPtr;
  unsigned long Key;
  int Pos;
  char Str17[18];
  char *UpCaseStr;
  extern time_t bantimer;

  if(NETBURSTING == FALSE)
  {
    Debug(DBGINFO, "Disabling info lines for a while due to netburst..");
    NETBURSTING = TRUE;
    bantimer = time(NULL);
  }

  if ((ServerCount * 2) > SERVERSEGSIZE)
  {
    sendtosock(Sock, ":%s WALLOPS :WARNING!!! I WASN'T ABLE TO ADD %s TO THE LIST OF SERVERS! SOMEONE MUST REDEFINE #DEFINE SERVERSEGSIZE %d TO SOMETHING LIKE %d AND THEN RECOMPILE ME! DO IT NOW!", IrcNick, SERVERSEGSIZE, (SERVERSEGSIZE * 2) - 1);
    return (1);
  }

  if (!(NewServ = (ServerListType) malloc(sizeof(struct ServStruct))))
     return (0);

  if (!(NewServ->Name = (char *) malloc(strlen(ServName) + 1)))
    return (0);

  if (!(NewServ->LinkName = (char *) malloc(strlen(ServLink) + 1)))
    return (0);

  strcpy(NewServ->Name, ServName);
  strcpy(NewServ->LinkName, ServLink);
  NewServ->NumOfUsers = 0;
  NewServ->Deleted = FALSE;
  UpCaseStr = UpCase(ServName);
  GetStr17(Str17, UpCaseStr);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % SERVERSEGSIZE;
  free(UpCaseStr);
  NewServ->Next = Servers->Table[Pos];
  Servers->Table[Pos] = NewServ;
  ServerCount++;
  return (1);
}

void RemoveDownLinks(char *Name)
{
  ServerListType ServPtr, TmpPtr;
  int i;

  Debug(DBGINFO, "RemoveDownLinks(%s)", Name);
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      TmpPtr = ServPtr->Next;
      if (0 == strcasecmp(ServPtr->LinkName, Name))
      {
	DeleteServer(ServPtr->Name);
/*	break; */
      }
      ServPtr = TmpPtr;
    }
  }
}

void DeleteServer(char *Name)
{
  ServerListType ServPtr;

  RemoveDownLinks(Name);
  ServPtr = GetServer(Name);
  if (ServPtr)
  {
    ServPtr->Deleted = TRUE;
    Debug(DBGINFO, "Marking %s as deleted!", ServPtr->Name);
  }
}

/*
   This goes though the list and purges any servers marked as deleted
 */
void PurgeDeletedServers()
{
  ServerListType ServPtr, TmpPtr;
  int i;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      TmpPtr = ServPtr->Next;
      if (ServPtr->Deleted == 1)
      {
	Debug(DBGINFO, "Purging deleted Server [%s]", ServPtr->Name);
	PurgeDeletedServer(ServPtr->Name);
      }
      ServPtr = TmpPtr;
    }
  }
}

/*
   This ACTUALY removes the server from the linked list
 */
void PurgeDeletedServer(char *Name)
{
  unsigned long Key;
  int Pos;
  ServerListType TailPtr, CurrPtr;
  char Str17[18];
  char *UpCaseStr;
  char *UpCaseName;

/*  Debug(DBGINFO, "Removing server %s\n", Name);
   RemoveDownLinks(Name);
 */
  DeleteAllUsersByServer(Name);

  TailPtr = NULL;

  UpCaseStr = UpCase(Name);
  GetStr17(Str17, UpCaseStr);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % SERVERSEGSIZE;

  CurrPtr = Servers->Table[Pos];
  while (CurrPtr)
  {
    if (!strcmp(UpCaseName = UpCase(CurrPtr->Name), UpCaseStr))
    {
      free(UpCaseName);
      break;
    }
    else
    {
      free(UpCaseName);
      TailPtr = CurrPtr;
      CurrPtr = CurrPtr->Next;
    }
  }
  if (!CurrPtr)
    return;
  if (!TailPtr)
    Servers->Table[Pos] = CurrPtr->Next;
  else
    TailPtr->Next = CurrPtr->Next;

  free(UpCaseStr);
  free(CurrPtr->Name);
  CurrPtr->Name = NULL;
  free(CurrPtr->LinkName);
  CurrPtr->Name = NULL;
  free(CurrPtr);
  CurrPtr = NULL;
  ServerCount--;
}


/*
   This is the front end to the server (and all servers behind)
   deletion function.  This should be called by parse.c 
 */
void RemoveServer(char *Name)
{
  DeleteServer(Name);
  PurgeDeletedServers();
}

void KillRouts()
{
  int i;
  ServerListType TempRoutPtr;

  for (i = 0; i < ROUTSEGSIZE; i++)
  {
    if (Routs->Table[i])
    {
      TempRoutPtr = Routs->Table[i];
      if (TempRoutPtr)
      {
	if (TempRoutPtr->Name)
	  free(TempRoutPtr->Name);
	if (TempRoutPtr->LinkName)
	  free(TempRoutPtr->LinkName);
	free(TempRoutPtr);
	Routs->Table[i] = NULL;
      }
    }
  }
}

void KillServerList()
{

  int i;
  ServerListType TempPtr, CurrPtr;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    CurrPtr = Servers->Table[i];
    while (CurrPtr)
    {
      TempPtr = CurrPtr;
      CurrPtr = CurrPtr->Next;
      free(TempPtr->Name);
      free(TempPtr->LinkName);
      free(TempPtr);
    }
    Servers->Table[i] = NULL;
  }
  ServerCount = 0;
}

int AddRout(char ServName[], char ServLink[])
{

  ServerListType NewRout;
  char Str17[18];
  int Pos;
  unsigned long Key;
  char *UpCaseName;


  NewRout = (ServerListType) malloc(sizeof(struct ServStruct));

  if (!NewRout)
    return (0);

  NewRout->Name = (char *) malloc(strlen(ServName) + 1);
  if (!NewRout->Name)
    return (0);

  NewRout->LinkName = (char *) malloc(strlen(ServLink) + 1);
  if (!NewRout->LinkName)
    return (0);

  strcpy(NewRout->Name, ServName);
  strcpy(NewRout->LinkName, ServLink);

  RoutCount++;
  UpCaseName = UpCase(ServName);
  GetStr17(Str17, UpCaseName);
  free(UpCaseName);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % ROUTSEGSIZE;

  NewRout->Next = Routs->Table[Pos];
  Routs->Table[Pos] = NewRout;
  return (1);
}

void Reroute()
{
  int i;
  ServerListType TempRoutPtr, TempServPtr;

  RoutsOk = true;
  for (i = 0; i < ROUTSEGSIZE; i++)
  {
    if (Routs->Table[i])
    {
      TempRoutPtr = Routs->Table[i];
      while (TempRoutPtr)
      {
	if (TempServPtr = GetServer(TempRoutPtr->Name))
	{
	  if (strcasecmp(TempRoutPtr->LinkName, TempServPtr->LinkName))
	  {
	    if (!strcasecmp(ServiceName, TempServPtr->LinkName))
	    {
	      TempRoutPtr = TempRoutPtr->Next;
	      continue;
	    }
	    RoutsOk = false;
	    if (!strcasecmp(TempServPtr->Name, Uplink))
	    {
	      sendtosock(Sock, ":%s SQUIT %s %ul :re routing", IrcNick, TempServPtr->LinkName, time(NULL));
	      sendtosock(Sock, ":%s CONNECT %s 6667 %s", IrcNick, Uplink, TempRoutPtr->LinkName);
              RemoveServer(TempServPtr->LinkName);
	    }
	    else
	    {
	      sendtosock(Sock, ":%s SQUIT %s %ul :re routing", IrcNick, TempServPtr->Name, time(NULL));
	      sendtosock(Sock, ":%s CONNECT %s 6667 %s", IrcNick, TempServPtr->Name, TempRoutPtr->LinkName);
              RemoveServer(TempServPtr->Name);
	    }
	  }
	}
	else
	{
	  RoutsOk = false;
	  sendtosock(Sock, ":%s CONNECT %s 6667 %s", IrcNick, TempRoutPtr->Name, TempRoutPtr->LinkName);
	}
	TempRoutPtr = TempRoutPtr->Next;
      }
    }
  }
  if (RoutsOk)
    sendtosock(Sock, ":%s WALLOPS :Routing seems to be ok.", IrcNick);
}
