
/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "sock.h"
#include "misc.h"
#include "chan.h"
#include "server.h"
#include "user.h"

void send_to_user(char *to, char *sendtext,...)
{
  va_list args;
  char buff[1000];
  int duh;

  va_start(args, sendtext);
  vsprintf(buff, sendtext, args);
  va_end(args);
  if (to[0] == '#')
    duh = sendtosock(Sock, ":%s PRIVMSG %s :%s", IrcNick, to, buff);
  else
    duh = sendtosock(Sock, ":%s NOTICE %s :%s", IrcNick, to, buff);
}

UserListType FindUser(char *UserID, char *Address, UserSegment TheSegment)
{

  int Pos;
  unsigned long Key;
  UserListType UserList, TempListPtr;
  char Str17[18];

  GetStr17(Str17, Address);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % USERSEGSIZE;

  if (TheSegment->Table[Pos])
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if ((!(strcmp(TempListPtr->Address, Address))) && (!(strcmp(TempListPtr->UserID, UserID))))
	return (TempListPtr);
      else
	TempListPtr = TempListPtr->NextA;
    }
  }
  return (NULL);
}


void UpdateCloneCount(char *UserID, char *Address, int NewCount, UserSegment TheSegment)
{

  int Pos;
  unsigned long Key;
  char Str17[18];
  UserListType UserList, TempListPtr;

  GetStr17(Str17, Address);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % USERSEGSIZE;

  if (TheSegment->Table[Pos])
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if ((!(strcmp(TempListPtr->UserID, UserID))) && (!(strcmp(TempListPtr->Address, Address))))
	TempListPtr->Clones = NewCount;
      TempListPtr = TempListPtr->NextA;
    }
  }
}

int GetNumOfClones(char *UserID, char *Address)
{

  UserListType UserNode;
  int CloneCount = 0;

  if (!(UserNode = FindUser(UserID, Address, UsersA)))
    return (CloneCount);

  return (UserNode->Clones + 1);

}

#ifdef HYBRIDTS4
char *NickFromCID(char *CID, UserSegment TheSegment)
{
   UserListType UsrPtr;

   if(!(UsrPtr = FindCID(CID, UsersC)))
   {
      Debug(DBGERROR, "CID [%s] not found in memory!", CID);
      return("~ERRNOTFOUND~");
   }
   else
      return(UsrPtr->Nick);
}

UserListType FindCID(char *CID, UserSegment TheSegment)
{
  int Pos;
  unsigned long Key;
  UserListType UserList, TempListPtr;
  char *TheCID;
  char *UpCaseCID;

  UpCaseCID = UpCase(CID);
  Key = strtol(UpCaseCID, (char **) NULL, STRTOLBASE);
  Pos = Key % USERSEGSIZE;
  if (TheSegment->Table[Pos] != NULL)
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if (!(strcmp(TheCID = UpCase(TempListPtr->CID), UpCaseCID)))
      {
        free(UpCaseCID);
        free(TheCID);
        return (TempListPtr);
      }
      else
      {
        free(TheCID);
        TempListPtr = TempListPtr->NextC;
      }
    }
  }
  free(UpCaseCID);
  return (NULL);
}
#endif /* HYBRIDTS4 */

UserListType FindNick(char *Nick, UserSegment TheSegment)
{

  int Pos;
  unsigned long Key;
  UserListType UserList, TempListPtr;
  char *TheNick;
  char *UpCaseNick;

  UpCaseNick = UpCase(Nick);
  Key = strtol(UpCaseNick, (char **) NULL, STRTOLBASE);
  Pos = Key % USERSEGSIZE;
  if (TheSegment->Table[Pos] != NULL)
  {
    TempListPtr = TheSegment->Table[Pos];
    while (TempListPtr)
    {
      if (!(strcmp(TheNick = UpCase(TempListPtr->Nick), UpCaseNick)))
      {
	free(UpCaseNick);
	free(TheNick);
	return (TempListPtr);
      }
      else
      {
	free(TheNick);
	TempListPtr = TempListPtr->NextN;
      }
    }
  }
  free(UpCaseNick);
  return (NULL);
}

boolean IsModeK(char *Nick)
{
  UserListType UserNode;

  if (UserNode = FindNick(Nick, UsersN))
  {
    if (UserNode->Modes & UMODEk || strcasecmp(Nick, IrcNick) == 0)
      return 1;
  }
  return 0;

}

boolean IsOper(char *Nick)
{

  UserListType UserNode;
  unsigned char OperMode;

  if (UserNode = (FindNick(Nick, UsersN)))
    return ((UMODEo & UserNode->Modes) != 0);
  return (FALSE);
}

boolean IsGodOper(char *Nick)
{

  UserListType UserNode;
  unsigned char OperMode;

  if (UserNode = (FindNick(Nick, UsersN)))
    return ((UMODEgod & UserNode->Modes) != 0);
  return (FALSE);
}

boolean IsOperByAdd(char *UserID, char *Address)
{

  int Pos;
  unsigned long Key;
  char Str17[18];
  UserListType UserList, TempListPtr;

  GetStr17(Str17, Address);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  Pos = Key % USERSEGSIZE;

  if (UsersA->Table[Pos])
  {
    TempListPtr = UsersA->Table[Pos];
    while (TempListPtr)
    {
      if ((!(strcmp(TempListPtr->UserID, UserID))) && (!(strcmp(TempListPtr->Address, Address))))
      {
	if ((UMODEo & TempListPtr->Modes) != 0)
	  return (true);
      }
      TempListPtr = TempListPtr->NextA;
    }
  }
  return (false);
}

#ifdef HYBRIDTS4
int AddUser(char *CID, char *Nick, char *UserID, char *Address, char *RealName, char *Serv, unsigned char Modes, OwnerListType Chans, char *Validated)
#else
int AddUser(char *Nick, char *UserID, char *Address, char *RealName, char *Serv, unsigned char Modes, OwnerListType Chans, char *Validated)
#endif /* HYBRIDTS4 */
{

  UserListType NewUser, UserNode;
  ServerListType ServPtr;
  unsigned long Key;
  int PosA, PosN;
#ifdef HYBRIDTS4
  int PosC;
#endif  /* HYBRIDTS4 */
  char Str17[18];
  char CleanUserID[12];
  int CloneCount;
  char *UpCaseNick;
  char buff[MAXLEN], reason[MAXLEN];


  if (FindNick(Nick, UsersN))
  {
    sendtosock(Sock, ":%s WALLOPS :WARNING!!! Nick collision on %s from server %s.  New user NOT added!!", IrcNick, Nick, Serv);
    return;
  }
  if ((UserCount * 2) > USERSEGSIZE)
  {
    sendtosock(Sock, ":%s WALLOPS :WARNING!!! NOT ABLE TO ADD USER TO THE LIST! SOMEONE NEEDS TO INCREASE MY #DEFINE USERSEGSIZE %d TO SOMETHING LIKE %d AND RECOMPILE ME! DO IT NOW!!!", IrcNick, USERSEGSIZE, (USERSEGSIZE * 2) - 1);
    return (1);
  }
  strncpy(CleanUserID, UserID, 11);
  CleanUserID[11] = '\0';
  CloneCount = GetNumOfClones(CleanUserID, Address);
  if (!(NewUser = (UserListType) malloc(sizeof(struct UserStruct))))
     return (0);
  if (!(NewUser->Address = (char *) malloc(strlen(Address) + 1)))
    return (0);
  if (!(NewUser->Serv = (char *) malloc(strlen(Serv) + 1)))
    return (0);
  if (!(NewUser->RealName = (char *) malloc(strlen(RealName) + 1)))                                 
    return (0);
  if (CloneCount > 0)
    UpdateCloneCount(CleanUserID, Address, CloneCount, UsersA);
  strcpy(NewUser->UserID, CleanUserID);
  strcpy(NewUser->Address, Address);
  strcpy(NewUser->Serv, Serv);
  strcpy(NewUser->RealName, RealName);
  strcpy(NewUser->IP, "\0\0");
  UpCaseNick = UpCase(Nick);
  strcpy(NewUser->Nick, Nick);
#ifdef HYBRIDTS4
  strcpy(NewUser->CID, CID);
#endif
  if (Validated)
  {
    if (!(NewUser->Validated = (char *) malloc(strlen(Validated) + 1)))
      return (0);
    strcpy(NewUser->Validated, Validated);
  }
  else
  {
    if (!(NewUser->Validated = (char *) malloc(2)))
      return (0);
    strcpy(NewUser->Validated, "\0");
  }
  NewUser->Modes = Modes;
  NewUser->Clones = CloneCount;
  NewUser->Chans = Chans;
  GetStr17(Str17, Address);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  PosA = Key % USERSEGSIZE;
  Key = strtol(UpCaseNick, (char **) NULL, STRTOLBASE);
  PosN = Key % USERSEGSIZE;
#ifdef HYBRIDTS4
  Key = strtol(CID, (char **) NULL, STRTOLBASE);
  PosC = Key % USERSEGSIZE;
#endif /* HYBRIDTS4 */
  free(UpCaseNick);

  NewUser->NextA = UsersA->Table[PosA];
  UsersA->Table[PosA] = NewUser;

  NewUser->NextN = UsersN->Table[PosN];
  UsersN->Table[PosN] = NewUser;

#ifdef HYBRIDTS4
  NewUser->NextC = UsersC->Table[PosC];
  UsersC->Table[PosC] = NewUser;
#endif /* HYBRIDTS4 */

  UserCount++;
  ServPtr = GetServer(Serv);
  if (ServPtr)
    ServPtr->NumOfUsers++;

  if (LastNickAdded[0] != '\0')
  {
    if (!(UserNode = FindNick(LastNickAdded, UsersN)))
    {
      strcpy(LastNickAdded, Nick);
      return (1);
    }

    if ((UserNode->Clones >= MAXCLONES) && (!(IsOperByAdd(UserNode->UserID, UserNode->Address))))
    {
      switch (CLONERES)
      {
      case 1:
	sendtosock(Sock, ":%s NOTICE %s :WARNING: You have more connections than allowed! Please try to disconnect one of your clients. Thank you.", IrcNick, LastNickAdded);
	sendtosock(Sock, ":%s WALLOPS :WARNING: %s!%s@%s has more connections than allowed!", IrcNick, UserNode->Nick, UserNode->UserID, UserNode->Address);
	break;
      case 2:
	sendtosock(Sock, ":%s KILL %s :Current clone limit is %d !", IrcNick, LastNickAdded, MAXCLONES);
	DeleteUser(LastNickAdded, false, true);
	break;
      case 3:
	sendtosock(Sock, ":%s GLINE * %s@%s %lu :Exceeded %d clones",
	 ServiceName, UserNode->UserID, UserNode->Address, 3600, MAXCLONES);
	sendtosock(Sock, ":%s WALLOPS :NOTICE: auto-glining %s@%s for exceeding %d maximum clones.",
		   IrcNick, UserNode->UserID, UserNode->Address, MAXCLONES);
	sprintf(buff, "%s@%s", UserNode->UserID, UserNode->Address);
	sprintf(reason, "Exceeded %d clones", MAXCLONES);
	ProcessAddGline(buff, 3600, reason);
      }
    }
  }
  strcpy(LastNickAdded, Nick);
  return (1);
}

void DeleteAllUsersByServer(char *ServerName)
{
  int i;
  UserListType UserPtr;
  UserListType TmpPtr;


  for (i = 0; i < USERSEGSIZE; i++)
  {
    UserPtr = UsersN->Table[i];
    while (UserPtr)
    {
      TmpPtr = UserPtr->NextN;
      if (!strcasecmp(UserPtr->Serv, ServerName))
      {
	DeleteUser(UserPtr->Nick, FALSE, FALSE);
      }
      UserPtr = TmpPtr;
    }
  }
}

void DeleteUser(char *Nick, boolean nickchange, boolean sendnotice)
{
  int PosN, PosA;
  unsigned long Key;
  char Str17[18];
  UserListType CurrPtrA, CurrPtrN, TailA = NULL, TailN = NULL;
  ServerListType ServPtr;
  char *UpCaseNick;
  OwnerListType ChanList, TempChanList;
  ChannelListType ChanPtr;
#ifdef HYBRIDTS4
  UserListType CurrPtrC, TailC = NULL;
  int PosC;
#endif /* HYBRIDTS4 */

#ifdef debugon
/* printf("Deleting user %s\n", Nick); */
#endif

  UpCaseNick = UpCase(Nick);
  Key = strtol(Nick, (char **) NULL, STRTOLBASE);
  PosN = Key % USERSEGSIZE;
  free(UpCaseNick);
  if (UsersN->Table[PosN])
  {
    CurrPtrN = UsersN->Table[PosN];
    while (CurrPtrN)
    {
      if (!(strcmp(CurrPtrN->Nick, Nick)))
	break;
      TailN = CurrPtrN;
      CurrPtrN = CurrPtrN->NextN;
    }
  }
  else
    return;
  if (!CurrPtrN)
    return;

  GetStr17(Str17, CurrPtrN->Address);
  Key = strtol(Str17, (char **) NULL, STRTOLBASE);
  PosA = Key % USERSEGSIZE;

  CurrPtrA = UsersA->Table[PosA];
  while (CurrPtrA)
  {
    if (!(strcmp(CurrPtrA->Nick, Nick)))
      break;
    TailA = CurrPtrA;
    CurrPtrA = CurrPtrA->NextA;
  }

#ifdef HYBRIDTS4
  Key = strtol(CurrPtrN->CID, (char **) NULL, STRTOLBASE);
  PosC = Key % USERSEGSIZE;

  CurrPtrC = UsersC->Table[PosC];
  while (CurrPtrC)
  {
    if(!(strcmp(CurrPtrC->Nick, Nick)))
      break;
    TailC = CurrPtrC;
    CurrPtrC = CurrPtrC->NextC;
  }
#endif /* HYBRIDTS4 */

  if (!nickchange)
  {
    if (IsOper(Nick))
      OperCount--;
    ChanList = CurrPtrN->Chans;
    while (ChanList)
    {
      SetLastSeen(Nick, ChanList->Channel, PART);
      if (IsChanOwner(Nick, ChanList->Channel) || IsChanCoOwner(Nick, ChanList->Channel))
      {
	ChanPtr = GetChannel(ChanList->Channel);
	if (ChanPtr->OwnersOnline == 1)
	{
	  ChanPtr->LastAccess = time(NULL);
//	  UpdateChanDatFile(ChanList->Channel);
	  WriteChanOpFile(ChanList->Channel);
	}
	ChanPtr->OwnersOnline--;
      }
      TempChanList = ChanList;
      ChanList = ChanList->Next;
      free(TempChanList->Channel);
      free(TempChanList);
    }
  }
  if (TailN == NULL)
    UsersN->Table[PosN] = CurrPtrN->NextN;
  else
    TailN->NextN = CurrPtrN->NextN;

  if (TailA == NULL)
    UsersA->Table[PosA] = CurrPtrA->NextA;
  else
    TailA->NextA = CurrPtrA->NextA;

#ifdef HYBRIDTS4
  if (TailC == NULL)
    UsersC->Table[PosC] = CurrPtrC->NextC;
  else
    TailC->NextC = CurrPtrC->NextC;
#endif /* HYBRIDTS4 */

  UserCount--;
  if (!strcmp(LastNickAdded, Nick))
    LastNickAdded[0] = '\0';
  if (CurrPtrN->Clones > 0)
    UpdateCloneCount(CurrPtrN->UserID, CurrPtrN->Address, CurrPtrN->Clones - 1, UsersA);

#ifdef notices
  if (!nickchange && sendnotice)
  {
    sendtosock(Sock, ":%s WALLCHOPS #Notices :Signoff %s!%s@%s (%s)",
	IrcNick, Nick, CurrPtrN->UserID, CurrPtrN->Address, CurrPtrN->Serv);
  }
#endif

  ServPtr = GetServer(CurrPtrN->Serv);
  if (ServPtr)
    ServPtr->NumOfUsers--;
  free(CurrPtrN->Address);
  free(CurrPtrN->Serv);
  free(CurrPtrN);
}

void KillUserList()
{

  int i;
  UserListType UserList, TempPtr;

  for (i = 0; i < USERSEGSIZE; i++)
  {

    if (UsersA->Table[i])
    {
      UserList = UsersA->Table[i];
      while (UserList)
      {
	TempPtr = UserList;
	UserList = UserList->NextA;
	DeleteUser(TempPtr->Nick, false, false);
      }
      UsersA->Table[i] = NULL;
    }
  }
  UserCount = 0;
}

void SetAway(char *Nick)
{
  UserListType UserNode;

  if(!(UserNode = FindNick(Nick, UsersN)))
    return;
  UserNode->Modes = UserNode->Modes | UMODEaway;
}

void ClearAway(char *Nick)
{
  UserListType UserNode;
  if(!(UserNode = FindNick(Nick, UsersN)))
    return;
  UserNode->Modes = UserNode->Modes & ~UMODEaway;
}

void ProcessUserMode(char *Nick, char *Modes)
{
/* i'm not realy satisfied with the way this works */

  char ch;
  char Operator = '+';
  int len, i;
  unsigned int UserMode;
  UserListType UserNode;

  if (!(UserNode = FindNick(Nick, UsersN)))
    return;

  UserMode = NOMODES;
  len = strlen(Modes);
  for (i = 0; i < len; i++)
  {
    ch = Modes[i];
    switch (ch)
    {
    case 'o':
      UserMode = UMODEo;
      break;
    case 'd':
      UserMode = UMODEd;
      break;
    case 'w':
      UserMode = UMODEw;
      break;
    case 's':
      UserMode = UMODEs;
      break;
    case 'i':
      UserMode = UMODEi;
      break;
#ifdef AFTERNET
    case 'g':
      UserMode = UMODEgod;
      break;
#else
    case 'g':
      UserMode = UMODEg;
      break;
#endif
    case 'k':
      UserMode = UMODEk;
      break;
    case 'x':
      UserMode = UMODEx;
      break;
    case '-':
      Operator = ch;
      break;
    case '+':
      Operator = ch;
      break;
    default:
      Debug(DBGWARNING, "ERROR: Unknown usermode %c by %s.", ch, Nick);
      break;
    }
    if (Operator == '-')
    {
      UserNode->Modes = UserNode->Modes & ~UserMode;
      if (UserMode == UMODEo)
      {
	OperCount--;
      }
    }
    else if (Operator == '+')
    {
      UserNode->Modes = UserNode->Modes | UserMode;
      if (UserMode == UMODEo)
      {
	OperCount++;
        /* Check if the oper is away for .IRCOPS */
        sendtosock(Sock, ":%s WHOIS %s %s", IrcNick, Nick, Nick);
      }
    }
  }
  if ((UMODEo & UserNode->Modes) != 0)
    return;
}

void ProcessNickChange(char *OldNick, char *NewNick)
{

  UserListType UserNode;
  int Pos;
  unsigned long Key;
  char Nick[MAXNICKLEN], UserID[12], *Address, *Serv, *RealName;
#ifdef HYBRIDTS4
  char CID[MAXCIDLEN];
#endif /* HYBRIDTS4 */
  unsigned char Modes;
  OwnerListType Chans;
  char Level;
  char *Validated;

  if (!(UserNode = FindNick(OldNick, UsersN)))
    return;
  strcpy(UserID, UserNode->UserID);
  Modes = UserNode->Modes;
  Address = (char *) malloc(strlen(UserNode->Address) + 1);
  Chans = UserNode->Chans;
  Serv = (char *) malloc(strlen(UserNode->Serv) + 1);
  RealName = (char *) malloc(strlen(UserNode->RealName) + 1);
  Validated = UserNode->Validated;

  if ((!Address) || (!Serv))
  {
    sendtosock(Sock, ":%s WALLOPS :WARRNING: NOT ABLE TO ALLOCATE MEMORY FOR A NICK CHANGE! THIS WILL CAUSE GENERAL MALFUNCTION!", IrcNick);
    return;
  }
  strcpy(Address, UserNode->Address);
  strcpy(Serv, UserNode->Serv);
  strcpy(RealName, UserNode->RealName);

#ifdef HYBRIDTS4
  strcpy(CID, UserNode->CID);
#endif /* HYBRIDTS4 */

  DeleteUser(UserNode->Nick, true, false);

#ifdef HYBRIDTS4
  AddUser(CID, NewNick, UserID, Address, RealName, Serv, Modes, Chans, Validated);
#else /* HYBRIDTS4 */
  AddUser(NewNick, UserID, Address, RealName, Serv, Modes, Chans, Validated);
#endif /* HYBRIDTS4 */
}


/* EOF */
